import bisect
import numpy
import time
from operator import itemgetter
from .        import __version__
from .field   import FieldList
from .utils   import equals, RTOL, ATOL, CfDict
from .value   import Value


_global_attributes = ('comment',
                      'Conventions',
                      'history',
                      'institution',
                      'references',
                      'source',
                      'title',
                      )

class AggregateError(Exception):
    '''

An exception that indicates that two fields can not be aggregated.
'''
    pass
#--- End: class


def aggregate(fields, rtol=None, atol=None, Ndim=True,
              verbose=False, messages=False, strict_units=False):
    '''
    
**Examples**

>>> len(f)
6
>>> f
[<CF Field: eastward_wind(110, 106)>,
 <CF Field: eastward_wind(110, 106)>,
 <CF Field: eastward_wind(110, 106)>,
 <CF Field: eastward_wind(110, 106)>,
 <CF Field: eastward_wind(110, 106)>,
 <CF Field: eastward_wind(110, 106)>]
>>> cf.aggregate(f)
>>> len(f)
1
>>> f
[<CF Field: eastward_wind(3, 2, 110, 106)>]
'''

    if atol is None:
        atol = ATOL()
    if rtol is None:
        rtol = RTOL()

    equivalent_field_units  = {}
    equivalent_coord_units  = {}
    equivalent_cm_units     = {}
    unique_cell_methods     = []
    unique_transform        = []

    aggregated_fields = FieldList()
        
    aggregating_group = {}

    history = '%s Aggregated by cf-python v%s' % \
        (time.strftime('%Y-%m-%d %H:%M:%S UTC', time.gmtime()),
         __version__)

    for f in fields:

        do_not_aggregate = 0

        # Raise an exception for missing functionality
        for attr in ('ancillary_variables', 'flag_values', 
                     'flag_masks', 'flag_meanings'):
            if hasattr(f, attr):
                if messages:
                    print \
"Can't aggregate %s: Code 1, %s" % (repr(f), attr)
                do_not_aggregate = 1
                break
        #--- End: for

        if do_not_aggregate:
            aggregated_fields.append(f)
            continue

        # Still here? Then find this field's aggregating name
        aggregating_name = f.getattr('standard_name', 
                                     f.getpriv('aggregating_name',
                                               None))
        if aggregating_name is None:
            # This field has no aggregating name, so it can not be
            # aggregated. Put it straight into the output list.
            if messages:
                print \
"Can't aggregate %s: Field has no standard/aggregating name" % repr(f)
            aggregated_fields.append(f)
            continue

        # Still here? Then try to make a hashable representation of
        # the field's space
        meta = CfDict()

        all_standard_names     = []
        meta.nd_standard_names = []
        meta.simple_attrs      = {}
        meta.key               = {}
        meta.datum             = {}
        meta.cell_measures     = [] # List of space keys which are cell measures

        for key, coord in f.space.iteritems():
            
            # Check for cell measures
            if key[0:2] == 'cm':
                meta.cell_measures.append(key)
                continue
             
            # Still here? Then check that there's a standard_name
            try:
                standard_name = coord.standard_name
            except AttributeError:                
                if messages:
                    print \
"Can't aggregate %s: Coordinate has no standard name" % repr(f)
                do_not_aggregate = 1
                break

            # Still here?
            all_standard_names.append(standard_name)        

            meta.key[standard_name] = key

            try:
                if coord.ndim <= 1:
                    dim = f.space.dimensions[key][0]                    
                else:
                    meta.simple_attrs[key] = sorted(coord._simple_attributes().iteritems())
                    meta.nd_standard_names.append(standard_name)
                    meta.datum[standard_name] = None
                    continue
            except AttributeError:
                # A space component has no data
                if messages:
                    print \
"Can't aggregate %s: Coordinate has no data" % repr(f)
                do_not_aggregate = 1
                break

            # Still here?
            meta.simple_attrs[key] = sorted(coord._simple_attributes().iteritems())
            if dim in meta:
                meta[dim]['1d_standard_names'].append(standard_name)
            else:
                # Initialize meta[dim]['1d_standard_names']
                meta[dim] = {'1d_standard_names': [standard_name]}

            try:
                meta.datum[standard_name] = float(coord.varray)
            except TypeError:
                meta.datum[standard_name] = None
        #--- End: for
 
        if do_not_aggregate:
            aggregated_fields.append(f)
            continue

        if len(all_standard_names) > len(meta.key):
            # Non-unique standard names
            aggregated_fields.append(f)
            continue

        # Each dimension in each space has at least one dimension
        # coordinate or 1-d auxiliary coordinate associated with it.
        for dim in f.space.dimension_sizes:
            if dim not in meta:
                # A dimension has no scalar or 1-d coordinates
                if messages:
                    print \
"Can't aggregate %s: Dimension has no 1-d coordinates" %repr(f)
                do_not_aggregate = 1
                break
        #--- End: for

        if do_not_aggregate:
            # A dimension has no scalar or 1-d coordinates
            aggregated_fields.append(f)
            continue

        # Sort each dimension's 1-d coordinates' standard names
        for value in meta.values():
            value['1d_standard_names'].sort()
        meta.nd_standard_names.sort()

        f.setpriv('_space_metadata', meta)

        # ------------------------------------------------------------
        # Create a key entry for the field's Units
        # ------------------------------------------------------------
        field_units = _units(f, aggregating_name,
                             equivalent_field_units, 
                             strict_units=strict_units)

        # ------------------------------------------------------------
        # Create a key entry for the field's cell methods. (Do this
        # after ascertaining that every coordinate has a standard name
        # so we can be sure that the 'name' keys are comparable.)
        # ------------------------------------------------------------
        field_cell_methods = getattr(f, 'cell_methods', None)
        if field_cell_methods is not None:
            # This field has cell methods. If we've seen this
            # equivalent cell_methods before then use the existing one
            # in the key            
            new_cell_methods = 1
            for cell_methods1 in unique_cell_methods:
                if len(field_cell_methods) != len(cell_methods1):
                    continue          
                for cm0, cm1 in zip(field_cell_methods, cell_methods1):
                    keys0 = set(cm0) - set(('dim', 'no_coords'))
                    keys1 = set(cm1) - set(('dim', 'no_coords'))
                    if keys0 != keys1:
                        continue                    
                    for key in keys0:
                        if not equals(cm0[key], cm1[key], rtol=rtol, atol=atol):
                            continue
                new_cell_methods = 0
                field_cell_methods = cell_methods1
                break
            #--- End: for
            if new_cell_methods:
                unique_cell_methods.insert(0, field_cell_methods)
        #--- End: if

        # ------------------------------------------------------------
        # Create a key entry for the field's public attributes
        # ------------------------------------------------------------
        field_simple_attrs = f._simple_attributes()

        for attr in _global_attributes:
            if attr in field_simple_attrs:
                del field_simple_attrs[attr]
        #--- End: for
        field_simple_attrs = tuple(sorted(field_simple_attrs.iteritems()))

        # ------------------------------------------------------------
        # Create a hashable description for each coordinate
        # construct. This comprises public attributes, transforms,
        # units and whether or not it has bounds (and whether or not
        # they are climatological.
        # ------------------------------------------------------------
        coordinate_attrs = []

        # Find any transforms that we haven't seen before
        if hasattr(f.space, 'transform'):
            for t in f.space.transform.values():
                if not t.has_key('grid_mapping_name'):
                    # Can't do anything with formula_terms, yet
                    if messages:
                        print \
"Can't aggregate %s: Coordinate has formula_terms" % repr(f)
                    do_not_aggregate = 1
                    break
                if t not in unique_transform:
                    unique_transform.append(t)
            #--- End: for
            if do_not_aggregate:
                aggregated_fields.append(f)
                continue
        #--- End: if
           
        try:
            unique_transform.extend(t 
                                    for t in f.space.transform.values()
                                    if t not in unique_transform)
        except AttributeError:
            pass

        for standard_name, key in sorted(meta.key.iteritems()):

            # Find any units that we haven't seen before
            coord_units = _units(f.space[key], standard_name,
                                 equivalent_coord_units,
                                 strict_units=strict_units)

            # Transform
            try:
                transform = f.space[key].transform
            except AttributeError:
                transform = 0
            else:
                transform = f.space.transform[transform]
                transform = tuple(sorted(transform.iteritems()))
            #--- End: try

            # Are there bounds? If so, are they climatological?
            bounds      = getattr(f.space[key], 'bounds', 0)
            climatology = 0
            if bounds is not 0:
                bounds = 1                
                climatology = getattr(f.space[key], 'climatology', 0)
                if climatology is not 0:
                    climatology = 1
            #--- End: if

            coordinate_attrs.append(
                (tuple(sorted(f.space[key]._simple_attributes().iteritems())), 
                 bounds,
                 climatology,
                 transform,
                 coord_units)
                )
        #--- End: for
        coordinate_attrs = tuple(coordinate_attrs)
            
        # ------------------------------------------------------------
        # Create a hashable description for each cell measure
        # construct. This comprises units and simple attributes.
        # ------------------------------------------------------------
        cell_measure_attrs = []
        for key in meta.cell_measures:
            # Find any units that we haven't seen before
            cm_units = _units(f.space[key], None, equivalent_cm_units,
                              strict_units=strict_units)
            
            cell_measure_attrs.append(
                cm_units,
                tuple(sorted(f.space[key]._simple_attributes().iteritems())),
                )
        #--- End: for
        cell_measure_attrs =  tuple(sorted(cell_measure_attrs, 
                                           key=itemgetter(0,1))
                                    )
                                    
        # ------------------------------------------------------------
        # Create the final dictionary key
        # ------------------------------------------------------------    
        key = (aggregating_name,
               len(f.space.dimension_sizes),
               field_simple_attrs,
               field_cell_methods,
               field_units,
               coordinate_attrs,
               cell_measure_attrs,
               )

        # ------------------------------------------------------------
        # Append this field to the group of similar fields
        # ------------------------------------------------------------    
        try:
            # Append this field to the group of fields with this
            # aggregating name
            aggregating_group[key].append(f)
        except KeyError:
            # Initialize the group of fields with this aggregating
            # name
            aggregating_group[key] = FieldList(f)
    #--- End: for    

    for k, g in aggregating_group.iteritems(): #aggregating_group.itervalues():

        name = g[0].getattr('standard_name', g[0].getpriv('aggregating_name', None))

        original_len_g = len(g)

        for f in [g]:
            M = 0
            while 1:
                len_f = len(f)
                i = 0
                while i < len(f):
                    j = i + 1
                    while j < len(f):
                        try:
                            # Attempt to aggregate two fields
                            f[i] = _aggregate_2_fields(f[i], f[j],
                                                       rtol=rtol, atol=atol, 
                                                       Ndim=Ndim)
                            # Successfully aggregated fields f[i] and f[j]
                            # into f[i], so get rid of f[j]
                            f.pop(j)
                        except AggregateError as error:
                            # Could not aggregate fields f[i] and f[j]
                            j += 1
                            if messages:
                                print \
"Can't aggregate %s: %s" % (name, error)
                    #--- End: while 
                    i += 1
                #--- End: while 
                if not Ndim or len(f) == len_f:           
                    break
                M += 1
            #--- End: while
            aggregated_fields.extend(f)
            if verbose:
                print "Aggregated %d '%s' field%s to %d field%s" % \
                    (original_len_g, name,
                     ('s' if original_len_g>1 else ''),                 
                     len_f, ('s' if len_f>1 else ''))
        #--- End: for

        # Set the history attribute
        if original_len_g > 1:
            f.setattr('history', history)
        
        # Delete the space metadata
        f.delpriv('_space_metadata')
    #--- End: for

    # Update input fields in-place
    fields[:] = aggregated_fields 
#--- End: def

def _units(var, aggregating_name, units, strict_units=False):
    '''

Return an equivalent Units object that has been encountered before, if
one exists. Otherwise return this variable's Units object and update
the `units` dictionary. If strict_units is False then assume that
variables with the same standard name but missing units all have equal
units.
'''
    var_units = var.Units

    if units.has_key(aggregating_name):
        if var_units:
            for u in units[aggregating_name]:
                if var_units.equivalent(u):
                    return u
            #--- End: for
            # Still here?
            units[aggregating_name].append(var_units)
#        elif var.dtype.kind = 'S':
#            return None
        elif not strict_units or var.dtype.kind == 'S':
            return None
        else:
            return var_units
    else:
        if var_units:
            units[aggregating_name] = [var_units]
            return var_units
#        elif var.dtype.kind = 'S':
#            return None
        elif not strict_units or var.dtype.kind == 'S':
            return None
        else:
            return var_units
#--- End: def   

def _equivalent_data(var0, var1, 
                     var0_dimensions, var1_dimensions,
                     meta0, meta1, key0, key1,
                     match_for_dim1, rtol=None, atol=None):
    '''

Return True if data of the two input variables are
equivalent. Otherwise, return False.
'''  
    # ----------------------------------------------------------------
    # Check that the variables' data sizes are the same
    # ----------------------------------------------------------------
    size = var0.size
    if size != var1.size:
        return

    # ----------------------------------------------------------------
    # Compare the variables' data, after making them compatible
    # ----------------------------------------------------------------
    x = var0._data.concatenate()._list[0]
    y = var1._data.concatenate()._list[0]

    new_order  = []
    full_slice = [slice(None)] * y.ndim
    _transpose = 0
    y1 = y

    ndim    = x.ndim
    x_order = range(ndim)

    if ndim > 1:
        # Reorder y so that its axes are in the same order as x
        new_order = [var0_dimensions.index(match_for_dim1[dim1])
                     for dim1 in var1_dimensions]
        if new_order != x_order:
            _transpose = 1
            y1 = numpy.transpose(y, axes=new_order)
    else:
        new_order = x_order
    #--- End: if

    if size > 1:
        # Reverse any of y's axes thst increase in the opposite
        # direction to the same axis in x.
        _reverse = 0
        for i, dim in enumerate(var0._data.order):
            if (var0._data.direction[dim] != 
                var0._data.direction[var1._data.order[new_order[i]]]):
                _reverse = 1
                full_slice[i] = slice(None, None, -1)
        #--- End: for
        if _reverse:
            y1 = y1[tuple(full_slice)]

        if not (Value(x, cfunits=var0.Units) == Value(y1, cfunits=var1.Units)).all():
            return

#        if not array_allclose(x, Units.conform(y1, var1.Units, var0.Units), rtol=rtol, atol=atol):
#            return
    else:
        if not (Value(x, cfunits=var0.Units) == Value(y1, cfunits=var1.Units)).all():
            return
#        if not numpy.allclose(x, Units.conform(y1, var1.Units, var0.Units), rtol=rtol, atol=atol):
#            return
    #--- End: if

    # ----------------------------------------------------------------
    # Compare the variables' bounds, after making them compatible
    # ----------------------------------------------------------------
    try:
        x = var0.bounds._data.concatenate()._list[0]
        y = var1.bounds._data.concatenate()._list[0]            
    except AttributeError: 
        pass
    else:
        y1 = y
        if _transpose:
            new_order.append(var1.ndim-1)
            y1 = numpy.transpose(y1, axes=new_order)

        if (var0.bounds._data.direction['bounds'] != 
            var1.bounds._data.direction['bounds']):
            full_slice.append(slice(None, None, -1))
            y1 = y1[tuple(full_slice)]
            
        if not (Value(x, cfunits=var0.Units) == Value(y1, cfunits=var1.Units)).all():
            return
#        if not numpy.allclose(x, Units.conform(y1, var1.Units, var0.Units), rtol=rtol, atol=atol):
#            return
    #--- End: try

    # Still here? Then these two variables have equivalent data
    return 1
#-- End: def

def _dimensions_are_paired(dimensions0, dimensions1, match_for_dim1):
    '''
'''
    if len(dimensions0) != len(dimensions1):
        # Components have different ranks
        return 
    
    # Still here?
    for dim1 in dimensions1:
        if match_for_dim1[dim1] not in dimensions0:
            # Component's dimensions are not matching pairs
            return 
    #--- End: for

    # Still here?
    return 1
#--- End: def

def _aggregate_2_fields(field0, field1,
                        rtol=None, atol=None, Ndim=1):

#- Both fields have the same standard_name, etc.

#- Each dimension in each space has at least one scalar or 1-d
#  coordinate (dimension or auxiliary) and each such coordinate's
#  standard_name is unique within its space.


    if field0.ndim < field1.ndim:
        field0, field1 = field1, field0

    space0 = field0.space
    space1 = field1.space

    meta0 = field0.getpriv('_space_metadata')
    meta1 = field1.getpriv('_space_metadata')

    # ================================================================
    # For each dimension in one of the spaces, there is a 'matching
    # dimension' in the other space with exactly the same list of
    # scalar or 1-d coordinate (dimension or auxiliary)
    # standard_names, although a standard_name may belong to different
    # types of coordinate in the two spaces. Call these coordinates a
    # pair of 'matching coordinates' for the matching dimensions
    # ================================================================
    matching_dimensions = {}
    match_for_dim1      = {'bounds': 'bounds'}
    dims1 = meta1.keys()
    
    for dim0 in meta0:        
        
        found_matching_dimension = 0
        standard_names0 = meta0[dim0]['1d_standard_names']        
        for dim1 in dims1:
            if standard_names0 == meta1[dim1]['1d_standard_names']:
                found_matching_dimension = 1
                dims1.remove(dim1)
                break
        #--- End: for
        if not found_matching_dimension:
            raise AggregateError, \
"Not all dimensions can be matched"

        # Add to the dictionaries which, given a dimension from one
        # field, easily return its matching dimension from the other
        # field. This is used to reorder variables from field1 when
        # checking that they have equivalent values and for inserting
        # one variable into another
        match_for_dim1[dim1] = dim0

        # Create a list of pairs of matching coordinates for this pair
        # of matching dimensions
        matching_dimensions[(dim0, dim1)] = [(meta0.key[sn], meta1.key[sn])
                                             for sn in standard_names0]
    #--- End: for

    # Initialize the list of pairs of matching N-d constructs. This
    # comprises all space components except 1-d coordinates. I.e. all
    # cell measures and all N-d (N>1) auxiliary coordinates.
    matching_nd_constructs = []

    # Find matching pairs of N-d auxiliary coordinate constructs
    for sn in meta0.nd_standard_names:
    
        # Check that potentially matching N-d auxiliary coordinate
        # constructs span equivalent dimensions
        key0 = meta0.key[sn]
        key1 = meta1.key[sn]

        if not _dimensions_are_paired(space0.dimensions[key0],
                                      space1.dimensions[key1],
                                      match_for_dim1):
            raise AggregateError, \
"N-d construct has no match"
                   
        # Add this pair to the list of matching N-d constructs
        matching_nd_constructs.append((key0, key1))
    #--- End: for

    # Find matching pairs of cell measure constructs
    cell_measures1 = meta1.cell_measures[:]
    for key0 in meta0.cell_measures:
        found_pair_of_cell_measures = 0
        for key1 in cell_measures1[:]:
            if _dimensions_are_paired(space0.dimensions[key0],
                                      space1.dimensions[key1],
                                      match_for_dim1):
                if found_pair_of_cell_measures:
                    raise AggregateError, \
"cm match oof"
                # check units/pubartrs here. ???
                found_pair_of_cell_measures = 1
                cell_measures1.remove(key1)
                continue
        #--- End: for
        if not found_pair_of_cell_measures:
            raise AggregateError, \
"cm match oof2"
            
        # Add this pair to the list of matching 
        matching_nd_constructs.append((key0, key1))
    #--- End: for

    # ================================================================
    # There is EXACTLY ONE pair of matching dimensions for which at
    # least one pair of matching 1-d coordinates do NOT share the same
    # values for their arrays and bounds. Call this dimension the
    # 'aggregating dimension' in each field.
    # ================================================================
    aggregating_dims = None
#    no_aggregating_dimension_coordinates = 1

    for (dim0, dim1), matching_1d_coords in matching_dimensions.iteritems():

        if aggregating_dims and (space0.dimension_sizes[dim0] != 
                                 space1.dimension_sizes[dim1]):
#            if aggregating_dims:
            raise AggregateError, \
"More than one pair of matching dimensions have different sizes"
#            else:
#                aggregating_dims = 1 #(dim0, dim1)
#                continue
        #--- End: if

#        aggregating_dims = None
#        no_aggregating_dimension_coordinates = 1
#        for key0, key1 in matching_1d_coords:
#           if key0[0] == 'd' or key1[0] == 'd':
#               no_aggregating_dimension_coordinates = 0
#               break
#        #--- End: for
#
#        if no_aggregating_dimension_coordinates:
#            if not _equivalent_data(space0[key0], space1[key1],
#                                    space0.dimensions[key0], 
#                                    space1.dimensions[key1], 
#                                    meta0, meta1, key0, key1,
#                                    match_for_dim1, rtol=rtol, atol=atol):
#
#                if aggregating_dims:
#                    raise AggregateError, \
#"More than one pair of dimensions have 1-d coordinates with different values"
#                else:
#                    # The name 'aggregating' is not already defined, this
#                    # pair of matching dimensions comprise the aggregating
#                    # dimensions, provided we don't find another pair with
#                    # unequal data.
#                    aggregating_dims = (dim0, dim1)
#                    continue
#        #--- End: if

        # Still here? Then this pair of matching dimensions has at
        # least one dimension coordinate associated with it. Therefore
        # we need to check for equivalent values.
        for key0, key1 in matching_1d_coords:

            if _equivalent_data(space0[key0], space1[key1],
                                space0.dimensions[key0], 
                                space1.dimensions[key1], 
                                meta0, meta1, key0, key1,
                                match_for_dim1, rtol=rtol, atol=atol):
                # These two coordinates have equivalent values
                continue                

            # Still here? Then this pair of matching coordinates have
            # non-equivalent values.
            if aggregating_dims:
                raise AggregateError, \
"More than one pair of matching dimensions have 1-d coordinates with different values"
            else:
                # The aggregating_dims has not yet been set, so this
                # pair of matching dimensions could comprise the
                # aggregating dimensions, provided we don't find
                # another pair of matching dimensions with unequal
                # data.
                aggregating_dims = (dim0, dim1)
                break
        #--- End: for
    #--- End: for

    # ----------------------------------------------------------------
    # Find which pairs of matching coordinates/cell measures (1-d and
    # Nd) span their aggregating dimension and which do not. Also
    # find, for each field, the key of its aggregating dimension.
    # ----------------------------------------------------------------
    if not aggregating_dims:
        # The name 'aggregating' is not defined because each pair of
        # matching coordinates has the same data values
        raise AggregateError, \
"No pair of aggregating dimensions"

    spanning_1d_coordinates = matching_dimensions[aggregating_dims]

    key0, key1 = spanning_1d_coordinates[0]
    adim_key0 = space0.dimensions[key0][0]
    adim_key1 = space1.dimensions[key1][0]

    spanning_nd_constructs     = []
    non_spanning_nd_constructs = []
    for key0, key1 in matching_nd_constructs:
        if adim_key0 in space0.dimensions[key0]:            
            spanning_nd_constructs.append((key0, key1))
        else:
            non_spanning_nd_constructs.append((key0, key1))
    #--- End: for

    # ================================================================
    # If either or both aggregating dimensions have a dimension
    # coordinate then there are no values for it and its matching
    # coordinate which appear in both spaces (if they run in different
    # senses or their ranges overlap, the aggregated dimension will
    # need to be sorted). [All other coords are concatenated.]
    # ================================================================

    # ----------------------------------------------------------------
    # Check that neither or both aggregating dimensions have a
    # dimension coordinate.
    # ----------------------------------------------------------------
    dim_coords = None
    for key0, key1 in spanning_1d_coordinates:
        if key0[0] == 'd' and key1[0] == 'd':
            # Found a pair of dimension coordinates
            dim_coords = (key0, key1)
            break
        if key0[0] == 'd' or key1[0] == 'd':
            raise AggregateError, \
"Can not join a dimension and an auxiliary coordinate"
    #--- End: for

    # ----------------------------------------------------------------
    # Find the insertion indices
    # ----------------------------------------------------------------
    if not dim_coords:
        # ------------------------------------------------------------
        # Neither field's aggregating dimension has a dimension
        # coordinate (because name 'dim_coord' is not
        # defined). Therefore the fields will be simply concatenated
        # along the aggregating dimension
        # ------------------------------------------------------------
        indices = None
    else:        
        # ------------------------------------------------------------
        # The aggregating dimensions of both fields have a dimension
        # coordinate. Therefore we need to find the insertion index in
        # field0 for each data element in field1.
        # ------------------------------------------------------------
        dim0 = space0[dim_coords[0]]
        dim1 = space1[dim_coords[1]].copy()
        dim1.Units = dim0.Units.copy()

        dim0._data.conform()
        dim1._data.conform()        
        try:
            dim1.bounds._data.conform()
        except AttributeError:
            pass

        mins0 = []
        maxs0 = []
        for array in dim0._data._list:
            try:
                first = numpy.asscalar(array)
                mins0.append(first)
                maxs0.append(first)
            except ValueError:
                size       = array.size
                first_last = array[slice(0, size, size-1)]
                mins0.append(min(first_last))
                maxs0.append(max(first_last))
        #--- End: for

        # Set the reversed_ flag to signify whether or not we had to
        # reverse the mins0 and maxs0 lists in order to get them into
        # ascending sequences.
        try:
            reversed_ = not dim0._data.direction.values()[0]
        except AttributeError:
            reversed_ = not dim0._data.direction
        if reversed_:
            mins0.reverse()
            maxs0.reverse()
        
        indices = []
        for i, array in enumerate(dim1._data._list):
            try:
                first = numpy.asscalar(array)
                min1  = first
                max1  = first
            except ValueError:
                size       = array.size
                first_last = array[slice(0, size, size-1)]
                min1       = min(first_last)
                max1       = max(first_last)

            index = bisect.bisect(mins0, min1)

            if reversed_:
                indices.append(len(mins0)-index)
            else:
                indices.append(index)

            # Update mins0 and maxs0
            mins0.insert(index, min1)
            maxs0.insert(index, max1)

            # Check for overlapping dimension coordinates (not allowed
            # for now). This check also serves to check for any shared
            # dimension coordinate values.
            if index == len(maxs0)-1:
                if min1 <= maxs0[index-1]:
                    raise AggregateError, \
"Overlapping or equal '%s' dimension coordinate values" % dim0.standard_name
            elif index == 0:
                if max1 >= mins0[index+1]:
                    raise AggregateError, \
"Overlapping or equal '%s' dimension coordinate values" % dim0.standard_name
            else:
                 if min1 <= maxs0[index-1] or max1 >= mins0[index+1]:
                     raise AggregateError, \
"Overlapping or equal '%s' dimension coordinate values" % dim0.standard_name
            #--- End: if

            # Check for overlapping dimension coordinate bounds. This
            # is a bit strict for now
            try:               
#                dim1.bounds._data.conform(i)
                bounds1 = dim1.bounds._data._list[i]
            except AttributeError:
                # Dimension coordinate knows no bounds
                pass
            else: 
                # Dimension coordinate has bounds
                if index == len(maxs0)-1:
                    if numpy.amin(bounds1) <= maxs0[index-1]:
                        raise AggregateError, \
"Overlapping (ish) '%s' dimension coordinate bounds" % dim0.standard_name
                elif index == 0:
                    if numpy.amax(bounds1) >= mins0[index+1]:
                        raise AggregateError, \
"Overlapping (ish) '%s' dimension coordinate bounds" % dim0.standard_name
                else:  
                    if (numpy.amin(bounds1) <= maxs0[index-1] or
                        numpy.amax(bounds1) >= mins0[index+1]):
                        raise AggregateError, \
"Overlapping (ish) '%s' dimension coordinate bounds" % dim0.standard_name
            #--- End: try
        #--- End: for
    #--- End: if

    # ================================================================
    # Check that N-d constructs which don't span the aggregating
    # dimension have the same data values
    # ================================================================
    for key0, key1 in non_spanning_nd_constructs:
        if not _equivalent_data(space0[key0], space1[key1],
                                space0.dimensions[key0], space1.dimensions[key1],
                                meta0, meta1, key0, key1,
                                match_for_dim1, rtol=rtol, atol=atol):
            # These two coordinates do not have equivalent values
            raise AggregateError, \
"Non spanning N-d constructs have different values"
    #--- End: for

    # ----------------------------------------------------------------
    # Check that all coordinates/field data/cell measures which span
    # the aggregating dimension are not already aggregated across a
    # different dimension and that, within each field, they all have
    # the same number of data elements. (Note that we're not checking
    # that each data element has an equivalent aggregating dimension
    # size. I'm assuming that this will be the case. But is this a
    # reasonable assumption?)
    # ----------------------------------------------------------------
    for adim_key, field in zip((adim_key0, adim_key1),
                               (field0   , field1)):

        try:
            adim = field.space.dimensions['data'].index(adim_key)
        except ValueError:
            # Field's data doesn't span the aggregating dimension
            try:
                field._data.adim
                # Field is aggregated along a dimension other than the
                # current aggregating dimension
                if not Ndim:
                    raise AggregateError, \
"Field's data are already aggregated along a different dimension" 
                field._concatenate_data()
            except AttributeError:
                # Field's data is not aggregated along any dimension
                pass
        else:
            # Field's data does span the aggregating dimension
            try:            
                if field._data.adim != adim:
                    # Field is aggregated along a dimension other than
                    # the current aggregating dimension
                    if not Ndim:
                        raise AggregateError, \
"Field's data are already aggregated along a different dimension"
                    field._concatenate_data()
            except AttributeError:
                # Field's data is not aggregated along any dimension
                pass
        #--- End: try
    #--- End: for
  
    adim_key   = adim_key0
    n_elements = len(field0._data)
    for i, space in enumerate((space0, space1)):
    
        # Check spanning 1-d coordinates
        for key_pair in spanning_1d_coordinates:
            if len(space[key_pair[i]]._data) != n_elements:
                raise AggregateError, \
"1-d coordinate has the wrong number of data elements %d %d" % (len(space[key_pair[i]]._data),n_elements)
        #--- End: for

        # Check spanning N-d constructs
        for key_pair in spanning_nd_constructs:
            key  = key_pair[i]
            if len(space[key]._data) != n_elements:
                raise AggregateError, \
"N-d construct has the wrong number of data elements"

        adim_key   = adim_key1
        n_elements = len(field1._data)
    #--- End: for

    # ================================================================
    # All of the aggregation tests have been passed and so field1 will
    # now be aggregated into field0 *in place*. Field1 might also be
    # changed *in place*.
    # 
    # There are no more trapped AggregateError exceptions from this
    # point onwards, as it is assumed that the fields have passed all
    # of the aggregation checeks and are therefore aggregatable.
    # ================================================================
    # Get 1-d coordinates ready for aggregation
    for key_pair in spanning_1d_coordinates:
        c = space0[key_pair[0]]
        try:
            c._data.adim
            continue
        except AttributeError:                
            pass

        c._data.adim = 0                            
        try:
            c.bounds._data.adim = 0
        except AttributeError:
            pass

        # if the coordinate in field0 is a a scalar then turn it into
        # a 1-d coordinate
        if not c._data.shape:
            c._data.shape     = (1,)
            c._data.direction = {adim_key0: c._data.direction}
            c._data.order     = [adim_key0]
            try:
                b = c.bounds._data
            except AttributeError:
                pass
            else:
                b.shape                = (1, 2)
                b.direction[adim_key0] = c._data.direction[adim_key0]
                b.order.insert(0, adim_key0)
        #--- End: if
    #--- End: for

    # Get N-d constructs ready for aggregation
    adim_key = adim_key0
    for i, space in enumerate((space0, space1)):
        for key_pair in spanning_nd_constructs:
            key = key_pair[i]
            c   = space[key]
            try:  
                c._data.adim
            except AttributeError:
                c._data.adim = space.dimensions[key].index(adim_key)            
                try:
                    c.bounds._data.adim = c._data.adim
                except AttributeError:
                    pass
            #--- End: try
        #--- End: for
        adim_key = adim_key1
    #--- End: for

    # Get the data ready for aggregation
    try:            
        field0._data.adim
    except AttributeError:
        try:
            # The field's data spans the aggregating dimension
            field0._data.adim = space0.dimensions['data'].index(adim_key0)
        except ValueError:
            # The field's data doesn't currently span the
            # aggregating dimension, so it needs to be inserted
            field0.space.dimensions['data'].insert(0, adim_key0)

            field0._data.adim  = 0
            field0._data.shape = (1,) + field0._data.shape
            field0._data.order.insert(0, adim_key0)
            try:
                field0._data.direction.update(space0[adim_key0]._data.direction)
            except AttributeError:
                # Field's data is currently a scalar
                field0._data.direction = field0.space[adim_key0]._data.direction.copy()
        #--- End: try
    #--- End: try

    # Insert 1-d coordinates which span the aggregating dimension
    for key0, key1 in spanning_1d_coordinates:
        space0[key0]._data_insert(indices, space1[key1],
                                  space0.dimensions[key0], space1.dimensions[key1],
                                  match_for_dim1)
        # Update dimension_sizes
        space0.dimension_sizes[adim_key0] = space0[key0].size

    # Insert N-d constructs which span the aggregating dimension
    for key0, key1 in spanning_nd_constructs:
        space0[key0]._data_insert(indices, space1[key1],
                                  space0.dimensions[key0], space1.dimensions[key1],
                                  match_for_dim1)
 
    # Insert data
    field0._data_insert(indices, field1,
                        space0.dimensions['data'], space1.dimensions['data'],
                        match_for_dim1)

    # ----------------------------------------------------------------
    # Return the aggregated field
    # ----------------------------------------------------------------
    return field0
#--- End: def
