#import numpy
from .utils    import RTOL, ATOL, equals, CfList, _parse_index
from .units    import Units
from .value    import Value
from .variable import Variable, VariableList, SliceVariable

# ====================================================================
#
# Variable objects
#
# ====================================================================

class Coordinate(Variable):
    '''

A CF coordinate object.

Refer to the :class:`~cf.Variable` and :class:`~cf.Field` classes and
the :mod:`cf` package for details.


**Reserved public attributes:**

In addition to the reserved public attributes inherited from the
:class:`~cf.Variable` class, the following are also reserved public
attributes:

===================== ========================
Attribute             Description             
===================== ========================
:attr:`axis`          As in the CF convention.
:attr:`positive`      As in the CF convention.
===================== ========================


**Reserved attributes:**

In addition to the reserved attributes inherited from the
:class:`~cf.Variable` class, the following attributes have special
definitions and it is recommended to set them accordingly.

=================== =================================================
Attribute           Description
=================== =================================================
:attr:`bounds`      The coordinate's :class:`~cf.CoordinateBounds`
                    object.

:attr:`climatology` An indicator to specify if the coordinate's
                    bounds are intervals of climatological time.

:attr:`transform`   A pointer to the coordinate's
                    :class:`~cf.Transform` object.
=================== =================================================

'''

    _special_attributes = Variable._special_attributes.union(set(('bounds',
                                                                  'climatology',
                                                                  'transform',
                                                                  )))
   
#
    @property
    def slice(self):
        '''

Slice the coordinate's data, returning a new coordinate. The
coordinate' bounds are also sliced.

**Examples:**

>>> c
<CF Coordinate: longitude(96)>
>>> c.bounds
<CF CoordinateBounds: longitude(96,2)>
>>> d=c.slice[1:]
>>> d
<CF Coordinate: longitude(95)>
>>> d.bounds
<CF CoordinateBounds: longitude(95,2)>
'''
        return SliceCoordinate(self)
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: Units
    # ----------------------------------------------------------------
    @property
    def Units(self):
        '''
The 'Units' object.
'''
        return Variable.Units.fget(self)
    #--- End: def

    @Units.setter
    def Units(self, value):
        '''
Set the 'Units' object.
'''
        Variable.Units.fset(self, value)

        # Set the Units object on the bounds
        try:
            self.bounds.Units = value.copy()
        except AttributeError:
            pass
    #--- End: def
    
    @Units.deleter
    def Units(self):
        '''
Delete the 'Units' object.
'''
        Variable.Units.fdel(self)
        
        # Delete the Units object from the bounds
        try:
            del self.bounds.Units
        except AttributeError:
            pass
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: units (a special attribute)
    # ----------------------------------------------------------------
    # DCH possible inconsistency when setting self.Units.units
    @property
    def units(self):
        '''
The CF 'units' attribute.
'''
        return Variable.units.fget(self)
    #--- End: def

    @units.setter
    def units(self, value):
        '''
Set the 'units' attribute
'''
        Variable.units.fset(self, value)

        # Set the units on the bounds
        try:
            self.bounds.units = value 
        except AttributeError:
            pass
    #--- End: def
    
    @units.deleter
    def units(self):
        '''
Delete the 'units' attribute
'''
        Variable.units.fdel(self)
        
        # Delete the units from the bounds
        if hasattr(self, 'bounds'):
            del self.bounds.units
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: calendar (a special attribute)
    # ----------------------------------------------------------------
    @property
    def calendar(self):
        '''
The CF 'calendar' attribute.
'''
        return Variable.calendar.fget(self)
    #--- End: def

    @calendar.setter
    def calendar(self, value):
        '''
Set the 'calendar' attribute
'''
        Variable.calendar.fset(self, value)
        
        # Set the calendar on the bounds
        try:
            self.bounds.calendar = value
        except AttributeError:
            pass
    #--- End: def

    @calendar.deleter
    def calendar(self):
        '''
Delete the 'calendar' attribute
'''
        Variable.calendar.fdel(self)
        
        # Delete the calendar from the bounds
        try:
            del self.bounds.calendar
        except AttributeError:
            pass
    #--- End: def
   
    # ----------------------------------------------------------------
    # Property attribute: axis (a simple attribute)
    # ----------------------------------------------------------------
    @property
    def axis(self):
        '''
The CF 'axis' attribute. 
'''
        return self.getattr('axis')
    #--- End: def
    @axis.setter
    def axis(self, value):
        '''
Set the axis attribute
'''
        self.setattr('axis', value)
    #--- End: def

    @axis.deleter
    def axis(self):        
        '''
Delete the axis attribute
'''
        self.delattr('axis')
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: positive (a simple attribute)
    # ----------------------------------------------------------------
    @property
    def positive(self):
        '''
The CF 'positive' attribute. 
'''
        return self.getattr('positive')
    #--- End: def

    @positive.setter
    def positive(self, value):
        '''
Set the positive attribute. 
'''
        self.setattr('positive', value)
    #--- End: def

    @positive.deleter
    def positive(self):
        '''
Delete positive attribute. 
'''
        self.delattr('positive')
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: month_lengths (a simple attribute)
    # ----------------------------------------------------------------
    @property
    def month_lengths(self):
        '''
The CF 'month_lengths' attribute. 
'''
        return self.getattr('month_lengths')
    #--- End: def

    @month_lengths.setter
    def month_lengths(self, value):
        '''
Set the month_lengths attribute. 
'''
        self.setattr('month_lengths', value)
    #--- End: def

    @month_lengths.deleter
    def month_lengths(self):
        '''
Delete month_lengths attribute. 
'''
        self.delattr('month_lengths')
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: leap_year (a simple attribute)
    # ----------------------------------------------------------------
    @property
    def leap_year(self):
        '''
The CF 'leap_year' attribute. 
'''
        return self.getattr('leap_year')
    #--- End: def

    @leap_year.setter
    def leap_year(self, value):
        '''
Set the leap_year attribute. 
'''
        self.setattr('leap_year', value)
    #--- End: def

    @leap_year.deleter
    def leap_year(self):
        '''
Delete leap_year attribute. 
'''
        self.delattr('leap_year')
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: bounds (a special attribute)
    # ----------------------------------------------------------------
    @property
    def bounds(self):
        '''
The coordinate's bounds object.
        '''
        return self._get_special_attr('bounds')
    #--- End: def
    @bounds.setter
    def bounds(self, value): self._set_special_attr('bounds', value)
    @bounds.deleter
    def bounds(self):        self._del_special_attr('bounds')

    # ----------------------------------------------------------------
    # Property attribute: climatology (a special attribute)
    # ----------------------------------------------------------------
    @property
    def climatology(self):
        '''

Indicator of whether the coordinate's bounds represent climatological
time intervals.
        '''
        return self._get_special_attr('climatology')
    #--- End: def
    @climatology.setter
    def climatology(self, value): self._set_special_attr('climatology', value)
    @climatology.deleter
    def climatology(self):        self._del_special_attr('climatology')
   
    # ----------------------------------------------------------------
    # Property attribute: transform
    # ----------------------------------------------------------------
    @property
    def transform(self):
        '''
Pointer to coordinate's transfrom.
        '''
        return self._getter('transform')
    #--- End: for
    @transform.setter
    def transform(self, value): self._setter('transform', value)
    @transform.deleter
    def transform(self):        self._deleter('transform')

    def dump(self, id=None, omit=()): 
        '''

Return a string containing a full description of the coordinate.

:keyword id: *Optional*. Set the common prefix of component names. If
             None then defaults to the class name.

:type id: str

:keyword omit: *Optional*. Omit the given attributes from the
               description.

:type omit: sequence


:returns: A string containing the description of the coordinate.

.. seealso:: :func:`cf.dump`
'''
        if id is None:
            id = self.__class__.__name__       

        default = getattr(self, 'ncvar', '')

        string = ['%s coordinate' % self.name(default=default, long_name=True)]
        string.append(''.ljust(len(string[0]), '-'))
        
        string.append(super(Coordinate,self).dump(id=id, omit=omit+('bounds',
                                                                    'transform',
                                                                    )))

        if hasattr(self, 'bounds'):
           string.append(self.bounds.dump(id=id+'.bounds', omit=omit+('bounds',
                                                                      'Units')))

        if hasattr(self, 'transform'):
            string.append('%s.transform = %s\n' % (id,
                                                   repr(self.transform)))
            
        return '\n'.join(string)
    #--- End: def

    def _data_insert(self, indices, other,
                     self_dimensions, other_dimensions, 
                     match_for_dim1):
        '''
'''
        # ------------------------------------------------------------
        # Check that either both coordinates have bounds or neither
        # coordinate has bounds
        # ------------------------------------------------------------
        try:
            self_bounds = self.bounds
        except AttributeError:
            try:
                other.bounds
                raise AttributeError, "Can't insert coordinate data: Missing bounds"
            except AttributeError:
                # Neither coordinate has bounds
                pass
        else:
            try:
                other_bounds = other.bounds
                # Both coordinates have bounds
            except AttributeError:
                raise AttributeError, "Can't insert coordinate data: Missing bounds"
        #--- End: try

        # Insert the coordinate's data
        super(Coordinate, self)._data_insert(indices, other,
                                             self_dimensions,
                                             other_dimensions,
                                             match_for_dim1)

        try:
            other_bounds
        except NameError:
            pass
        else:
            # Insert the coordinate's bounds
            self_bounds._data_insert(indices, other.bounds,
                                     self_dimensions+['bounds'], 
                                     other_dimensions+['bounds'], 
                                     match_for_dim1)        
        #--- End: try

    #--- End: def

    def _concatenate_data(self):
        '''

Concatenate a coordinate's data
'''
#        print 'dch _concatenate_data ',self.__class__.__name__
        super(Coordinate, self)._concatenate_data()
        try: 
            # Concatenate the coordinate's bounds
            self.bounds._concatenate_data()
        except AttributeError:
            # The coordinate knows no bounds
            pass
    #--- End: for
 
    def mask(self, m):
        '''
mask coordinate data
'''
        super(Coordinate, self).mask(m)
        try: 
            # Concatenate the coordinate's bounds
            self.bounds.mask(m)
        except AttributeError:
            # The coordinate knows no bounds
            pass
    #--- End: for
 
#--- End: class

class CoordinateBounds(Variable):
    '''
    
A CF coordinate's bounds object containing cell boundaries or
intervals of climatological time. The containing coordinate's
`climatology` attribute indicates which type of bounds are present.

Refer to the parent :class:`~cf.Variable` class for all details.

'''
    pass
#--- End: class


# ====================================================================
#
# Variable list objects
#
# ====================================================================

class CoordinateList(VariableList):
    '''

An ordered sequence of coordinates stored in a single, list-like
object. Refer to Variable Lists in the :mod:`cf` package for details.

**Methods and attributes defined here:**
'''
    pass
#--- End: class


# ====================================================================
#
# Slice objects
#
# ====================================================================

class SliceCoordinate(SliceVariable):

    __slots__ = []

    def __getitem__(self, index):
        '''
x.__getitem__(index) <==> x[index]
'''
        coord = self.variable

        # Copy the coordinate, omitting the data
        #dch but the bounds/climatology are still deep copied .....
        new = coord.copy(data=False)

        # Parse the index
        index, dummy = _parse_index(coord, index)

        try:
            new._data = coord._data.slice[index]
        except AttributeError:
            pass

        # Slice the bounds, if there are any
        try:
            new.bounds = coord.bounds.slice[index+(slice(None),)]
        except AttributeError:
            try:
                new.climatology = coord.climatology.slice[index+(slice(None),)]
            except AttributeError:
                pass
      
        # Return the new coordinate
        return new
    #--- End: def

#--- End: class
