import copy
import numpy
import re
from .utils import RTOL, ATOL, equals, CfList, CfDict, _parse_index
from .units import Units
from .value import Value
from .variable   import (Variable, VariableList, 
                         SliceVariable, SliceVariableList)
from .coordinate import CoordinateList
from .space      import Space

# ====================================================================
#
# Variable objects
#
# ====================================================================

class Field(Variable):
    '''

Represent a field construct according to the CF data model.

Refer to the :class:`~cf.Variable` class for details.

Refer to the :mod:`cf` module for details on data access by indexing
and data storage.


:keyword public_attributes: *Optional*. The new instance is given the
                            public attributes named and valued from
                            the key, value pairs of this
                            dictionary. The values are deep copied.

:type public_attributes: dict

:keyword private_attributes: *Optional*. The new instance is given the
                             private attributes named and valued from
                             the key, value pairs of this
                             dictionary. The values are deep copied.

:type private_attributes: dict


**Reserved attributes:**

In addition to the reserved attributes inherited from the
:class:`~cf.Variable` class, the following attributes have special
definitions and it is recommended to set them accordingly.

==================== ============================================
Attribute            Description             
==================== ============================================
:attr:`cell_methods` The field's :class:`~cf.CellMethods` object.
:attr:`space`        The field's :class:`~cf.Field` object.
==================== ============================================

**Methods and attributes defined here:**
'''
                              
    _special_attributes = Variable._special_attributes.union(set(('ancillary_variables',
                                                                  'cell_methods',
                                                                  'flag_values',
                                                                  'flag_masks',
                                                                  'flag_meanings',
                                                                  )))
   
#
    def __str__(self):
        '''
x.__str__() <==> str(x)
'''
        string = []

        try:
            space = self.space
        except AttributeError:
            # Field has no space
            dim_names = self.shape
        else:     
            # Field has a space
            dim_names = []
            for dim in space.dimensions['data']:
                try:
                    dim_names.append(space[dim].name(ncvar=True, 
                                                     default=dim))
                except (KeyError, AttributeError):
                    try:                        
                        dim_names.append(space.nc_dimensions[dim])
                    except (KeyError, AttributeError):
                        dim_names.append(dim)
            #--- End: for
            dim_names = tuple(dim_names)
        # ---End: try

        dim_names = str(dim_names).replace("'", "")
        dim_names = dim_names.replace(',)', ')', 1)

        string = ['Data            : %s%s' %(self.name(long_name=True, 
                                                       ncvar=True, 
                                                       default=''),
                                             dim_names)]
        
        # Cell methods
        try:
            string.append(str(self.cell_methods))
        except AttributeError:
            # Field has no cell methods
            pass

        # Space
        try:
            string.append(str(space))
        except NameError:
            # Space has no space
            string.append('Dimensions      : \nAuxiliary coords: \nCell measures   : ')

        return '\n'.join(string)
    #--- End def

    def _concatenate_data(self):
        '''

Concatenate a field's data
'''
        original_n_elements = len(self._data)

        # Concatenate the data elements of the field
        super(Field, self)._concatenate_data()

        # Concatenate the data elements of the space components which
        # share a dimension with the field's data
        try:
            space = self.space
        except AttributeError:
            pass
        else:
            if original_n_elements > 1:
                for key in space:
                    space[key]._concatenate_data()
        #--- End: try

    #--- End: def

#    def _to_Values(self, other):
#        '''
#
#Like Variable._to_Values, but reorderin dimensions if required.
#'''
#        if not isinstance(other, self.__class__):
#            return super(Field, self)._to_Values(other)
#
#        # Still here? The other and self are both Fields
#        
#
#        
#        # Still here? 
#        raise TypeError, "Can't combine '%s' with '%s'" % \
#            (self.__class__.__name__, other.__class__.__name__)
#    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: add_offset
    # ----------------------------------------------------------------
    @property
    def add_offset(self):
        '''
The  CF 'add_offset' attribute. 
'''
        return self.getattr('add_offset')
    #--- End: def
    @add_offset.setter
    def add_offset(self, value): self.setattr('add_offset', value)
    @add_offset.deleter
    def add_offset(self):        self.delattr('add_offset')

    # ----------------------------------------------------------------
    # Property attribute: scale_factor
    # ----------------------------------------------------------------
    @property
    def scale_factor(self):
        '''
The  CF 'scale_factor' attribute. 
'''
        return self.getattr('scale_factor')
    #--- End: def
    @scale_factor.setter
    def scale_factor(self, value): self.setattr('scale_factor', value)
    @scale_factor.deleter
    def scale_factor(self):        self.delattr('scale_factor')

    # ----------------------------------------------------------------
    # Property attribute: ancillary_variables
    # ----------------------------------------------------------------
    @property
    def ancillary_variables(self):
        return self._get_special_attr('ancillary_variables')
    @ancillary_variables.setter
    def ancillary_variables(self, value):
        self._set_special_attr('ancillary_variables', value)
    @ancillary_variables.deleter
    def ancillary_variables(self):
        self._del_special_attr('ancillary_variables')


    # ----------------------------------------------------------------
    # Property attribute: Flags
    # ----------------------------------------------------------------
    @property
    def Flags(self):        
        return self._get_special_attr('Flags')
    @Flags.setter
    def Flags(self, value): 
        self._set_special_attr('Flags', value)
    @Flags.deleter
    def Flags(self):        
        self._del_special_attr('Flags')

    # ----------------------------------------------------------------
    # Property attribute: flag_values
    # ----------------------------------------------------------------
    @property
    def flag_values(self):      
        '''
'''
        try:
            return self.Flags.values
        except AttributeError:
            raise AttributeError, "Can't get '%s' attribute 'flag_values'" % \
                self.__class__.__name__
    #--- End: def
    @flag_values.setter
    def flag_values(self, value): 
        '''
'''
        try:
            self.Flags.values = value
        except AttributeError:
            self.Flags = Flags(values=value)
    #--- End: def
    @flag_values.deleter
    def flag_values(self): 
        '''
'''
        try:
            del self.Flags.values
        except AttributeError:
            raise AttributeError, "Can't delete '%s' attribute 'flag_values'" % \
                self.__class__.__name__
        else:
            if not self.Flags:
                del self.Flags
    #--- End: def 

    # ----------------------------------------------------------------
    # Property attribute: flag_masks
    # ----------------------------------------------------------------
    @property
    def flag_masks(self):      
        '''
'''
        try:
            return self.Flags.masks
        except AttributeError:
            raise AttributeError, "Can't get '%s' attribute 'flag_masks'" % \
                self.__class__.__name__
    #--- End: def
    @flag_masks.setter
    def flag_masks(self, value): 
        '''
'''
        try:
            self.Flags.masks = value
        except AttributeError:
            self.Flags = Flags(masks=value)
    #--- End: def
    @flag_masks.deleter
    def flag_masks(self): 
        '''
'''
        try:
            del self.Flags.masks
        except AttributeError:
            raise AttributeError, "Can't delete '%s' attribute 'flag_masks'" % \
                self.__class__.__name__  
        else:
            if not self.Flags:
                del self.Flags
    #--- End: def 

    # ----------------------------------------------------------------
    # Property attribute: flag_meanings
    # ----------------------------------------------------------------   
    @property
    def flag_meanings(self):      
        '''
'''
        try:
            return self.Flags.meanings
        except AttributeError:
            raise AttributeError, "Can't get '%s' attribute 'flag_meanings'" % \
                self.__class__.__name__
    #--- End: def
    @flag_meanings.setter
    def flag_meanings(self, value): 
        '''
'''
        try:
            self.Flags.meanings = value
        except AttributeError:
            self.Flags = Flags(meanings=value)
    #--- End: def
    @flag_meanings.deleter
    def flag_meanings(self): 
        '''
'''
        try:
            del self.Flags.meanings
        except AttributeError:
            raise AttributeError, "Can't delete '%s' attribute 'flag_meanings'" % \
                self.__class__.__name__  
        else:
            if not self.Flags:
                del self.Flags
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: cell_methods
    # ----------------------------------------------------------------
    @property
    def cell_methods(self): 
        '''

The field's CellMethods object.
        '''
        value = self._get_special_attr('cell_methods')
#        if isinstance(value, basestring):
#            cm = CellMethods()
#            value = cm.parse(value, field=self)
        return value
    #--- End: def

    @cell_methods.setter
    def cell_methods(self, value): 
        self._set_special_attr('cell_methods', value)

    @cell_methods.deleter
    def cell_methods(self):
        self._del_special_attr('cell_methods')

    # ----------------------------------------------------------------
    # Property attribute: space
    # ----------------------------------------------------------------
    @property
    def space(self):
        '''

The field's Space object
'''
        return self._get_special_attr('space')
    #--- End: def

    @space.setter
    def space(self, value): 
        if not isinstance(value, Space):
            raise TypeError, "asdasdasdsasdas dch "
        self._set_special_attr('space', value)

    @space.deleter
    def space(self):
        self._del_special_attr('space')

#    def reorder(self, other):
#
#
#        
#        # Reorder y so that its axes are in the same order as x
#        new_order = [var0_dimensions.index(match_for_dim1[dim1])
#                     for dim1 in var1_dimensions]
#        if new_order != range(var0.ndim):
#            _transpose = 1
#            y1 = numpy.transpose(y, axes=new_order)
        

    def reverse_axes(self, *names):
        '''

Reverse the direction of named axes *IN PLACE*. The reverse does not
actually occur until the data are accessed.
'''
        dimensions = self.space.dimensions.copy()
        data_dims  = dimensions.pop('data')

        for name in names:
            dim = self.coord(name, key=True)

            try:
                i = data_dims.index(dim)
            except ValueError:
                # Scalar
                pass
            else:
                try:
                    ddim = self._data.order[i]
                except AttributeError:
                    # No data
                    pass
                else:
                    # 1d or Nd
                    self._data.direction[ddim] = not self._data.direction[ddim]
                    try:
                        if self._data.adim is i:
                            self._data._list.reverse()
                            self._data._elements.reverse()
                    except AttributeError:
                        pass
            #--- End: try
        
            for key, dims in dimensions.iteritems():
                try:
                    i = dims.index(dim)
                except ValueError:
                   pass
                else:                
                    try:
                        ddim = self.space[key]._data.order[i]
                    except IndexError:
                        # Scalar
                        self.space[key]._data.direction = \
                            not self.space[key]._data.direction
                    except AttributeError:
                        # No data
                        pass
                    else:
                        # 1d or Nd
                        self.space[key]._data.direction[ddim] = \
                            not self.space[key]._data.direction[ddim]
                        try:
                            if self.space[key]._data.adim is i:
                                self.space[key]._data._list.reverse()
                                self.space[key]._data._elements.reverse()
                        except AttributeError:
                            pass
                #--- End: try
            #--- End: for
        #--- End: for
    #--- End: def

    def squash(self):
        '''

Squash axes *IN PLACE*. I.e. squeeze size 1 dimensions from the field
and its dimension coordinates if and only if a size 1 dimension is not
spanned by an auxuliary coordinate or a cell measures variable.
'''
        for dim in self.space:
            # --------------------------------------------------------
            # Only consider squeezing this dimension if it is a 1d,
            # size 1 dimension coordinate
            # --------------------------------------------------------
            if not dim.startswith('d'):
                continue

            c = self.space[dim]._data
            if not (c.size == 1 and c.ndim == 1):
                continue

            # --------------------------------------------------------
            # Do not squeeze this dimension if it is being used by an
            # auxiliary coordinate or cell measures
            # --------------------------------------------------------
            squeeze = True
            for key in self.space:
                if key.startswith('d'):
                    continue
                if dim in self.space.dimensions[key]:
                    squeeze = False
                    break
            #--- End: for
            if not squeeze:
                continue

            # --------------------------------------------------------
            # Convert the dimension coordinate to a scalar
            # --------------------------------------------------------
            c.shape                 = ()            
            c.order                 = []
            try:
                c.direction         = c.direction.values()[0] # c.direction  -> scalar bool
            except AttributeError:
                pass

            c.elements[0].order     = []
            c.elements[0].direction = []
            c.elements[0].shape     = []
            try:
                b = self.space[dim].bounds._data
            except AttributeError:
                pass
            else:
                b.shape                 = (2,)                
                b.order                 = ['bounds']
                b.direction             = {'bounds': b.direction['bounds']}
                b.elements[0].order     = ['bounds']
                b.elements[0].shape     = [2]
                b.elements[0].direction.pop(0)
            #--- End: try

            # --------------------------------------------------------
            # Drop this dimension from the field's data
            # --------------------------------------------------------
            try:
                index = self.space.dimensions['data'].index(dim)
            except ValueError:
                pass
            else:
                self.space.dimensions['data'].pop(index)
                self._data.remove_dimension(index)
        #--- End: for
    #--- End: def


#    def special_attrs(self):
#        return self._private['special_attrs'].copy()
#        out = super(Field, self).special_attrs()
#        for attr in ('Flags', 'cell_methods', 'ancillary_variables',
#                     'space'):
#            if hasattr(self, attr):
#                out[attr] = getattr(self, attr)
#        return out
    #--- End: def

    def dump(self, id=None, omit=()):
        '''

Return a string containing a full description of the field.

:keyword id: *Optional*. Set the common prefix of variable
             component names. If None then defaults to the class
             name.

:keyword id: str

:keyword omit: *Optional*. Omit the given attributes the
               description.

:type omit: sequence


:returns: A string containing the description of the field.

.. seealso:: :func:`cf.dump`
'''
        if id is None:
            id = self.__class__.__name__

        # Produce summary
        string = ['field summary\n-------------']
        string.append(str(self))
        string.append('')

        # Produce dump
        name = self.name(default='', long_name=True)
        string.append('%s field' % name)
        string.append(''.ljust(len(string[-1]), '-'))

        string.append(super(Field, self).dump(id=id, omit=('space',)))

        if hasattr(self, 'space'):
            string.append(self.space.dump(id='%(id)s.space' % locals()))

        return '\n'.join(string)
    #--- End: def

    def coord(self, standard_name=None, role=None, key=False, exact=False, 
              axis=None, ncvar=None, long_name=None, position=None, **kwargs):
        '''

Find a field's coordinate by name.

The given `name` argument is an abbreviation for (or equal to if the
`exact` parameter is True) the standard name of desired coordinate If
the `key` parameter is True then return the coordinate's space key
name instead of the Coordinate variable. If a coordinate does not have
`standard_name` attribute, then `name` will be matched against its
`ncvar` attribute is used.

Note that the returned coordinate is an object identity to the
coordinate stored in the space so, for example, a coordinate's
attributes may be changed in-place as follows:

>>> s.coord('height').long_name
AttributeError: 'Coordinate' object has no attribute 'long_name'
>>> s.coord('hei').long_name = 'HEIGHT'
>>> s.coord('heigh').long_name
'HEIGHT'

Or a deep copy may be made with the coordinate's
:meth:`~cf.Coordinate.copy` method:

>>> h = s.coord('height').copy()

:argument name: The string to identify a coordinate by name.

:type name: str

:keyword exact: *Optional*. If True then assume that the value of the
                `name` argument is equal to exactly one coordinate's
                name.

:type exact: bool

:keyword key: *Optional*. Return the space key name instead of the
              coordinate.

:type key: str

:keyword role: *Optional*. Restrict the search to coordinates of the
               given role. Valid values are 'dim' and 'aux', for
               dimension and auxiliary coordinate types
               respectively. If None then both types of coordinates
               will be searched.

:type role: str or None

  
:returns: If a coordinate has been identified, return either a
          :class:`~cf.Coordinate` instance or, if the `keys` parameter
          is True, a space key name string. otherwise, return None.


**Examples:**

>>> s.coord('lon')
<CF Coordinate: longitude(128)>
>>> s.coord('lon', key=True)
'dim2'
>>> s.coord('lonX', key=True)
None
>>> s.coord('lon', exact=True)
None
>>> s.coord('longitude', exact=True)
<CF Coordinate: longitude(128)>
'''
        try:
            return self.space.coord(standard_name=standard_name, axis=axis, role=role,
                                    ncvar=ncvar, position=position,
                                    key=key, exact=exact, **kwargs)
        except AttributeError:
            return None
    #--- End: def

    @property
    def slice(self):
        '''

Slice the field's data, returning a new field. The field's space is
also sliced.

Slicing by index
----------------

Refer to the 'Copying and resizing' in the :mod:`cf` package.

Slicing by coordinate criteria
------------------------------

Coordinates are identified by keyword parameters of their standard
names. If the optional first positional argument `exact` evaluates to
`False` (the default) then unambiguous abbreviations of the standard
names are accepted, otherwise exact standard names are required:

>>> s.slice(longitude=0.0)
<CF Field: air_temperature(12, 64, 1)
>>> s.slice(lon=0.0)
<CF Field: air_temperature(12, 64, 1)
>>> s.slice(l=0.0)
ValueError: Can't identify coordinate from keyword 'l'
>>> s.slice(False, lon=0.0)
<CF Field: air_temperature(12, 64, 1)
>>> s.slice(True, lon=0.0)
ValueError: Can't identify coordinate from keyword 'lon'

The coordinate criteria are given by the keyword parameters
values. Four different methods are available:

1. A scalar. The dimension is sliced where coordinate equals (using
   the coordinate's == operator) the scalar value:

   >>> s.slice(lat=0.0)

2. A sequence. The dimension is sliced where the coordinate equals
   (using the coordinate's == operator) the sequence's elements. All
   elements of the sequence must be found in the coordinate, but the
   sequence may contain fewer elements than the coordinate's data:

   >>> s.slice(lon=[0.0, 2.8125])

3. A :class:`~cf.Comparison` object (as, for example, returned by the
   :func:`~cf.lt` function). The dimension is sliced where the
   comparison for the coordinate evaluates to True:

   >>> s.slice(lat=cf.lt(0.0))

4. An numpy array of booleans with the same shape as the
   coordinate. The dimension is sliced where elements of this array
   are True:

   >>> s.slice(lat=s.coord('lat')<0.0)
  
If a coordinate is not identified with a keyword parameter then its
dimension is returned unchanged.

The order in which coordinates are given in the parameter list is
irrelevant.

A call with no keyword parameters is equivalent to indexing with an
Ellipsis (...).


**Examples:**

>>> s.slice(lat=0.0, lon=0.0)
<CF Field: air_temperature(12, 1, 1)
>>> s.slice(lon=cf.lt(0.0), lon=[0, 2.8125])
<CF Field: air_temperature(12, 32, 2)
>>> s.slice(lon=[0, 1])
ValeuError: 'longitude' coordinate doesn't contain the value 1
>>> s.slice(lon=1)
ValueError: No indices found for 'longitude' coordinate
'''
        return SliceField(self)
    #--- End: def

#--- End: class


class CellMethods(CfList):
    '''

A CF cell methods object to describe the characteristic of a field
that is represented by cell values.

Each cell method is stored in a dictionary and these dictionaries are
stored in a list-like object. Similarly to a CF 'cell_methods' string,
the order of cell methods in the list is important.

The dictionary representing each cell method recognizes the following
keys (where all keys are optional; the referenced sections are from
`NetCDF Climate and Forecast (CF) Metadata Conventions Version 1.5
<http://cf-pcmdi.llnl.gov/documents/cf-conventions/1.5/cf-conventions.html>`_;
and words in double quotes ("...") refer to CF cell_methods components
in the referenced sections):

========= ========================================================
Key       Value
========= ========================================================
method    A "method" component. Refer to section 7.3.

name      A list of all of the "name" components involved with the
          "method" component. Each element is either:

              1. The standard_name of the cell method's dimension
              2. The string 'area'          
              3. A netCDF variable name or netCDF dimension name

          Option 3. only occurs if the cell method's dimension
          either has no dimension coordinate or has a dimension
          coordinate with no standard_name.

          This is a deviation from the "name" components as
          described in sections 7.3, 7.3.1 and 7.3.4, in that
          standard_names are stored whenever possible, rather than
          only in the special cases described in section
          7.3.4. Inspection of the 'no_coords' list retains the
          information required to fully interpret the cell method.

dim       A list of space dimension names corresponding to the
          'name' list. A 'name' list value of 'area' always
          corresponds to a 'dim' list value of None.

no_coords A list of booleans, corresponding to the 'name' list,
          indicating whether or not a particular cell method is
          relevant to the data in a way which may not be precisely
          defined by the corresponding dimension or dimensions. Refer
          to section 7.3.4.

interval  A list of "interval" component values corresponding to
          the 'name' list. Refer to section 7.3.2.

units     A list of "interval" component units corresponding to
          the 'interval' list. Refer to section 7.3.2.

comment   A "comment" or non-standardised information,
          component. Refer to sections 7.3.2 and 7.4

within    A "within" climatology component. Refer to section 7.4.

over      An "over" cell portion or "over" climatology component.
          Refer to sections 7.3.3 and 7.4.

where     A "where" component. Refer to section 7.3.3.
========= ========================================================

Note that the above table assumes that the cell methods have been
constructed in the context of a field, the only way in which the 'dim'
and 'no_coords' keys may take sensible values. If this is not the
case, as would happen if the `parse` method were called without its
`field` keyword, then elements of the 'dim' key default to False and
elements of the 'no_coords' key default to None.

:argument sequence: *Optional*. Initialize new list from
                    sequence's items.

:type sequence: sequence

**Examples:**

>>> c = cf.CellMethods()
>>> c = c.parse('time: minimum within years time: mean over years (ENSO years)')    
>>> print c
Cell methods    : time: minimum within years
                  time: mean over years (ENSO years)
>>> list(c)
[
 {'method'   : 'minimum',
  'name'     : ['time'],     
  'dim'      : [False],
  'no_coords': [None], 
  'within'   : 'years'
 },
 {'method'   : 'mean'
  'name'     : ['time'],
  'dim'      : [False], 
  'no_coords': [None],     
  'comment'  : 'ENSO years', 
  'over'     : 'years'
 }
]

>>> c = c.parse('lat: lon: standard_deviation')
>>> list(c)
[
 {'dim'      : [False, False],
  'method'   : 'standard_deviation'
  'name'     : ['lat', 'lon'],
  'no_coords': [None, None],
 }
]
 

**Overloaded operators**

Refer to :class:`~cf.CfList`.


**Methods:**

================================== =====================================================
Method                             Description
================================== =====================================================
:meth:`~cf.CellMethods.append`     As for a built-in list.
:meth:`~cf.CellMethods.copy`       Create a deep copy.
:meth:`~cf.CellMethods.count`      As for a built-in list but using numerically tolerant
                                   equality.
:meth:`~cf.CellMethods.dump`       Return a string containing a full description of the
                                   object.
:meth:`~cf.CellMethods.equals`     Determine whether two lists are congruent
                                   element-wise.
:meth:`~cf.CellMethods.extend`     As for a built-in list.
:meth:`~cf.CellMethods.index`      As for a built-in list but using numerically tolerant
                                   equality.
:meth:`~cf.CellMethods.insert`     As for a built-in list.
:meth:`~cf.CellMethods.parse`      Parse a netCDF(CF) cell_methods string.
:meth:`~cf.CellMethods.pop`        As for a built-in list.
:meth:`~cf.CellMethods.properties` Return a set of writable, public attributes.
:meth:`~cf.CellMethods.remove`     As for a built-in list.
:meth:`~cf.CellMethods.reverse`    As for a built-in list.
:meth:`~cf.CellMethods.sort`       As for a built-in list.
:meth:`~cf.CellMethods.strings`    Convert each element to a string.
================================== =====================================================


**Methods and attributes defined here:**
'''

    def __init__(self, string=None):
        '''
'''
        super(CellMethods, self).__init__()
        self.parse(string=string)
    #--- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)
'''
        return '<CF %s: %s>' % (self.__class__.__name__, 
                                ' '.join(self.strings()))
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)
'''        
        strings = self.strings()

        try:
            string = ['Cell methods    : %s' % strings[0]]
        except IndexError:
            return  'Cell methods    : '

        for s in strings[1:]:
            string.append('                  %s' % s)
            
        return '\n'.join(string)
    #--- End: def

    def dump(self, id=None, omit=()):
         '''

Return a string containing a full description of the cell methods.

If a cell methods 'name' is followed by a '*' then that cell method is
relevant to the data in a way which may not be precisely defined its
corresponding dimension or dimensions.

:keyword id: *Optional*. Set the common prefix of each line of the
             output. If None then defaults to the class name.

:type id: str

:keyword omit: *Optional*. Dummy argument required for consistency
               with the :func:`cf.dump` function.

:type omit: sequence


:returns: A string containing the description of the cell methods.

.. seealso:: :func:`cf.dump`
'''
         if id is None:
             id = self.__class__.__name__
                              
         string = [] #'cell methods\n------------']
         
         for i, s in enumerate(self.strings()):
             string.append('%s[%d] -> %s' % (id, i, s))
             
         return '\n'.join(string)
    #--- End: def

    def parse(self, string=None):
        '''

Parse a CF cell_methods string into this CellMethods instance.

:argument string: The CF cell_methods string to be parsed into the
                  CellMethods object.

:type string: str

:keyword field: *Optional*. Set the 'name', 'dim' and 'no_coord' key
                values appropriately for the given field. Refer to
                :class:`~cf.CellMethods`.

:type field: Field

:returns: A cell methods list.

**Examples:**

>>> c = cf.CellMethods()
>>> c = c.parse('time: minimum within years time: mean over years (ENSO years)')    
>>> print c
Cell methods    : time: minimum within years
                  time: mean over years (ENSO years)

'''

#        result = type(self)()

        # Split the cell_methods string into a list of strings ready
        # for parsing into the result list. E.g.
        #   'lat: mean (interval: 1 hour)'
        # maps to 
        #   ['lat:', 'mean', '(', 'interval:', '1', 'hour', ')']
        cell_methods = re.sub('\((?=[^\s])' , '( ', string)
        cell_methods = re.sub('(?<=[^\s])\)', ' )', cell_methods).split()

        while cell_methods:

            # Create a new element for this 
            self.append(CfDict())            

            # List of names
            self[-1]['name']      = CfList()
            self[-1]['dim']       = CfList()
            self[-1]['no_coords'] = CfList()

            while cell_methods:
                if not cell_methods[0].endswith(':'):
                    break
                self[-1]['name'].append(cell_methods.pop(0)[:-1])            
                self[-1]['dim'].append(False)
                self[-1]['no_coords'].append(None)
            #--- End: while

            if not cell_methods:
                break

            # Method
            self[-1]['method'] = cell_methods.pop(0)

            if not cell_methods:
                break

            # Climatological statistics and statistics which apply to
            # portions of cells
            while cell_methods[0] in ('within', 'where', 'over'):
                term = cell_methods.pop(0)
#                if term == 'over' and cell_methods[0] in ('years', 'days'):
#                    self[-1]['c'+term] = cell_methods.pop(0)
#                else:
                self[-1][term] = cell_methods.pop(0)
                if not cell_methods:
                    break
            #--- End: while

            if not cell_methods: 
                break

            # interval and comment
            if cell_methods[0].endswith('('):
                cell_methods.pop(0)

                if not (re.search('^(interval|comment):$', cell_methods[0])):
                    cell_methods.insert(0, 'comment:')
                           
                while not re.search('^\)$', cell_methods[0]):
                    term = cell_methods.pop(0)[:-1]

                    if term == 'interval':
                        if not self[-1].has_key('interval'):
                            self[-1]['interval'] = CfList()
                        interval = float(cell_methods.pop(0))
                        interval = Value(interval)
                        if cell_methods[0] != ')':
                            interval.units = cell_methods.pop(0)
                        self[-1]['interval'].append(interval)
                        continue
                    #--- End: if

                    if term == 'comment':
                        comment = CfList()
                        while cell_methods:
                            if cell_methods[0].endswith(')'):
                                break
                            if cell_methods[0].endswith(':'):
                                break
                            comment.append(cell_methods.pop(0))
                        #--- End: while
                        self[-1]['comment'] = ' '.join(comment)
                    #--- End: if

                #--- End: while 

                if cell_methods[0].endswith(')'):
                    cell_methods.pop(0)
            #--- End: if
        #--- End: while

        return self
    #--- End: def

    def strings(self):
        '''

Retun a list of a netCDF(CF)-like string representation of each cell
method.

Note that if the intention is to concatenate the output list into a
string for creating a netCDF(CF) cell_methods attribute, then the cell
methods "name" components may need to be modified, where appropriate,
to reflect netCDF variable names.

:returns: A list of cell method strings.

**Examples:**

>>> c = cf.CellMethods()
>>> c = c.parse('time: minimum within years time: mean over years (ENSO years)')    
>>> print c
Cell methods    : time: minimum within years
                  time: mean over years (ENSO years)
>>> c.strings()
['time: minimum within years',
 'time: mean over years (ENSO years)']
'''
        strings = []
        
        for cm in self:

            string = []

            for j in xrange(len(cm['name'])):
                string.append('%s: ' % cm['name'][j])

            string.append(cm['method'])

            x = []
            try:
                x.extend(('within', cm['within'])) 
            except KeyError:
                try:
                    x.extend(('where', cm['where'])) 
                except KeyError:
                    pass
            try:
                x.extend(('over', cm['over'])) 
            except KeyError:
                pass
            if x:
                string.append(' %s' % ' '.join(x))

            if cm.has_key('interval'):
                string.append(' (')
                for j in xrange(len(cm['interval'])):
                    string.append('interval: %s' % str(cm['interval'][j]))
#                    string.append('interval: %s' % repr(cm['interval'][j]))
#                    try:
#                        string.append(' %s' % cm['units'][j].units)
#                    except KeyError:
#                        pass
                if cm.has_key('comment'):
                    string.append(' comment: %s' % cm['comment'])
                string.append(')')

            elif cm.has_key('comment'):
                string.append(' (%s)' % cm['comment'])
            
            strings.append(''.join(string))
        #--- End: for

        return strings
    #--- End: def

#--- End: class


# ====================================================================
#
# List of Fields object
#
# ====================================================================

class FieldList(VariableList):
    '''

An ordered sequence of fields stored in a single, list-like
object. Refer to Variable Lists in the :mod:`cf` package for details.

**Methods and attributes defined here:**
'''
    @property
    def slice(self):
        '''

Slice each field in the list, returning a new list of fields. Slicing
by indices or by coordinate values are both allowed.

**Examples:**

>>> sl
[<CF Variable: air_temperature(73, 96)>,
 <CF Variable: air_temperature(73, 96)>]
>>> sl.slice[0,0]
[<CF Variable: air_temperature(1,1)>,
 <CF Variable: air_temperature(1,1)>]
>>> sl.slice(longitude=0, latitude=0)
[<CF Variable: air_temperature(1,1)>,
 <CF Variable: air_temperature(1,1)>]
'''
        return SliceFieldList(self)
    #--- End: def

    def coord(self, *args, **kwargs):
        '''

Return the requested coordinate from each field in the list.
'''    
        return CoordinateList([v.coord(*args, **kwargs) for v in self._list])
    #--- End: def
    
    def squash(self):
        '''
 
Return a built-in list of the result of the match on each element of
the list of variables, with unambiguous abbreviations of phenomenon
names allowed.
'''  
        for f in self._list:
            f.squash()
    #--- End: def
       
#--- End: class


# ====================================================================
#
# Field slice object
#
# ====================================================================

class SliceField(SliceVariable):
    '''

Return a new Field instance whose data and grid are sliced according
to indices or call arguments to its `slice` attribute. The latter
converts criteria on coordinate values into indices and so the two
methods a formally equivalent.

**Examples:**

>>> print s
Data            : air_temperature(time, latitude, longitude)
Cell methods    : time: mean (interval: 1.0 month)
Dimensions      : time(7070) -> 450-11-16 00:00:00 to 1049-12-16 12:00:00
                : latitude(64) -> -87.8638 to 87.8638 degrees_north
                : longitude(128) -> 0 to 357.1875 degrees_east
                : height(1) -> 2 m
Auxiliary coords:
>>> s.slice[:,32, 64]
<CF Field: air_temperature(12, 1, 1)
>>> s.slice(longitude=0.0, latitude=0.0)
<CF Field: air_temperature(12, 1, 1)

Slicing by index
----------------
Refer to the 'Copying and resizing' of the `cf` help.

Slicing by coordinate criteria
------------------------------

Coordinates are identified by keyword parameters of their standard
names. If the optional positional argument `exact` evaluates to `False`
(the default) then unambiguous abbreviations of the standard names are
accepted, otherwise exact standard names are required:

>>> s.slice(longitude=0.0)
<CF Field: air_temperature(12, 64, 1)
>>> s.slice(lon=0.0)
<CF Field: air_temperature(12, 64, 1)
>>> s.slice(l=0.0)
ValueError: Can't identify coordinate from keyword 'l'
>>> s.slice(False, lon=0.0)
<CF Field: air_temperature(12, 64, 1)
>>> s.slice(True, lon=0.0)
ValueError: Can't identify coordinate from keyword 'lon'

The coordinate criteria are given by the keyword parameters
values. Four different methods are available:

1. A scalar. The dimension is sliced where coordinate equals (using
   the coordinate's == operator) the scalar value.

   **Example**
   >>> s.slice(lat=0.0)

2. A sequence. The dimension is sliced where the coordinate equals
   (using the coordinate's == operator) the sequence's elements. All
   elements of the sequence must be found in the coordinate, but the
   sequence may contain fewer elements than the coordinate's data.

   **Example**
   >>> s.slice(lon=[0.0, 2.8125])

3. A :class:`~cf.Comparison` object (as, for example, returned by the
   :func:`~cf.lt` function). The dimension is sliced where the
   comparison for the coordinate evaluates to True.

   **Example**
   >>> s.slice(lat=cf.lt(0.0))

4. An numpy array of booleans with the same shape as the
   coordinate. The dimension is sliced where elements of this array
   are True.

   **Example**
   >>> s.slice(lat=s.coord('lat')<0.0)
  
If a coordinate is not identified with a keyword parameter then its
dimension is returned unchanged.

The order in which coordinates are given in the parameter list is
irrelevant.

A call with no keyword parameters is equivalent to indexing with an
Ellipsis (...).

**Examples:**

>>> f.slice(lat=0.0, lon=0.0)
<CF Field: air_temperature(12, 1, 1)
>>> f.slice(lon=cf.lt(0.0), lon=[0, 2.8125])
<CF Field: air_temperature(12, 32, 2)
>>> f.slice(lon=[0, 1])
ValeuError: 'longitude' coordinate doesn't contain the value 1
>>> f.slice(lon=1)
ValueError: No indices found for 'longitude' coordinate
'''
    __slots__ = []

    def __getitem__(self, index):
        '''
x.__getitem__(index) <==> x[index]
''' 
        field = self.variable

        # Copy the field

        # It'd be nice not to deep copy the space coordinates and cell
        # measures now but do this in the for loop below. Would be a
        # bit more efficient.

        # Copy the coordinate
        new = field.copy(data=False)

        # Parse the index
        index, dummy = _parse_index(field, index)

        # Slice the field's data
        new._data = field._data.slice[index]

        # Break ancillary variables
        full_slice = (slice(None),) * field.ndim
        if index != full_slice:
            try:
                del new.ancillary_variables
            except AttributeError:
                pass        

        try:
            new.space
        except AttributeError:
            # Return now if the field doesn't have a space
            return new
        
        # -----------------------------------------------------------
        # Still here? Then slice the space
        # -----------------------------------------------------------
        i = -1
        for dim in new.space.dimensions['data'][:]:

            i += 1

            # Continue if no slicing to be done on this dimension
            if index[i] == slice(None, None, None): # dch ??
                continue

            for key in new.space:

                # Continue if this space component does not span the
                # current dimension
                if dim not in new.space.dimensions[key]:
                    continue

                dice          = [slice(None)] * new.space[key].ndim
                pos           = new.space.dimensions[key].index(dim)

                dice[pos]     = index[i]
                new.space[key] = new.space[key].slice[tuple(dice)]

                # Update dimension size
                new.space.dimension_sizes[dim] = new.space[key].shape[pos]
            #--- End: for
        #--- End: for
                                
        try:
            transforms = new.space.transform.values()
        except AttributeError:
            pass
        else:
            if index != full_slice:
                # Break formula terms
                for transform in transforms:
                    try:
                        transform['grid_mapping_name']
                        continue
                    except KeyError:
                        pass
                    for term, value in transform.iteritems():
                        if isinstance(value, Variable): 
                            transform[term] = None
            #--- End: if
        #--- End: try

        return new
    #--- End: def

    def __call__(self, *args, **kwargs):
        '''

Slice by coordinate value. refer to :meth:`cf.Field.slice` for
details.
'''

        field = self.variable       

        if not kwargs:
            return copy.deepcopy(field)

        if args:
            exact = args[0]
        else:
            exact = False

        dimensions = []

        # Initialize index
        index  = [slice(None)] * field.ndim

        # Loop round slice criteria
        for name, value in kwargs.iteritems():
                                    
            key = field.coord(name, key=True, exact=exact)

            if key is None:
                raise ValueError, "Can't identify coordinate from keyword '%(name)s'" % locals()

            if key.startswith('dim'):
                # Dimension coordinate
                dim = key
            else:
                # Auxiliary coordinate
                dim = field.space.dimensions[key][0]

                if field.space[key].ndim > 1:
                    raise ValueError, "Can't define a slice for a multidimensional auxiliary coordinate '%s'" % field.coord(name, exact=exact)
            #--- End: if

            if dim in dimensions:
                raise ValueError, "Two coordinates share the same dimension"
            dimensions.append(dim)

            if dim not in field.space.dimensions['data']:
                raise IndexError, "Data doesn't have a '%s' dimension" % \
                    field.coord(name, exact=exact)

            # Define c, the coordinate for the specified dimension
            c = field.space[key]
            slice_ = []
            try:
                # value is a Comparison object
                slice_ = value.evaluate(c)
            except AttributeError:

                try:
                    # value is a numpy array
                    if value.dtype.kind == 'b':
                        # value is a boolean numpy array
                        if value.shape != c.shape:
                            raise ValueError, "Boolean numpy array must have shape as coordinate"
                        slice_ = value
                    else:
                        # value is a non-boolean numpy array
                        try:
                            # value is a 1d numpy array
                            for v in value:
                                x = numpy.where(c == v)[0]
                                if not x.size:
                                    raise ValueError, "'%s' coordinate doesn't contain the value %s" % \
                                        (field.coord(name, exact=exact), v)
                                slice_ = x
                        except TypeError:
                            # value is a scalar numpy array
                            slice_ = numpy.where(c == v)[0]
                except AttributeError:
                    # value is not a numpy array
                    try:
                        # value is a sequence
                        value.__iter__
                        for v in value:
                            x = numpy.where(c == v)[0]
                            if not x.size:
                                raise ValueError, "'%s' coordinate doesn't contain the value %s" % \
                                    (field.coord(name, exact=exact), v)
                            slice_ = x
                    except AttributeError:
                        # value is a scalar
                        slice_ = numpy.where(c == value)[0]
                    #--- End: try
                #--- End: try
            #--- End: try
                
            new_dim_size = len(slice_)
            if not new_dim_size:
                raise ValueError, "No indices found for '%s' coordinate" % \
                    field.coord(name, exact=exact)
            #--- End: if

            # Add slice_ to the right place in the data slice
            pos        = field.space.dimensions['data'].index(dim)
            index[pos] = slice_        
        #--- End: for

        return field.slice[tuple(index)]
    #--- End: def

#--- End: class


# ====================================================================
#
# List of Fields slice object
#
# ====================================================================

class SliceFieldList(SliceVariableList):
    '''
'''
    __slots__ = []

    def __call__(self, *arg, **kwargs):
        '''

Slice by fields by coordinate value. refer to :meth:`cf.Field.slice`
for details.
'''
        variablelist = self.variablelist
        return type(variablelist)([f.slice(*arg, **kwargs) 
                                   for f in variablelist])
    #--- End: def

#--- End: class

class Flags(object):
    '''
    '''
    __slots__ = ['_values',
                 '_meanings',
                 '_masks',
                 ]

    def __init__(self, **kwargs):
        '''
'''
        for attr, value in kwargs.iteritems():
            if value is not None:
                setattr(self, attr, value)
        #--- End: if

        # Sort the attributes
        self.sort()
    #--- End: def

    def __eq__(self, other):
        '''
x.__eq__(y) <==> x==y
'''
        return self.equals(other)
    #--- End: def

    def __ne__(self, other):
        '''
x.__ne__(y) <==> x!=y
'''
        return not self.equals(other)
    #--- End: def

    def __nonzero__(self):
        '''
x.__nonzero__() <==> x !=0
'''
        for attr in self.__slots__:
            if hasattr(self, attr):
                return True
        return False        
    #-- End: def

    @property
    def values(self):
        try:
            return self._values
        except AttributeError:
            raise AttributeError, "'%s' object has no attribute 'values'" % \
                self.__class__.__name__

    @values.setter
    def values(self, value):
        if not isinstance(value, numpy.ndarray):
            value = numpy.array(value)
        self._values = vale
    @values.deleter
    def values(self):
        try:
            del self._values
        except AttributeErorr:
            raise AttributeError, "Can't delete '%s' attribute 'values'" % \
                self.__class__.__name__

    @property
    def masks(self):
        try:
            return self._masks
        except AttributeError:
            raise AttributeError, "'%s' object has no attribute 'masks'" % \
                self.__class__.__name__

    @masks.setter
    def masks(self, value):
        if not isinstance(value, numpy.ndarray):
            value = numpy.array(value)
        self._masks = value
    @masks.deleter
    def masks(self):
        try:
            del self._masks
        except AttributeErorr:
            raise AttributeError, "Can't delete '%s' attribute 'masks'" % \
                self.__class__.__name__

    @property
    def meanings(self):
        try:
            return self._meanings
        except AttributeError:
            raise AttributeError, "'%s' object has no attribute 'meanings'" % \
                self.__class__.__name__

    @meanings.setter
    def meanings(self, value):
        if isinstance(value, basestring):
            value =  numpy.array(value.split())
        self._meanings = value
    @meanings.deleter
    def meanings(self):
        try:
            del self._meanings
        except AttributeErorr:
            raise AttributeError, "Can't delete '%s' attribute 'meanings'" % \
                self.__class__.__name__

    def equals(self, other):
        '''
'''        
        for attr in self.__slots__:
            try:
                x = getattr(self , attr)
            except AttributeError:
                if hasattr(other, attr):
                    return False
            else:
                try:
                    y = getattr(other, attr)
                except AttributeError:
                    return False
                try:
                    if not (x == y).all():
                        return False
                except AttributeError:
                    # This will be the case if the two arrays have different lengths
                    return False
        #--- End: for

        return True
    #--- End: def
                
    def __repr__(self):
        '''
x.__repr__() <==> repr(x)
'''  

        string = ['<CF %s: ' % self.__class__.__name__]
        if hasattr(self, 'masks'):
            string.append(str(self.masks))
        if hasattr(self, 'values'):
            string.append(str(self.values))
        if hasattr(self, 'meanings'):            
            string.append(str(self.meanings))
        string.append('>')
        return ''.join(string)
    #--- End: def

    def sort(self):
        '''
Sort the flags by flag value *in place*
'''
        # Sort all three attributes
        if hasattr(self, 'values'):
            indices = numpy.argsort(self.values)
            for attr in self.__slots__:
                if hasattr(self, attr):
                    array      = getattr(self, attr).view()
                    array[...] = array[indices]
        #--- End: if
    #--- End: def

    def copy(self):
        '''
'''
        return copy.deepcopy(self)
    #--- End: def
#--- End: class
