import netCDF4
import numpy
import re
import copy
from .coordinate   import Coordinate, CoordinateBounds
from .cellmeasures import CellMeasures
from .space        import Space, Transform
from .field        import Field, FieldList, CellMethods, Flags
from .data         import Data, set_Data_attributes, dimension_direction
from .utils        import CfDict
from .units        import Units


def read(filename, close=False):
    ''' 

Read fields from an input netCDF file on disk or from an OPeNDAP
server location.

All data arrays from all field components are not read from disk
but are stored in the returned fields as file pointers.

NetCDF dimension names are stored in the :attr:`nc_dimensions`
attribute of a field's space and netCDF variable names are stored
in the :attr:`ncvar` attributes of the field and its space
components (coordinates, coordinate bounds, cell measures and
transformations).


:argument filename: A string giving the file name or OPenDAP URL
                    from which to read fields.

:type filename: `str`


:returns: A list of fields (a `cf.FieldList` object).

**Examples**

>>> import cf
>>> s = cf.netcdf.read('file.nc')
>>> type(s)
<class 'cf.field.FieldList'>
>>> s
[<CF Field: pmsl(30, 24)>,
 <CF Field: z-squared(17, 30, 24)>,
 <CF Field: temperature(17, 30, 24)>,
 <CF Field: temperature_wind(17, 29, 24)>]

>>> cf.netcdf.read('file.nc')[0:2]
[<CF Field: pmsl(30, 24)>,
 <CF Field: z-squared(17, 30, 24)>]

>>> cf.netcdf.read('file.nc', units='K')
[<CF Field: temperature(17, 30, 24)>,
 <CF Field: temperature_wind(17, 29, 24)>]

>>> cf.netcdf.read('file.nc').extract(units='K')
[<CF Field: temperature(17, 30, 24)>,
 <CF Field: temperature_wind(17, 29, 24)>]

>>> cf.netcdf.read('file.nc')[0]
<CF Field: pmsl(30, 24)>

.. seealso:: `~cf.read` for more examples of field selection by
             phenomena criteria.

'''
    
    # Read the netCDF file
    nc = netCDF4.Dataset(filename, 'r')

    # ----------------------------------------------------------------
    # Put the file's global attributes into the global
    # 'global_attributes' dictionary
    # ----------------------------------------------------------------
    global_attributes = {}
    for attr in map(str, nc.ncattrs()):
        try:
            value = nc.getncattr(attr)
            if isinstance(value, basestring):
                try:
                    global_attributes[attr] = str(value)
                except UnicodeEncodeError:
                    global_attributes[attr] = value           
            else:
                global_attributes[attr] = value     
        except UnicodeDecodeError:
            pass
    #--- End: for
    
    # ----------------------------------------------------------------
    # Create a list of the names of all of the variables in the
    # file. E.g. variables=['lon','lat','tas']
    # ----------------------------------------------------------------
    variables = map(str, nc.variables)

    # ----------------------------------------------------------------
    # Create a dictionary keyed by nc variable names where each key's
    # value is a dictionary of that variable's nc
    # attributes. E.g. var_attr['tas']['units']='K'
    # ----------------------------------------------------------------
    var_attr = {}
    for ncvar in variables:
        var_attr[ncvar] = {}
        for attr in map(str, nc.variables[ncvar].ncattrs()):
            try:
                var_attr[ncvar][attr] = nc.variables[ncvar].getncattr(attr)
                if isinstance(var_attr[ncvar][attr], basestring):
                    try:
                        var_attr[ncvar][attr] = str(var_attr[ncvar][attr])
                    except UnicodeEncodeError:
                        var_attr[ncvar][attr] = var_attr[ncvar][attr]                   
            except UnicodeDecodeError:
                pass
        #--- End: for

        # Check for bad units
        try:
            Units(units=var_attr[ncvar]['units'])
        except TypeError:
            print "WARNING: Unsupported units: '%s'" % var_attr[ncvar]['units']
            var_attr[ncvar]['nonCF_units'] = var_attr[ncvar].pop('units')
        except KeyError:
            pass

        # Check for bad calendar
        try:
            Units(calendar=var_attr[ncvar]['calendar'])
        except TypeError:
            print "WARNING: Unsupported calendar: '%s'" % var_attr[ncvar]['calendar']
            var_attr[ncvar]['nonCF_calendar'] = var_attr[ncvar].pop('calendar')
        except KeyError:
            pass
    #--- End: for

    # ----------------------------------------------------------------
    # Remove everything bar data variables from the list of
    # variables. I.e. remove dimension and auxiliary coordinates,
    # their bounds and grid_mapping variables
    # ----------------------------------------------------------------
    for ncvar in variables[:]:

        # Remove dimension coordinates and their bounds
        if ncvar in map(str, nc.dimensions):

            try:
                variables.remove(ncvar)
            except ValueError:
                # This ncvar has already been removed. For example, it
                # could have been an auxiliary coordinate to another
                # variable
                pass
            else:
                for attr in ('bounds', 'climatology'):
                    try:
                        bounds = var_attr[ncvar][attr]
                    except KeyError:
                        continue
                
                    # Check the dimensionality of the coordinate's
                    # bounds. If it is not right, then it can't be a
                    # bounds variable and so promote to an independent
                    # data variable
                    try:
                        b = nc.variables[bounds]
                    except KeyError:                    
                        print "Missing bounds variable '%(bounds)s' in %(filename)s" % \
                            locals()
                        del var_attr[ncvar][attr]
                    else:
                        if b.ndim == nc.variables[ncvar].ndim+1:
                            variables.remove(bounds)
                        else:
                            del var_attr[ncvar][attr]
                        break
                #--- End: for
            #--- End: try
            continue
        #--- End: if
    
        # Remove auxiliary coordinates and their bounds
        try:
            var_attr[ncvar]['coordinates']
        except KeyError:
            pass
        else:
            # Allow for (incorrect) comma separated lists
            for aux in re.split('\s+|\s*,\s*', var_attr[ncvar]['coordinates']):
                try:
                    variables.remove(aux)
                except ValueError:
                    pass
                else:
                    for attr in ('bounds', 'climatology'):
                        try:
                            bounds = var_attr[aux][attr]
                        except KeyError:
                            continue

                        # Check the dimensionality of the coordinate's
                        # bounds. If it is not right, then it can't be
                        # a bounds variable and so promote to an
                        # independent data variable
                        try:
                            b = nc.variables[bounds]
                        except KeyError:                    
                            print "Missing bounds variable '%(bounds)s' in %(filename)s" % \
                                locals()
                            del var_attr[aux][attr]
                        else:
                            if b.ndim == nc.variables[aux].ndim+1:
                                variables.remove(bounds)
                            else:
                                del var_attr[aux][attr]
                            break
                    #--- End: for
            #--- End: for
        #--- End: try
    
        # Remove grid mapping variables
        try:
            variables.remove(var_attr[ncvar]['grid_mapping'])
        except (KeyError, ValueError):
            pass
        #--- End: try
    #--- End: for

    # ----------------------------------------------------------------
    # Everything left in the variables list is now a proper data
    # variable, so make a list of fields, each of which contains one
    # data variable and the relevant shared metadata. Reset
    # 'seen_in_file' to an empty dictionary.
    # ----------------------------------------------------------------
    seen_in_file            = {}
    increasing              = {}
    formula_terms           = {}
    ancillary_variables     = set()
    fields_in_file          = FieldList()
    for data_var in variables:

        s = _create_a_field(nc, data_var, var_attr, seen_in_file,
                            formula_terms, 
                            ancillary_variables, 
                            increasing, 
                            global_attributes)
        
        s.setpriv('file', filename)
        
        fields_in_file.append(s)                                    
    #--- End: for

    # ----------------------------------------------------------------
    # If any transform equation terms are other fields then find out
    # which fields they are
    # ----------------------------------------------------------------
    if formula_terms:

        #find the ft fields
        ncvar_to_field = {}
        ft_ncvars      = []
        for transform in formula_terms.itervalues():
            for value in transform.itervalues():
                try:
                    ft_ncvars.append(value['ncvar'])
                except (TypeError, KeyError):
                    continue
        #--- End: for

        index = 0
        while ft_ncvars:
            try:
                f = fields_in_file[index]
            except IndexError:
                break

            # Still here?
            ncvar = f.ncvar            
            if ncvar in ft_ncvars:
                # Remove formula_terms transforms from fields
                # contained within formula_terms transforms
                if hasattr(f.space, 'transform'):
                    transforms = f.space.transform
                    for key in transforms.keys():
                        if transforms[key].isformula_terms:
                            del transforms[key]
                            for k in f.space:
                                if hasattr(f.space[k], 'transform'):
                                    del f.space[k].transform
                #--- End: if
                ft_ncvars.remove(ncvar)
                ncvar_to_field[ncvar] = f
                fields_in_file.pop(index)
            else:
                index += 1
        #--- End: while

        for f in fields_in_file:
            try:
                transforms = f.space.transform.values()
            except AttributeError:
                continue
        
            # Still here? Then loop round this field's transforms
            for transform in transforms:
                if hasattr(transform, 'grid_mapping_name'):
                    continue
                # Still here? Then loop round this formula_terms
                # transform's terms and values
                for term, value in transform.iteritems():
                    try:
                        transform[term] = ncvar_to_field[value['ncvar']].copy()
                    except (TypeError, KeyError): #(IndexError, TypeError, KeyError):
                        continue
                #--- End: for
            #--- End: for
        #--- End: for
    #--- End: if
        
    # ----------------------------------------------------------------
    # For each field that has ancillary variables, replace its list of
    # netCDF variable names with a FieldList.
    # ----------------------------------------------------------------
    if ancillary_variables:
        ncvar_to_field = {}
        index          = 0
        while ancillary_variables:
            try:
                f = fields_in_file[index]
            except IndexError:
                break
            
            # Still here?
            ncvar = f.ncvar            
            if ncvar in ancillary_variables:
                ancillary_variables.discard(ncvar)
                ncvar_to_field[ncvar] = f
                fields_in_file.pop(index)
            else:
                index += 1
        #--- End: while
        
        for f in fields_in_file:
            if not hasattr(f, 'ancillary_variables'):
                continue        
            av_fields = FieldList()
            for ncvar in f.ancillary_variables:
                av_fields.append(ncvar_to_field[ncvar].copy())
            f.ancillary_variables = av_fields
        #--- End: for
    #--- End: if

    # ----------------------------------------------------------------
    # Close the file
    # ----------------------------------------------------------------
    if close:
        nc.close()
    
    return fields_in_file
#--- End: def


def _create_a_field(nc, data_var, var_attr, seen_in_file,
                    formula_terms, 
                    ancillary_variables,
                    increasing,
                    global_attributes):
    '''Private function'''
    
    def _create_Coordinate(f, key, nc, ncvar, var_attr):
        '''

Create a coordinate variable
'''
        c = Coordinate(attributes = var_attr[ncvar])

        c.ncvar = ncvar
        
        data = nc.variables[ncvar]
 
        # ------------------------------------------------------------
        # Collapse a string valued auxiliary
        # coordinate. E.g. ['a','b','c'] becomes ['abc']
        # ------------------------------------------------------------
        if data.dtype.kind is 'S':
            new_shape = data.shape[0:-1]
            new_size  = numpy.prod(new_shape)
            strlen    = data.shape[-1]
            data      = data[...].reshape((new_size, strlen))
            
            # Make sure everything is ok if data is a masked array
            try:
                data.set_fill_value(' ')
                data = data.filled()
            except AttributeError:
                pass

            new = [''.join(x).rstrip() for x in data]
            new = numpy.array(new, dtype='|S%(strlen)d' % locals())

            data = new.reshape(new_shape)

            del new
        #--- End: if

        c._data = Data(data,
                       units      = c.Units,
                       _FillValue = getattr(c, '_FillValue', None))

        # ---------------------------------------------------------
        # Add any bounds
        # ---------------------------------------------------------
        if hasattr(c, 'climatology'):
            c.bounds      = c.climatology
            c.climatology = True

        if hasattr(c, 'bounds'):
            ncbounds = c.bounds

            b = CoordinateBounds(attributes = var_attr[ncbounds])
            b.ncvar = ncbounds
            
            b._data = Data(nc.variables[ncbounds],
                           units      = c.Units,
                           _FillValue = getattr(b, '_FillValue', None))

            if b.hasattr('formula_terms'):
                b.delattr('formula_terms')

            c.bounds = b
        #--- End: if

        if key.startswith('d'): # d for dim
            direction              = dimension_direction(c)
            increasing[ncvar]      = direction
            f.space.increasing[key] = direction
        
        # ---------------------------------------------------------
        # Set the coordinate and bounds' Data objects' metadata
        # ---------------------------------------------------------
        c = set_Data_attributes(f, key, c)
        
        # ---------------------------------------------------------
        # Return the coordinate
        # ---------------------------------------------------------
        return c
    #--- End: def

    def _create_CellMeasures(f, key, nc, ncvar, var_attr):
        '''

Create a cell measures variable
'''

        cell = CellMeasures(attributes = var_attr[ncvar])
        cell.ncvar = ncvar

        cell._data = Data(nc.variables[ncvar],
                          units      = cell.Units,
                          _FillValue = getattr(cell, '_FillValue', None))
        
        # ---------------------------------------------------------
        # Set the cell measures' Data objects' metadata
        # ---------------------------------------------------------
        cell = set_Data_attributes(f, key, cell)
        
        return cell
    #--- End: def

    # Add global attributes to the data variable's properties, unless
    # the data variables already has a property with the same name.
    for attr, value in global_attributes.iteritems():
        try:
            var_attr[data_var][attr]
        except KeyError:
            var_attr[data_var][attr] = value

    # Take cell_methods out of the data variable's properties as it
    # will need special processing once the space has been defined
    if var_attr[data_var].has_key('cell_methods'):
        cell_methods = CellMethods(var_attr[data_var].pop('cell_methods'))
    else:
        cell_methods = None

    # Take add_offset and scale_factor out of the data variable's
    # properties as they will be dealt with by the variable's Data
    # object
    for attr in ('add_offset', 'scale_factor'):
        if var_attr[data_var].has_key(attr):
            del var_attr[data_var][attr]

    # ----------------------------------------------------------------
    # Initialize the field with the data variable and its attributes
    # ----------------------------------------------------------------
    t = Field(attributes = var_attr[data_var])

    t.ncvar = data_var
    t.space = Space()
    t._data = Data(nc.variables[data_var],
                   units      = t.Units,
                   _FillValue = getattr(t, '_FillValue', None))

    # Initialize mappings of nc variable and dimension names to
    # internal field names. E.g. ncvar_to_var['geo_region']='aux0',
    # ncdim_to_dim['depth']='dim1'
    ncvar_to_var       = {}
    ncdim_to_dim       = {}
    t.space.increasing = {}
        
    t.space.dimensions['data'] = []
    t.space.nc_dimensions      = {}
    
    # ----------------------------------------------------------------
    # Add dimensions and dimension coordinates to the field
    # ----------------------------------------------------------------
    dimN = 0
#    uncN = 0

    for ncdim in map(str, nc.variables[data_var].dimensions):
        
        # Set the internal field name for this dimension and map the
        # nc name to it
        dim = 'dim%(dimN)d' % locals()
        dimN += 1         # Increment dimension number
        ncvar_to_var[ncdim] = dim
        ncdim_to_dim[ncdim] = dim
        
        # Update data dimension name and set dimension size
        t.space.nc_dimensions[dim]   = ncdim
        t.space.dimension_sizes[dim] = len(nc.dimensions[ncdim])
        t.space.dimensions['data'].append(dim)
        
        # If a dimension coordinate exists then create it in the field
        if nc.variables.has_key(ncdim):

            t.space.dimensions[dim] = [dim]

            try:
                t.space[dim] = copy.deepcopy(seen_in_file[ncdim])
                t.space.increasing[dim] = increasing[ncdim]
            except KeyError: 
                pass
                coord = _create_Coordinate(t, dim, nc, ncdim, var_attr)

                seen_in_file[ncdim] = coord                
                t.space[dim]        = coord
        
#            # --------------------------------------------------------
#            # Compression by gathering
#            # --------------------------------------------------------
#            try:
#                t.space[dim].compress = CoordinateList(t.space[dim].compress.split())
#            except AttributeError:
#                pass
#            else:
#                # Create the uncompressed coordinates
#                for i in xrange(len(t.space[dim].compress)):
#                    try:
#                        t.space[dim].compress[i] = copy.deepcopy(seen_in_file[ncvar])
#                    except KeyError:
#                        pass
#                        t.space[dim].compress[i] = _create_Coordinate(t, dim, 
#                                                                     nc, ncvar, var_attr)
#                        seen_in_file[ncvar] = t.space[dim].compress[i]
#                #--- End: for
#
#                t.space[dim].compress = t.space[dim].compress.split()
#                for i in xrange(len(t.space[dim].compress)):
#                    # Create the coordinate without updating the
#                    # dimension_sizes or dimensions dictionaries
#                    unc = 'unc'+str(uncN)
#                    ncvar = t.space[dim].compress[i]
#                    t.space[dim].compress[i] = unc
#                    t.space.nc_dimensions[unc] = ncvar
#                    try:
 #                        t.space[unc] = copy.deepcopy(seen_in_file[ncvar])
#                    except KeyError:
#                        t.space[unc] = _create_Coordinate(nc, ncvar, var_attr)
#                        seen_in_file[ncvar] = t.space[unc] 
#
#                    # Increment ?? number
#                    uncN += 1
#                #--- End: for

        else:
            # Dimension coordinate for this dimension does not exist
            increasing[ncdim]       = True
            t.space.increasing[dim] = True
        #--- End: if

    #--- End: for

    # ----------------------------------------------------------------
    # Add auxiliary coordinates to the field
    # ----------------------------------------------------------------
    auxN = 0
    try:
        coordinates = t.getattr('coordinates')
    except AttributeError:
        pass
    else:
        # Allow for (incorrect) comma separated lists
        for ncvar in re.split('\s+|\s*,\s*', coordinates):

            # Skip dimension coordinates which are in the list
            if ncvar in nc.variables[data_var].dimensions:
                continue

            # Set the internal field name for this auxiliary
            # coordinate and map the nc name to it
            aux = 'aux%(auxN)d' % locals()
            auxN += 1       # Increment auxiliary coordinate number
            ncvar_to_var[ncvar] = aux

            # If the auxiliary coordinate variable exists then create
            # a new coordinate variable for it in the field dictionary
            if nc.variables.has_key(ncvar):

                # Set auxiliary coordinate dimension names
                t.space.dimensions[aux] = []
                for ncdim in nc.variables[ncvar].dimensions:
                    try:
                        t.space.dimensions[aux].append(ncdim_to_dim[ncdim])
                    except KeyError:
                        # Why try here? because of things like bounds, strlen
                        pass
                #--- End: for

                key = aux

                # ----------------------------------------------------
                # Create a dimension for a numeric scalar coordinate
                # and turn it inot a dimension coordinate.
                # ----------------------------------------------------
                if not t.space.dimensions[aux]:
                    dim = 'dim%(dimN)d' % locals()
                    dimN += 1         # Increment dimension number
                    
                    t.space.dimension_sizes[dim] = 1
                    t.space.dimensions[dim]      = [dim]

                    if nc.variables[ncvar].dtype.kind is 'S':
                        # String valued scalar coordinate                        
                        t.space.dimensions['aux'] = dim
                        t.space.increasing['dim'] = True
                    else:  
                        # Numeric valued scalar coordinate                           
                        del t.space.dimensions[aux]
                        key = dim
                #--- End: if

                try:                   
                    t.space[key] = copy.deepcopy(seen_in_file[ncvar])
                except KeyError:
                    coord = _create_Coordinate(t, key, nc, ncvar, var_attr)

                    seen_in_file[ncvar] = coord
                    t.space[key]        = coord
                #--- End: try
            #--- End: if

        #--- End: for

        t.delattr('coordinates')
    #--- End: try

    # ----------------------------------------------------------------
    # Now that we've created all of the dimension and auxiliary
    # coordiantes, set the field's Data objects's metadata
    # ----------------------------------------------------------------
    t = set_Data_attributes(t, 'data', t)

    # ----------------------------------------------------------------
    # Add formula_terms to coordinates' Transform attributes
    # ----------------------------------------------------------------
    transN = 0
    for coord in t.space.keys():
 
        coord_ncvar = t.space[coord].ncvar

        try:
            transform = formula_terms[coord_ncvar]
        except KeyError:
            pass
        else:
            # This coordinate's formula_terms transform has already
            # been created, so just copy it. (We have to do this
            # because the 'formula_terms' public attribute on the
            # coordinate has been deleted.)
            transform_id = 'trans%(transN)d' % locals()
            transN += 1       # Increment auxiliary coordinate number
            t.space[coord].transform = transform_id
            if not hasattr(t.space, 'transform'):
                t.space.transform = CfDict()
            t.space.transform[transform_id] = copy.deepcopy(transform)
            continue
        #-- End: try

        form_terms = t.space[coord].getattr('formula_terms', None)
        if form_terms is None:
            # This coordinate doesn't have a formula_terms attribute
            continue

        # Still here? Then this coordinate has a formula_terms
        # transform that needs creating
        transform      = Transform()
        transform.name = t.space[coord].standard_name
        
        # Add the equation terms and references to their values to to
        # new auxiliary coordinate's transform.
        ft = re.split('\s+|\s*:\s+', form_terms)
        for term, ncvar in zip(ft[0::2], ft[1::2]):
            try:
                # Try a coordinate variable from this field
                value = ncvar_to_var[ncvar]
            except KeyError:
                # Assume it's another field. We don't know what this
                # field is yet, so we'll give it its ncvar name and
                # work out which field it is later, if we
                # can. E.g. value={'ncvar':'pstar'}
                value = {'ncvar': ncvar}
    
            transform[term] = value
        #--- End: for 

        transform_id = 'trans%(transN)d' % locals()
        transN += 1       # Increment transform number

        t.space[coord].transform = transform_id

        try:
            t.space.transform
        except AttributeError:
            t.space.transform = CfDict()

        t.space.transform[transform_id] = transform

        coord_ncvar = t.space[coord].ncvar
        formula_terms[coord_ncvar] = transform

        t.space[coord].delattr('formula_terms')
    #--- End: for

    # ----------------------------------------------------------------
    # Add grid mapping to coordinates' Transform attributes
    # ----------------------------------------------------------------
    try:
        grid_mapping = t.getattr('grid_mapping')
    except AttributeError:
        # No grid mapping for this field
        pass
    else:
        # Got a grid_mapping for this field
        try:
            t.space.transform
        except AttributeError:
            t.space.transform = CfDict()                

        for gm in grid_mapping.split():

            try:
                var_attr[gm]
            except KeyError:
                continue

            try:
                var_attr[gm]['grid_mapping_name']
            except KeyError:
                var_attr[gm]['grid_mapping_name'] = gm
            grid_mapping_name = var_attr[gm]['grid_mapping_name']

            transform_id = 'trans%(transN)d' % locals()
            transN += 1     

            t.space.transform[transform_id] = Transform(**var_attr[gm])
            t.space.transform[transform_id].ncvar = gm
            t.space.transform[transform_id].name  = grid_mapping_name

            # Work out which coordinates need the transform attached
            # to them
            if grid_mapping_name in ('rotated_latitude_longitude'):
                for coord in t.space.get_keys('^dim|^aux'):
                    if t.space[coord].name() in ('longitude', 
                                                'latitude',
                                                'grid_longitude',
                                                'grid_latitude'):
                        t.space[coord].transform = transform_id

            elif grid_mapping_name == 'latitude_longitude':
                for coord in t.space.get_keys('^dim|^aux'):
                    if t.space[coord].name() in ('longitude', 
                                                'latitude'):
                        t.space[coord].transform = transform_id

            else:
                for coord in t.space.get_keys('^dim|^aux'):
                    if t.space[coord].name() in ('projection_x_coordinate', 
                                                'projection_y_coordinate'):
                        t.space[coord].transform = transform_id
            #--- End: if
        #--- End: for

        t.delattr('grid_mapping')
    #--- End: try

    # ----------------------------------------------------------------
    # Add cell measures to the field
    # ----------------------------------------------------------------
    try:
        # Try to parse a cell measures attribute
        cell_measures = re.split('\s*(\w+):\s*', t.getattr('cell_measures'))
    except AttributeError:
        # No cell measures for this field
        pass
    else:
        # Got cell measures
        cell_measures.pop(0)
        cmN = 0
        for measure, ncvar in zip(cell_measures[::2], 
                                  cell_measures[1::2]):

            cm = 'cm%(cmN)d' % locals()
            cmN += 1                # Increment cell measures number

            # Set cell measures' dimension names
            try:
                t.space.dimensions[cm] = [ncdim_to_dim[ncdim]
                                          for ncdim in nc.variables[ncvar].dimensions]
            except KeyError:
                # This cell measure variable does exist in the file
                cmN -= 1
                continue

            try:
                t.space[cm] = copy.deepcopy(seen_in_file[ncvar])
            except KeyError:

                try:             
                    var_attr[ncvar]
                except KeyError:
                    del t.space.dimensions[cm]
                else:                        
                    cell = _create_CellMeasures(t, cm, nc, ncvar, var_attr)
                    
                    try:
                        cell.standard_name
                    except AttributeError:
                        cell.standard_name = measure

                    # ------------------------------------------------
                    # Keep a record of field components which exist in
                    # this file so that they may be shared easily
                    # ------------------------------------------------
                    seen_in_file[ncvar] = cell
                    t.space[cm]         = cell
            #--- End: try    
        #--- End: for

        t.delattr('cell_measures')
    #--- End: try

    # Make dimension coordinate with no aux, cell_measures scalar and
    # drop their dimensions from the data - moved to read.py
#    t.squash()

    # -----------------------------
    # Add cell methods to the field
    # -----------------------------
    if cell_methods is not None:
        t.cell_methods = _parse_cell_methods(cell_methods, t)

    # ----------------------------------------------------------------
    # Parse ancillary variables to a list of netCDF variable names,
    # which will get converted to a FieldList later.
    # ----------------------------------------------------------------
    if hasattr(t, 'ancillary_variables'):
        t.ancillary_variables =  t.ancillary_variables.split()
        ancillary_variables.union(set(t.ancillary_variables))

    # ----------------------------------------------------------------
    # Parse flag values, flag masks and flag meanings into the field's
    # Flags attribute
    # ----------------------------------------------------------------
    if hasattr(t, 'Flags'):
        print 'sorting flags ...'
        t.Flags.sort()

    del t.space.increasing

    # Return the finished field
    return t
#--- End: def

def _parse_cell_methods(cell_methods, field):
    '''
'''
    # Change each 'name' value to a standard_name (or space coordinate
    # key) and create the 'dim' key
        
    # From the CF conventions (1.5): In the specification of this
    # attribute, name can be a dimension of the variable, a scalar
    # coordinate variable, a valid standard name, or the word 'area'.
    for cell_method in cell_methods:
        for i in xrange(len(cell_method['name'])):

            # Since input scalar coordinate variables have now been
            # converted into 1d dimension coordinates, name being "a
            # dimension of the variable, a scalar coordinate variable"
            # may be checked by testing name against
            # field.space.nc_dimensions.values().
            found_ncdimension = False
            for dim, ncdim in field.space.nc_dimensions.iteritems():
                if cell_method['name'][i] == ncdim:
                    cell_method['no_coords'][i] = False
                    cell_method['dim'][i]       = dim
                    try:                            
                        cell_method['name'][i] = getattr(field.space[dim], 
                                                         'standard_name',
                                                         ncdim)
                    except KeyError:
                        cell_method['name'][i] = ncdim
                    found_ncdimension = True
                    break
            #--- End: for
            if found_ncdimension:
                continue

            # Still here? If the 'name' is the netCDF variable name of
            # a coordinate, then set 'dim' and, if possible, replace
            # 'name' with the coorindate's standard name.
            found_ncvar = 0
            for key in field.space.get_keys('dim|aux'):
                ncvar = field.space[key].ncvar
                if cell_method['name'][i] == ncvar:
                    cell_method['name'][i] = getattr(field.space[key], 
                                                     'standard_name',
                                                     ncvar)
                    cell_method['dim'][i]  = key
                    found_ncvar = 1
                    break
            #--- End: for
            if found_ncvar:
                continue

            # Still here? If the 'name' is the standard name of a
            # coordinate (or the string "area"), then set 'dim'
            cell_method['no_coords'][i] = True
            if cell_method['name'][i] == 'area':
                cell_method['dim'][i] = None
            else:
                for key in field.space.get_keys('dim|aux'):
                    try:
                        sn = field.space[key].standard_name
                    except AttributeError:
                        pass
                    else:
                        if cell_method['name'][i] == sn:
                            dim = field.space.coord(sn, key=True, exact=True)
                            cell_method['dim'][i] = dim
                            break
                #--- End: for
            #--- End: if

        #--- End: for
    #--- End: for

    return cell_methods
#--- End: def
