import glob
import os
import struct
from .utils     import _files
from .field     import FieldList
from .netcdf    import read      as netcdf_read
from .pp        import read      as pp_read
from .aggregate import aggregate as cf_aggregate

def _file_format(openfile):
    '''

Private function to read a file's magic number.

Takes a file object as input and assumes that we're currently
pointing to the beginning of the file. The file is rewound to the
beginning after reading the magic number.

Returns a string giving the format of the input file.
''' 

    if openfile.name.startswith('http://'):
        # Assume that a URL is netCDF
        return 'nc'

    # Read the magic number
    try:
        magic_number = struct.unpack('=L', openfile.read(4))[0]
    except struct.error:
        raise IOError, "File %s is empty (contains fewer than 4 words)" % \
            openfile.name

    # Reset the pointer to the beginning of the file
    openfile.seek(0, os.SEEK_SET)

    # ----------------------------------------------------------------
    # netCDF
    # ----------------------------------------------------------------
    if magic_number in (21382211, 1128547841, 38159427):
        return 'nc'

    # ----------------------------------------------------------------
    # PP
    # ----------------------------------------------------------------
    if magic_number in (256, 65536):
        return 'pp'

    # ----------------------------------------------------------------
    # Add more file formats here ...
    # ----------------------------------------------------------------

    # Still here?
    raise IOError, "File %s has unsupported format" % openfile.name
#--- End: def

def _read_a_file(filename, ignore_ioerror=False, verbose=False,
                 close=False,
                 aggregate=True,
                 aggregate_options={},
                 squash=True,
                 ex={},
                 extract={}):

    # Find this input file's format
    if filename.startswith('http://'):
        format = 'nc'
    else:        
        try:
            openfile = open(filename, 'rb')             
            _files(openfile)
        except IOError as error:
            if ignore_ioerror:
                if verbose:
                    print 'Ignoring IOError: %(error)s' % locals()
                return FieldList()
            raise IOError, error
              
        try:
            format = _file_format(openfile)        
        except IOError as error:
            if ignore_ioerror: 
                if verbose:
                    print 'Ignoring IOError: %(error)s' % locals()
                return FieldList()
            raise IOError, error

    # ----------------------------------------------------------------
    # Read the file into fields
    # ----------------------------------------------------------------
    tmp_aggregate_options = aggregate_options
    if format == 'nc':
        fields = netcdf_read(filename, close=close,)
        
    elif format == 'pp': 
        fields = pp_read(openfile, close=close)

        # For PP fields, the default is strict_units = False
        tmp_aggregate_options = aggregate_options.copy()
        if not tmp_aggregate_options.has_key('strict_units'):
            tmp_aggregate_options['strict_units'] = False
    
    # Add more file formats here ...

    # ----------------------------------------------------------------
    # Extract matching fields before aggregation
    # ----------------------------------------------------------------
#    if ex:
#        fields = fields.ex(ex)
#    if extract:
#        fields = fields.extract(extract)

    # ----------------------------------------------------------------
    # Squash the fields, i.e. if a size 1 dimension coordinate has no
    # aux, cell_measures then make it scalar and drop its dimension
    # from the field's data.
    # ----------------------------------------------------------------
    if squash:
        fields.squash()

    # ----------------------------------------------------------------
    # Aggregate fields
    # ----------------------------------------------------------------
    if aggregate and len(fields) > 0:
        cf_aggregate(fields, **tmp_aggregate_options)

    return fields
#--- End: def

def read(files, verbose=False, index=None, 
         ignore_ioerror=False, close=False,
         aggregate=True,
         squash=True,
         ex={}, extract={}):
    '''

Read fields from input files from disk or from an OPeNDAP server.

All data arrays from all field components are not read from disk but
are stored in the returned fields as file pointers.

NetCDF dimension names are stored in the *nc_dimensions* attribute of
a field's space and netCDF variable names are stored in the private
attribute *ncvar* of the field and each of its space components.

Fields referenced by formula_terms transforms are not included in the
returned list of fields. A field may be referenced in the transforms
of multiple output fields, so in-place changes to that field in one
transform will be seen in the other fields which reference it.

:argument files: A string or sequence of strings giving the file names
                 or OPenDAP URLs from which to read fields. For files
                 on disk, the file names may contain wild cards as
                 understood by the :mod:`glob` module.

:type files: str or sequence

:parameter index: *Optional*. If None then return all possible fields
                  from the input files. If a non-negative integer then
                  only return the field with this index in the
                  otherwise full list of fields.

:type index: None or int

:parameter verbose: *Optional*. If True then print information to
                    stdout.

:type verbose: bool

:parameter close: *Optional*. If True then close each file after it
                  has been read.

:type close: bool

:parameter ignore_ioerror: *Optional*. If True then ignore files which
                           produce IO errors (such as empty files, for
                           example).

:type ignore_ioerror: bool

:parameter aggregate: *Optional*. If True then aggregate the fields
                      read in from all input files into as few fields
                      as possible using the CF aggregation
                      rules. (Note that this code has not yet been
                      optimsed for speed or memory usage.)

:type aggregate: bool


:returns: A list of fields in a :class:`~cf.FieldList` object.

**Examples:**

>>> f = cf.read('file*.nc')
>>> type(f)
<class 'cf.field.FieldList'>
>>> f
[<CF Field: pmsl(30, 24)>,
 <CF Field: z-squared(17, 30, 24)>,
 <CF Field: temperature(17, 30, 24)>,
 <CF Field: temperature_wind(17, 29, 24)>]

>>> cf.read('file*.nc')[0:2]
[<CF Field: pmsl(30, 24)>,
 <CF Field: z-squared(17, 30, 24)>]

>>> cf.read('file*.nc', index=0)
[<CF Field: pmsl(30, 24)>]

>>> cf.read('file*.nc')[-1]
<CF Field: temperature_wind(17, 29, 24)>

>>> cf.read('file*.nc').extract(units='K')
[<CF Field: temperature(17, 30, 24)>,
 <CF Field: temperature_wind(17, 29, 24)>]

>>> cf.read('file*.nc').extract(standard_name='.*pmsl*', 
                                units='K|Pa')[0]
<CF Field: pmsl(30, 24)>
'''

    # Initialize the output list of fields
    field_list = FieldList()

    aggregate_options = {}
    if isinstance(aggregate, dict):
        aggregate_options = aggregate
        aggregate         = True

    # Count the number of fields (in all files) and the number of
    # files
    field_counter = -1
    file_counter  = 0

    if isinstance(files, basestring):
        files = (files,)

    for file_glob in files:

        if file_glob.startswith('http://'):
            # Do not glob a URL
            files2 = (file_glob,)
        else:
            # Do glob files on disk
            files2 = glob.iglob(file_glob)

        for filename in files2:

            # Print some informative messages
            if verbose and index is None:
                print 'File: %(filename)s' % locals()

            # --------------------------------------------------------
            # Read the file into fields
            # --------------------------------------------------------
            fields = _read_a_file(filename,
                                  ignore_ioerror=ignore_ioerror,
                                  verbose=verbose,
                                  close=close,
                                  aggregate=aggregate,
                                  aggregate_options=aggregate_options,
                                  squash=squash,
                                  ex=ex,
                                  extract=extract)
            
            # --------------------------------------------------------
            # Add this file's fields to those alread read from other
            # files
            # --------------------------------------------------------
            field_list.extend(fields)
   
            field_counter = len(field_list)
            file_counter += 1

            # Print some informative messages
            if verbose and index is None:
                i = field_counter - len(fields)
                for f in fields:
                    print '%d: %s' % (i, repr(f))
                    i += 1
            #--- Endf: if

            # --------------------------------------------------------
            # If we only want one field from all input files then
            # break now if we have got it
            # --------------------------------------------------------
            if index is not None and field_counter >= index+1:
                break
        #--- End: for
            
        # ------------------------------------------------------------
        # If we only one field from all input files then break now if
        # we have got it
        # ------------------------------------------------------------
        if index is not None and field_counter >= index:
            break
    #--- End: for     

    # Error check
    if not ignore_ioerror:
        if not file_counter:
            raise RuntimeError('No files found')
        if not field_list:
            raise RuntimeError('No fields found from '+str(file_counter)+' files')
    #--- End: if
    if index >= len(field_list):
        raise IndexError('FieldList index='+str(index)+' is out of range')

    # Print some informative messages
    if verbose:
        if index is None:  
            print 'Read %d field%s from %d file%s' % \
                (field_counter, ('s' if field_counter!=1 else ''),
                 file_counter , ('s' if file_counter !=1 else ''))
        else:
            print 'File: %(filename)s' % locals()
            print '%d: %s' % (0, repr(field_list[index]))
            print 'Read 1 field from 1 file'
    #--- End: if
    
    # ----------------------------------------------------------------
    # 
    # ----------------------------------------------------------------
    try:
        field_list = field_list[index]
    except TypeError:
        pass

    # ----------------------------------------------------------------
    # Aggregate the output fields
    # ----------------------------------------------------------------    
    if file_counter > 1 and aggregate:
        cf_aggregate(field_list, **aggregate_options)
        if verbose:
            nfields = len(field_list)
            print 'Fields aggregated into %d field%s' % \
                (nfields,
                 ('' if nfields==1 else 's'))
    #--- End: def

    return field_list
#--- End: def
