from .utils import RTOL, ATOL, equals, CfDict
from .units import Units


class Space(CfDict):
    '''

A CF space object defining a space's dimensionality, coordinates, cell
measures and transformations in a dictionary-like object. Refer to the
`Space structure` section of the :mod:`cf` module for details.

:argument *args: Keys and values are initialized exactly as for a
                 built-in dict.

:argument **kwargs: Keys and values are initialized exactly as for a
                    built-in dict.

**Overloaded operators**

The ``in`` (set membership) operator is overloaded to use numerically
tolerant equality.

**Special attributes:**

======================= =================================================
Attribute               Description
======================= =================================================
:attr:`dimension_sizes` The size of each dimension.

:attr:`dimensions`      The dimensions belonging to each space component.

:attr:`nc_dimensions`   The netCDF dimension name for each dimension
                        in the `dimension_sizes` attribute. If
                        present when the space is written to a netCDF
                        file, then used for output netCDF dimension
                        names.

:attr:`transform`       A dictionary of :class:`~cf.Transform` objects.
======================= =================================================

**Methods and attributes defined here:**
'''

    def __init__(self, *args, **kwargs):
        '''

Set the built-in dictionary from the arguments and initialize other
attributes.
'''
        # Create the dimension dictionaries
        self.dimension_sizes = {}
        self.dimensions      = {}

        super(Space, self).__init__(*args, **kwargs)
    #--- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)
'''
        return '<CF %s: %s>' % (self.__class__.__name__, 
                                tuple(self.dimension_sizes.values()))
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)
'''
        def _print_coord(space, key, dimension_coord):
            '''Private function called by __str__'''
            try: 
                space[key]
            except KeyError: 
                size = space.dimension_sizes[key]
                try:
                    name = space.nc_dimensions[key]
                except (KeyError, AttributeError):
                    name = key
                return '%s(%d)' % (name, size)

            # Still here?
            if dimension_coord:
                # Dimension coordinate
                shape = space[key].shape
                if shape:
                    shape = '(%d)' % space[key].size
                else:
                    shape = ''
                name = space[key].name(ncvar=True, default=key)
            else:
                # Auxiliary coordinate
                try:
                    name = space[key].name(ncvar=True, default=key)
                except TypeError:
                    name = key
                shape = space.dimensions[key][:]
                for i in xrange(len(shape)):
                    try:
                        shape[i] = space[shape[i]].name(ncvar=True, default=shape[i])
                    except KeyError:
                        shape[i] = space.nc_dimensions[shape[i]]
                shape = str(tuple(shape)).replace("'", "")
                shape = shape.replace(',)', ')')
            #--- End: if
            x = [name]
            x.append(shape)

            try:
                space[key].compress
            except AttributeError:
                units = ''
                if space[key].Units and space[key].Units != Units('1'):
                    units = space[key].units

                first = space[key].first_datum
                last  = space[key].last_datum


#                if space[key].istime():
#                    # Time valued coordinate's first and last values
#                    if space[key].size == 1:
#                        x.append(' -> %(first)s' % locals())
#                    else:
#                        x.append(' -> %(first)s to %(last)s' % locals())
                if space[key].dtype.kind == 'S':
                    # String valued coordinate's first and last values
                    if space[key].size == 1:
                        x.append(" -> '%s' %s" % (first, units))
                    else:
                        x.append(" -> '%s' to '%s' %s" %  (first, last, units))
                else:
                    # Numeric valued coordinate's first and last values
                    if space[key].size == 1:
                        x.append(' -> %.7g %s' % (first, units))
                    else:
                        x.append(' -> %.7g to %.7g %s' % (first, last, units))

            else:
                x.append(' -> compressed ')
                compressed = []
                for unc in space[key].compress:
                    shape = str(unc.size)
                    compressed.append(unc.name(ncvar=True,
                                               default='unc')+'('+shape+')')
                x.append(', '.join(compressed))

            return ''.join(x)
        #--- End: def

        string = ['Dimensions      : ']
        x = []
        for dim in sorted(self.dimension_sizes):
            x.append(_print_coord(self, dim, True))
        string.append('\n                : '.join(x))
        string.append('\n')

        string.append('Auxiliary coords: ')
        x = []
        for aux in sorted(self.get_keys('^aux')):
            x.append(_print_coord(self, aux, False))
        string.append('\n                : '.join(x))
        string.append('\n')

        # Cell measures
        cell_measures = self.get_keys('^cm')
        if cell_measures:
            string.append('Cell measures   : ')
            x = []
            for cm in sorted(cell_measures):
                x.append(_print_coord(self, cm, False))
            string.append('\n                : '.join(x))
            string.append('\n')

        # Virtual auxiliary coordinates
        try:
            transforms = self.transform.values()
        except AttributeError:
            pass
        else:
            string.append('Virtual coords  : ')
            x = []
            for transform in transforms:
                try:
                    transform['grid_mapping_name']
                    continue
                except KeyError:
                    x.append(repr(transform))
            #-- End: for
            string.append('\n                : '.join(x))

        return ''.join(string)
    #--- End: def

    def dump(self, id=None, omit=()):
        '''
        
Return a string containing a full description of the space.

:keyword id: *Optional*. Set the common prefix of variable component
             names. If None then defaults to the class name.

:keyword id: str

:keyword omit: *Optional*. Omit the given attributes the description.

:type omit: sequence

:returns: A string containing the description of the space.

.. seealso:: :func:`cf.dump`
'''
        if id is None:
            id = self.__class__.__name__

        try:
            string = ['%s space' % self.name]
        except AttributeError:
            string = ['space']
        string.append(''.ljust(len(string[0]), '-'))

        string.append('%s.dimension_sizes = %s\n' % (id, self.dimension_sizes))
        
        for key in sorted(self.dimensions):
            dims = self.dimensions[key]
            string.append("%(id)s.dimensions['%(key)s'] = %(dims)s" % locals())

        string.append('')

        for key in sorted(self.get_keys('^dim')) + \
                sorted(self.get_keys('^aux')) + \
                sorted(self.get_keys('^cm')):
            string.append(self[key].dump(id="%(id)s['%(key)s']" % locals()))
           
        try:
            for key, transform in self.transform.iteritems():
                string.append(transform.dump( \
                    id="%(id)s.transform['%(key)s']" % locals()))
        except AttributeError:
            pass               
        
        return '\n'.join(string)
    #--- End: def

    def equals(self, other, rtol=None, atol=None):
        '''

Return True if two instances are congruent in that

    1. For each group of space component types (dimension coordinate,
       auxiliary coordinate, cell measures, etc.) there are an equal
       number of keys and each key's value equals a value in the other
       space. Note that within a group of component types, the key
       names need not be the same.
  
    2. Each dimension in one space has an equally sized dimension in
       the other space.

    3. The unordered set of transforms in one space equals that of the
       other space. Noe that The keys of the `transform` attribute
       dictionary need not be the same.

Equality of numbers is to within a tolerance. Refer to :mod:`cf` for
details.

:argument other: The variable to compare against for equality.

:type other: object
             
                                                 
:keyword atol: *Optional*. If None then use the default method for
               setting the absolute tolerance for numerical equality
               (refer to :mod:`cf` for details). If a number then set
               the absolute tolerance to this value for all such
               comparisons.

:type atol: None or float

:keyword rtol: *Optional*. If None then use the default method for
               setting the relative tolerance for numerical equality
               (refer to :mod:`cf` for details). If a number then set
               the relative tolerance to this value for all such
               comparisons.

:type rtol: None or float

:returns: True if the two objects are congruent, False otherwise.

    '''
        if self is other:
            return True
        
        # Check that each instance is the same type
        if type(self) != type(other):
            return False

        try:
            if set(self.dimension_sizes.values()) != \
                    set(other.dimension_sizes.values()):
                return False
        except AttributeError:
            pass

               
        if rtol is None:
            rtol = RTOL()
        if atol is None:
            atol = ATOL()              

        try:
            transform0 = self.transform
        except AttributeError:
            try:
                other.transform
                return False
            except AttributeError:
                pass
        else:
            try:
                transform1 = other.transform
            except AttributeError:
                return False

            self_keys  = sorted(transform0.keys())
            other_keys = sorted(transform1.keys())
            for key0 in self_keys:
                found_match = False
                for key1 in other_keys: 
                    if equals(transform0[key0], transform1[key1],
                              rtol=rtol, atol=atol):
                        found_match = True
                        other_keys.remove(key1)
                        break
                #--- End: for
                if not found_match:
                    return False
            #--- End: for
        #--- End: try

        for coord_type in ('^dim', '^aux', '^cm'):
            self_keys  = sorted(self.get_keys(coord_type))
            other_keys = sorted(other.get_keys(coord_type))
            for coord0 in self_keys:
                found_match = False
                for coord1 in other_keys: 
                    if equals(self[coord0], other[coord1], 
                              rtol=rtol, atol=atol):
                        found_match = True
                        other_keys.remove(coord1)
                        break
                #--- End: for
                if not found_match:
                    return False
            #--- End: for
        #--- End: for

        return True
    #--- End: def

    def coord(self, standard_name=None, role=None, key=False, exact=False, 
              axis=None, ncvar=None, long_name=None, position=None,
              **kwargs):
        '''

Find a coordinate of the space by name. Refer to
:meth:`cf.Field.coord` for details.

'''       
        dim, aux = True, True
        if role == 'dim':
            aux = False
        elif role == 'aux':
            dim = False

        found_name = 0

        if dim:
            dim_list = self.get_keys('^dim')
        else:
            dim_list = []
        if aux:
            aux_list = self.get_keys('^aux')       
        else:
            aux_list = []

        if position is not None:
            try:
                space_key = self.dimensions['data'][position]
            except (IndexError, TypeError, KeyError):
                return None
            if key:
                return space_key            
            return self[space_key]
        #--- End: if
 
#        if isinstance(standard_name, (int, long)):
#            try:
#                space_key = self.dimensions['data'][standard_name]
#            except (IndexError, KeyError):
#                return None
#            if key:
#                return space_key            
#            return self[space_key]
#        #--- End: if
 
        # Still here?
        y=None
        for x, attr in zip((standard_name, long_name, ncvar, axis),
                           ('standard_name', 'long_name', 'ncvar', 'axis')):

            if x is None:
                continue           

            found_name = 0
            for k in dim_list + aux_list:
                value = getattr(self[k], attr, None)
                if value is None: 
                    continue
                if ((exact and x == value) or
                    (not exact and value.startswith(x))):
                    if found_name:
                        return None
                    found_name += 1
                    space_key    = k
                    if y is not None and space_key != y:
                        return None
                    y = space_key
                #--- End: for
            #--- End: for
        #--- End: if      

        if y is not None:
            if key:
                return y
            return self[y]
        #--- End: if

        return None
    #--- End: def

#--- End: clas


class Transform(CfDict):
    '''

A CF transform object defining a coordinate's transformation in a
dictionary-like object.

The named parameters and their values of the transformation (i.e. the
transformation's mappings) comprise the object's key-value pairs.

A transformation is equivalent to either a netCDF(CF) 'formula_terms'
or 'grid_mapping' property. The latter is identified by the presence
of the 'grid_mapping_name' key which contains the mapping's name.

In the 'formula_terms' case, a mapping to a coordinate (as opposed to
another field) uses the coordinate's space key name as a pointer
rather than a copy of the coordinate itself.

Refer to the `Space structure` section of the :mod:`cf` module for
details.

:argument *args: Keys and values are initialized exactly as for a
                 built-in dict.

:argument **kwargs: Keys and values are initialized exactly as for a
                    built-in dict.


**Overloaded operators:**

The ``in`` (set membership) operator is overloaded to use numerically
tolerant equality.


**Special attributes:**

============ =======================================================
Attribute    Description
============ =======================================================
:attr:`name` The identifying name of the transformation. Typically a
             netCDF(CF) 'formula_terms' or 'grid_mapping_name'.
============ =======================================================


**Examples:**

>>> t
<CF Transform: atmosphere_sigma_coordinate>
>>> dump(t)
atmosphere_sigma_coordinate transform
-------------------------------------
Transform['ps'] = <CF Field: surface_air_pressure(73, 96)>
Transform['ptop'] = 0.05
Transform['sigma'] = 'dim0'

>>> t
<CF Transform: rotated_latitude_longitude>
>>> dump(t)
rotated_latitude_longitude transform
------------------------------------
Transform['grid_mapping_name'] = 'rotated_latitude_longitude'
Transform['grid_north_pole_latitude'] = 33.67
Transform['grid_north_pole_longitude'] = 190.0

**Methods and attributes defined here:**
'''
    def __repr__(self):
        '''
x.__repr__() <==> repr(x)
''' 
        try:
            return '<CF %s: %s>' % (self.__class__.__name__, self.name)
        except AttributeError:
            return '<CF %s: >' % self.__class__.__name__
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)
'''    
        return 'Transform       : %s' % repr(self)
    #--- End: def

    def dump(self, id=None, omit=()):
        '''

Return a string containing a full description of the transformation.

:keyword id: *Optional*. Set the common prefix of component names. If
             None then defaults to the class name.

:type id: str

:keyword omit: *Optional*. Omit the given attributes from the
               description.

:type omit: sequence


:returns:  A string containing the description of the object.

.. seealso:: :func:`cf.dump`
'''
        if id is None:
            id = self.__class__.__name__
            
        try:
            string = ['%s transform' % self.name]
        except AttributeError:
            string = [' transform']

        string.append(''.ljust(len(string[0]), '-'))

        for key in sorted(self.keys()):
            string.append("%s['%s'] = %s" % (id, key, repr(self[key])))

        return '\n'.join(string)
    #--- End: def

    @property
    def isformula_terms(self):
        '''
'''
        if 'grid_mapping_name' in self:
            return False
        return True
    #--- End: def

    @property
    def isgrid_mapping(self):
        '''
'''
        if 'grid_mapping_name' in self:
            return True
        return False
    #--- End: def

#--- End: class
