import copy
import numpy
import re
import textwrap
from .utils import RTOL, ATOL, equals, CfList
from .units import Units
from .value import Value

# ====================================================================
#
# Variable object
#
# ====================================================================

class Variable(object):
    '''
    
Base class for a field and its components.

Refer to the :ref:`variable data storage, access and manipulation
<variable_data_storage_access_and_manipulation>` in the top level
:mod:`cf` documentation for more details.

:keyword public_attributes: *Optional*. The new variable is given the
                            public attributes named and valued from
                            the key, value pairs of this
                            dictionary. The values are deep copied.

:type public_attributes: dict

:keyword private_attributes: *Optional*. The new variable is given the
                             private attributes named and valued from
                             the key, value pairs of this
                             dictionary. The values are deep copied.

:type private_attributes: dict

**Read only attributes**

The following attributes are read only on the variable and may not be
set.

============== ========================================  
Attribute      Description                             
============== ========================================
:attr:`array`  A numpy array deep copy of the data.    
:attr:`dtype`  Describes the format of the             
               elements in the data array (refer to          
               :obj:`numpy.ndarray.dtype`).                 
:attr:`ndim`   The data array's number of dimensions   
               (refer to :obj:`numpy.ndarray.ndim`).        
:attr:`shape`  The shape of the data array             
               (refer to :obj:`numpy.ndarray.shape`).       
:attr:`size`   The number of elements in the data array
               (refer to :obj:`numpy.ndarray.size`).        
:attr:`varray` A numpy view of the data.              
============== ========================================


**Reserved attributes:**

The following attributes have special definitions and it is
recommended to set them accordingly.

============== ======================================================
Attribute      Description                         
============== ======================================================
:attr:`_data`  The variable's :class:`~cf.Data` object.
:attr:`Units`  The variable's :class:`~cf.Units` object. Setting this
               changes the variable's `calendar` and `units` public
               attributes, and vice versa.
============== ======================================================


**Special private attributes:**

The following private attributes have special definitions and it is
recommended to set them accordingly.

============== ======================================================
Attribute      Description                         
============== ======================================================
:attr:`file`   DCH
:attr:`ncvar`  The variable's netCDF variable name.
============== ======================================================


**Reserved public attributes:**

The following public attributes have identical descriptions as in the
CF convention and may be set directly on the instance as well as with
the :meth:`setpub` or :meth:`pub` methods and retrieved directly from
the instance as well as with the :meth:`getpub` or :meth:`pub`
methods.

===================== ========================
Attribute             Description             
===================== ========================
:attr:`calendar`      As in the CF convention.
:attr:`_FillValue`    As in the CF convention.
:attr:`long_name`     As in the CF convention.
:attr:`missing_value` As in the CF convention.
:attr:`standard_name` As in the CF convention.
:attr:`units`         As in the CF convention.
===================== ========================


**Methods:**

A variable has the following methods.

================================ ================================================
Method                           Description
================================ ================================================
:meth:`~cf.Variable.copy`        Create a deep copy of the variable.

:meth:`~cf.Variable.dump`        Return a string containing a full description of
                                 the variable.

:meth:`~cf.Variable.equals`      Determine whether two instances are congruent
                                 with each other.

:meth:`~cf.Variable.match`       Determine whether a variable matches phenomena
                                 criteria.

:meth:`~cf.Variable.first_datum` Return the first element of the data without
                                 replacing a file pointer with a numpy array.

:meth:`~cf.Variable.last_datum`  Return the last element of the data without
                                 replacing a file pointer with a numpy array.

:meth:`~cf.Variable.name`        Return the variable's name.

:meth:`~cf.Variable.type`        The type of the data stored in the variable's
                                 :class:`~cf.Data` object.

:meth:`~cf.Variable.pubattrs`    Return a dictionary of the variable's public
                                 attributes.

:meth:`~cf.Variable.privattrs`   Return a dictionary of the variable's private
                                 attributes.

:meth:`~cf.Variable.istime`
:meth:`~cf.Variable.ematch`
:meth:`~cf.Variable.extract`
:meth:`~cf.Variable.eextract`
:meth:`~cf.Variable.priv`        Set or get a private attribute.        
:meth:`~cf.Variable.setpriv`     Set a private attribute.            
:meth:`~cf.Variable.getpriv`     Get a private attribute.            
:meth:`~cf.Variable.delpriv`     Delete a private attribute.         
:meth:`~cf.Variable.pub`         Set or get a public attribute.
:meth:`~cf.Variable.setpub`      Set a public attribute.
:meth:`~cf.Variable.getpub`      Get a public attribute.
:meth:`~cf.Variable.delpub`      Delete a public attribute.         
================================ ================================================


**Methods and attributes defined here:**
'''

    _private_attrs = ('_private_attrs',
                      '_data_insert'
                      '_concatenate_data',
                      '_getter',
                      '_setter',
                      '_deleter',
                      '_match',
                      '_to_Values',
                      'attributes',
                      'privattrs',
                      'first_datum',
                      'last_datum',
                      'type',
                      'copy',
                      'istime',
                      'dump',
                      'equals',
                      'match',
                      'ematch',
                      'extract',
                      'eextract',
                      'setpriv',
                      'getpriv',
                      'delpriv',
                      'pub',
                      'setattr',
                      'getattr',
                      'delattr',
                      'hasattr',
                      'name',
                      'override_Units',
                      )
    
    _reserved_public_attrs = set(('units',
                                  'calendar',
                                  '_FillValue', 
                                  ))

    _special_attributes = set(('units',
                               'calendar',
                               '_FillValue', 
                               ))

#    _special_attributes = set(('_data', 'Units', '_FillValue'))

    def __init__(self, attributes={}):
        '''

Set the public and private attributes from the arguments.
'''
        # Initialize the _private dictionary with an empty Units
        # object
        self._private = {'special_attributes': {'Units': Units()},
                         'simple_attributes' : {},
                         }

        for attr, value in attributes.iteritems():
            self.setattr(attr, copy.deepcopy(value))

#        for attr, value in public_attributes.iteritems():
#            self.setpub(attr, copy.deepcopy(value))        
#
#        for attr, value in private_attributes.iteritems():
#            self.setpriv(attr, copy.deepcopy(value))        
    #--- End: def

    def __setattr__(self, attr, value):
        '''
x.__setattr__(attr, value) <==> x.attr=value
'''
        if attr in self._private_attrs:
            raise AttributeError, "Can't set reserved attribute '%s'" % attr

        super(Variable, self).__setattr__(attr, value) 
    #--- End: def

    def __delattr__(self, attr):
        '''
x.__delattr__(attr) <==> del x.attr
'''
        if attr in self._private_attrs:
            raise AttributeError, "Can't delete reserved attribute '%s'" % attr

        super(Variable, self).__delattr__(attr) 
    #--- End: def

    def __getitem__(self, index):
        '''
x.__getitem__(index) <==> x[index]
'''

        if index == slice(None):
            return self

        try:
            if (index.start not in (0, -1) or
                index.stop != 1 or
                (index.step is not None or index.step < 1)):
                raise IndexError, "Bad '%s' index: %s" % \
                    (self.__class__.__name__, index)
        except AttributeError:
            if index not in (0, -1):
                raise IndexError, "%s index out of range: %s" % \
                    (self.__class__.__name__, index)
                
        return self
    #--- End: def

    def __setitem__(self, index, value):
        '''
x.__setitem__(index, value) <==> x[index]=value
'''
        pass # for now! DCH
    #--- End: def

    def __len__(self):
        '''
x.__len__() <==> len(x)
'''
        return 1
    #--- End: def

    def __deepcopy__(self, memo):
        '''

Used if copy.deepcopy is called on the variable.

.. seealso:: :meth:`~cf.Variable.copy`.
''' 
        return self.copy()
    #--- End: def

#    def _value_units(self, operator, other):
#        '''
#
#Return an object for the right hand side of a variable operation which
#has been conformed to appropiate units; and set the correct units on
#self.
#'''
##        try:
##            value = other.varraya
##        except AttributeError:
##            try:
##                value = other.Value
##            except AttributeError:
##                # other is not a Variable nor a ValueUnit
##                return other
##        #--- End: try
#        
#        try:
#            value = other.varray
#        except AttributeError:
#            # other is not a Variable nor a ValueUnit
#            return other
#        
#        v = ValueUnits(value, cfunits=self.Units)
#
#        # Still here? Then use the appropiate ValueUnits._???_work
#        # method to conform 'value' and get the correct units for self
#        self_vu = ValueUnits(cfunits=self.Units)
#
#        method  = '_%(operator)s_work' % locals()
#
#        value   = getattr(ValueUnits, method)(self_vu, value, other.Units)
#
#        self.override_Units(self_vu.Units)
#
#        # Return the (posibbly conformed) value
#        return value
#    #--- End: def

    # ----------------------------------------------------------------
    # Binary arithmetic operators and augmented arithmetic assignments
    # ---------------------------------------------------------------- 
    def _to_Values(self, other):
        '''
'''
        x = Value(value=self.varray, cfunits=self.Units)

        if isinstance(other, self.__class__):
            y = Value(value=other.varray, cfunits=other.Units)
            return x, y

        elif isinstance(other, Value):
            if isinstance(other.value, (float, int, long)):
                return x, other
            raise TypeError, "Can't combine '%s' with '%s' array" % \
                (self.__class__.__name__, other.__class__.__name__)

        elif isinstance(other, (float, int, long)):
            return x, Value(value=other, cfunits=self.Units)

        # Still here? 
        raise TypeError, "Can't combine '%s' with '%s'" % \
            (self.__class__.__name__, other.__class__.__name__)
    #--- End: def

    def __add__(self, other):
        '''
x.__add__(y) <==> x+y
'''
        x, y = self._to_Values(other)

        z = x + y

        new                = self.copy(data=False)
        new._data._list[0] = z.varray
        new.override_Units(z.Units)
        return new
    #--- End: def

    def __iadd__(self, other):
        '''
x.__iadd__(y) <==> x+=y
'''
        x, y = self._to_Values(other)
        x += y

        self.override_Units(x.Units)
        return self
    #--- End: def

    def __radd__(self, other):
        '''
x.__radd__(y) <==> y+x
'''
        return self + other
    #--- End: def

    def __sub__(self, other):
        '''
x.__sub__(y) <==> x-y
'''
        x, y = self._to_Values(other)

        z = x - y

        new                = self.copy(data=False)
        new._data._list[0] = z.varray
        new.override_Units(z.Units)
        return new
    #--- End: def

    def __isub__(self, other):
        '''
x.__isub__(y) <==> x-=y
'''
        x, y = self._to_Values(other)
        
        x -= y

        self.override_Units(x.Units)
        return self
    #--- End: def

    def __rsub__(self, other):
        '''
x.__rsub__(y) <==> y-x
'''
        x, y = self._to_Values(other)

        z = x.__rsub__(y)

        new                = self.copy(data=False)
        new._data._list[0] = z.varray
        new.override_Units(z.Units)
        return new
    #--- End: def

    def __mul__(self, other):
        '''
x.__mul__(y) <==> x*y
'''
        x, y = self._to_Values(other)

        z = x * y

        new                = self.copy(data=False)
        new._data._list[0] = z.varray
        new.override_Units(z.Units)
        return new
    #--- End: def

    def __imul__(self, other):
        '''
x.__imul__(y) <==> x*=y
'''  
        x, y = self._to_Values(other)
        
        x *= y

        self.override_Units(x.Units)
        return self
    #--- End: def
    
    def __rmul__(self, other):
        '''
x.__rmul__(y) <==> y*x
'''  
        return self * other
    #--- End: def
    
    def __div__(self, other):
        '''
x.__div__(y) <==> x/y
'''
        x, y = self._to_Values(other)

        z = x / y

        new                = self.copy(data=False)
        new._data._list[0] = z.varray
        new.override_Units(z.Units)
        return new
    #--- End: def

    def __idiv__(self, other):
        '''
x.__idiv__(y) <==> x/=y
'''          
        x, y = self._to_Values(other)
        
        x /= y

        self.override_Units(x.Units)
        return self
    #--- End: def
    
    def __rdiv__(self, other):
        '''
x.__rdiv__(y) <==> y/x
'''              
        x, y = self._to_Values(other)

        z = x.__rdiv__(y)

        new                = self.copy(data=False)
        new._data._list[0] = z.varray
        new.override_Units(z.Units)
        return new
    #--- End: def

    def __floordiv__(self, other):
        '''
x.__floordiv__(y) <==> x//y
'''
        x, y = self._to_Values(other)

        z = x // y

        new                = self.copy(data=False)
        new._data._list[0] = z.varray
        new.override_Units(z.Units)
        return new
    #--- End: def

    def __ifloordiv__(self, other):
        '''
x.__ifloordiv__(y) <==> x//=y
'''
        x, y = self._to_Values(other)
        
        x //= y

        self.override_Units(x.Units)
        return self
    #--- End: def
    
    def __rfloordiv__(self, other):
        '''
x.__rfloordiv__(y) <==> y//x
'''       
        x, y = self._to_Values(other)

        z = x.__rfloordiv__(y)

        new                = self.copy(data=False)
        new._data._list[0] = z.varray
        new.override_Units(z.Units)
        return new
    #--- End: def
    
    def __truediv__(self, other):
        '''
x.__truediv__(y) <==> x/y
'''
        x, y = self._to_Values(other)

        z = x.__truediv__(y)

        new                = self.copy(data=False)
        new._data._list[0] = z.varray
        new.override_Units(z.Units)
        return new
    #--- End: def

    def __itruediv__(self, other):
        '''
x.__itruediv__(y) <==> x/=y
'''  
        x, y = self._to_Values(other)
        
        x.__itruediv__(y)

        self.override_Units(x.Units)
        return self
    #--- End: def
    
    def __rtruediv__(self, other):
        '''
x.__rtruediv__(y) <==> y/x
'''
        x, y = self._to_Values(other)

        z = x.__rtruediv__(y)

        new                = self.copy(data=False)
        new._data._list[0] = z.varray
        new.override_Units(z.Units)
        return new
    #--- End: def
    
    def __pow__(self, other, modulo=None):
        '''
x.__pow__(y) <==> x**y
    '''  
        if modulo is not None:
            raise NotImplementedError, "3-argument power not supported for '%s'" % \
                self.__class__.__name__

        x, y = self._to_Values(other)

        z = x ** y

        new                = self.copy(data=False)
        new._data._list[0] = z.varray
        new.override_Units(z.Units)
        return new
    #--- End: def

    def __ipow__(self, other, modulo=None):
        '''
x.__pow__(y) <==> x**=y
    '''  
        if modulo is not None:
            raise NotImplementedError, "3-argument power not supported for '%s'" % \
                self.__class__.__name__

        x, y = self._to_Values(other)
        
        x **= y

        self.override_Units(x.Units)
        return self
    #--- End: def
    
    def __rpow__(self, other):
        return NotImplemented
    #--- End: def

    def __mod__(self, other):
        '''
x.__mod__(y) <==> x%y      
    '''
        x, y = self._to_Values(other)
        z    = x.__mod__(y)
        
        new                = self.copy(data=False)
        new.override_Units(z.Units)
        new._data._list[0] = z.varray

        return new
    #--- End: def

    def __imod__(self, other):
        '''
x.__imod__(y) <==> x%=y
'''        
        x, y = self._to_Values(other)
        x.__imod__(y)
        self.override_Units(x.Units)

        return self
    #--- End: def

    def __and__(self, other):
        '''
x.__and__(y) <==> x&y
    '''
        x, y = self._to_Values(other)
        z    = x & y
        
        new                = self.copy(data=False)
        new.override_Units(z.Units)
        new._data._list[0] = z.varray

        return new
    #--- End: def

    def __iand__(self, other):
        '''
x.__iand__(y) <==> x&=y
    '''
        x, y = self._to_Values(other)
        x &= y

        self.override_Units(x.Units)
        return self
    #--- End: def

    def __or__(self, other):
        '''
x.__or__(y) <==> x|y
    '''
        x, y = self._to_Values(other)
        z    = x | y
        
        new                = self.copy(data=False)
        new.override_Units(z.Units)
        new._data._list[0] = z.varray

        return new
    #--- End: def

    def __ior__(self, other):
        '''
x.__ior__(y) <==> x|=y
    '''
        x, y = self._to_Values(other)
        x |= y

        self.override_Units(x.Units)
        return self
    #--- End: def

    def __xor__(self, other):
        '''
x.__xor__(y) <==> x^y
    '''
        x, y = self._to_Values(other)
        z    = x ^ y
        
        new                = self.copy(data=False)
        new.override_Units(z.Units)
        new._data._list[0] = z.varray

        return new
    #--- End: def

    def __ixor__(self, other):
        '''
x.__and__(y) <==> x^=y
    '''
        x, y = self._to_Values(other)
        x ^= y

        self.override_Units(x.Units)
        return self
    #--- End: def

    # ----------------------------------------------------------------
    # Unary arithmetic operations
    # ----------------------------------------------------------------
    def __neg__(self):
        '''
x.__neg__() <==> -x
    '''
        new                = self.copy(data=False)
        new._data._list[0] = -self.varray
        return new
    #--- End: def

    def __pos__(self):
        '''
x.__pos__() <==> +x
    '''
        return self.copy()
    #--- End: def

    def __abs__(self):
        '''
x.__abs__() <==> abs(x)
    '''
        new                = self.copy(data=False)
        new._data._list[0] = numpy.abs(self.varray)
        return new   
    #--- End: def

    def __invert__(self):
        '''
x.__invert__() <==> ~x
    '''
        new                = self.copy(data=False)
        new._data._list[0] = ~self.varray
        return new   
    #--- End: def

    
#    def __eq__(self, other):
#        '''
#x.__eq__(y) <==> x==y
#    '''            
#        return self.__data_equals__(other)
#    #--- End: def
#
#    def __data_equals__(self, other, rtol=None, atol=None):
#        '''
#
#As :meth:`~cf.Variable.__eq__` but allowing the numeric equality
#tolerances to be set explicitly.
#
#    '''
#        if rtol is None:
#            rtol = RTOL()
#        if atol is None:
#            atol = ATOL()
#                
#        try:
#            other._data
#
#        except AttributeError:
#            # -------------------------------
#            # `other` is a scalar or an array
#            # -------------------------------
#            try:
#                # Numeric values
#                return numpy.less_equal(numpy.absolute(self.varray-other),
#                                        atol + rtol*numpy.absolute(other))
#            except TypeError:
#                # String values
#                return self.varray == other  
#        
#        else: 
#            # ---------------------
#            # `other` is a Variable
#            # ---------------------
#            # Check for file pointers being the same or for one array
#            # being a view of the other.
#            if self._data is other._data:
#                a = numpy.empty(self.shape, dtype=bool)
#                a.fill(True)
#                return a
#
#            try:
#                self._data.base
#                other._data.base
#            except AttributeError:
#                pass
#            else:
#                if ((self._data.base is not None and self._data.base is other._data) or
#                    (other._data.base is not None and other._data.base is self._data[0])):
#                    a = numpy.empty(self.shape, dtype=bool)
#                    a.fill(True)
#                    return a
#
#            x = self.varray
#            y = other.varray
#            try:
#                # Numeric values
#                return numpy.less_equal(numpy.absolute(x - y),
#                                        atol + rtol*numpy.absolute(y))
#            except TypeError:
#                # String values
#                return x == y
#    #--- End: def


    def __eq__(self, other):
        '''
x.__eq__(y) <==> x==y
'''
        x, y = self._to_Values(other)
        return x == y
    #--- End: def

    def __ne__(self, other):
        '''
x.__ne__(y) <==> x!=y
'''
        x, y = self._to_Values(other)
        return x != y
    #--- End: def

    def __lt__(self, other):
        '''
x.__lt__(y) <==> x<y
'''
        x, y = self._to_Values(other)
        return x < y            
    #--- End: def

    def __le__(self, other):
        '''
x.__le__(y) <==> x<=y
'''
        x, y = self._to_Values(other)
        return x <= y
    #--- End: def

    def __ge__(self, other):
        '''
x.__ge__(y) <==> x>=y
'''
        x, y = self._to_Values(other)
        return x >= y
    #--- End: def

    def __gt__(self, other):
        '''
x.__gt__(y) <==> x>y
'''
        x, y = self._to_Values(other)
        return x > y
    #--- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)
'''   
        try:
            name = self.name(default='', long_name=True, ncvar=True)
        except TypeError:
            # DCH This is so, very bad. The whol cfatts things need rethinking
            name = self.name
        try:
            shape = str(self.shape).replace(',)', ')', 1)
            return '<CF %s: %s%s>' % (self.__class__.__name__, name, shape)
        except AttributeError:
            return '<CF %s: %s>' % (self.__class__.__name__, name)
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)
'''   
        return self.__repr__()
    #--- End: def

    def _data_insert(self, indices, other,
                     self_dimensions, other_dimensions, 
                     match_for_dim1):
        '''
'''
        # ------------------------------------------------------------
        # Create a mapping between order names in other and those in
        # self. E.g. if self and other have dimension standard names
        # repectively:
        #
        #     ['time','latitude','longitude',-1]
        #     ['longitude','time',-1]
        #     
        # and
        #     
        #     self._data.order  = ['dim0','dim1','dim2','bounds']
        #     other._data.order = ['dim8','dim6','bounds']
        #     
        # then
        #    
        #     order_map['dim8']   = 'dim2'
        #     order_map['dim6']   = 'dim0'
        #     order_map['bounds'] = 'bounds'
        # ------------------------------------------------------------
#        print '\ndch sn=', self.pub('standard_name'), self.shape, other.shape
#        new_order  = []
#        for dim1 in other_dimensions:
#            try:
#                new_order.append(self_dimensions.index(match_for_dim1[dim1]))
#            except ValueError:
#                pass     
        #--- End: for
#        print 'dch self_dimensions =', self_dimensions, self._data.order 
#        print 'dch other_dimensions=', other_dimensions, other._data.order
#        print 'dch match_for_dim1=',match_for_dim1
#        print 'dch new_order=', new_order    
        order_map = {}

        for x in other._data.order:
            order_map[x] = match_for_dim1[x]

#        for i, n in enumerate(new_order):
#            try:
#                order_map[other._data.order[i]] = self._data.order[n]
#            except IndexError:
#                pass
            
        adim  = self._data.adim
        shape = list(self._data.shape)

        # If no indices, conctenate
        if not indices:
            len_self = len(self)
            indices = range(len_self, len_self+len(other))

        for i, n in enumerate(indices):

            # Insert the data and its metadata
            self._data.insert(n, other._data[i])

            # Change the dimension names of the newly inserted data to
            # conform with 'self._data'
            self._data.elements[n].order = [order_map[d]
                                            for d in self._data.elements[n].order]

            # update self.shape
            try:
                j = self._data.elements[n].order.index(self._data.order[adim])
            except ValueError: 
                shape[adim] += 1
            else:
                shape[adim] += self._data.elements[n].shape[j]
        #--- End: for

        self._data.shape = tuple(shape)
    #--- End: def

    def _concatenate_data(self):
        '''

'''
        try:
            data = self._data
        except AttributeError:
            pass
        else:
            data[:] = data.concatenate()
    #--- End:

    def _get_special_attr(self, attr):
        '''
        '''
        try:
            return self._private['special_attributes'][attr]
        except KeyError:
            raise AttributeError, "'%s' object doesn't have attribute '%s'" % \
                (self.__class__.__name__, attr)          
    #--- End: def

    def _set_special_attr(self, attr, value):
        '''
'''
        self._private['special_attributes'][attr] = value
    #--- End: def

    def _del_special_attr(self, attr):
        try:
            del self._private['special_attributes'][attr]
        except KeyError:
            raise AttributeError, "'%s' object doesn't have attribute '%s'" % \
                (self.__class__.__name__, attr)
    #--- End: def

    def _getter(self, attr):
        '''

Private function for getting attributes from the _private dictionary.
'''

        try:
            return self._private[attr]
        except (KeyError, AttributeError):
            raise AttributeError, "Can't get '%s' attribute '%s'" % \
                (self.__class__.__name__, attr)       
    #--- End: def

    def _setter(self, attr, value):
        '''

Private function for setting attributes in the _private dictionary.
'''
        self._private[attr] = value
    #--- End: def

    def _deleter(self, attr):
        '''

Private function for deleting attributes from the _private dictionary.
'''
        try:
            del self._private[attr]
        except KeyError:
            raise AttributeError, "Can't delete '%s' attribute '%s'" % \
                (self.__class__.__name__, attr)       
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: _data
    # ---------------------------------------------------------------- 
    @property
    def _data(self):        
        '''

The variable's Data object.
'''
        try:
            return self._private['_data']
        except KeyError:
            raise AttributeError, "'%s' object doesn't have attribute '_data'" % \
                self.__class__.__name__
#        return self._get_special_attr('_data')
    #--- End: def
    @_data.setter
    def _data(self, value):
        '''
'''
#        self._set_special_attr('_data', value)
        self._private['_data'] = value  
        for attr in ('Units', '_FillValue'):
            try:
                del self._private['special_attributes'][attr]
            except KeyError:
                pass
        #--- End: for
    #--- End: def

    @_data.deleter
    def _data(self):
        for attr in ('Units', '_FillValue'):
            try:
                self._private['special_attributes'][attr] = getattr(self._data, attr)
            except AttributeError:
                pass
        #--- End: for
        try:
            del self._private['_data']
        except KeyError:
            raise AttributeError, "'%s' object doesn't have attribute '_data'" % \
                self.__class__.__name__
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: Units
    # ----------------------------------------------------------------
    @property
    def Units(self):
        '''

The variable's `Units` object.
'''
        try:
            return self._data.Units
        except AttributeError:
            return self._get_special_attr('Units')
#    try:
#            return self._data.Units
#        except AttributeError:
#            try:
#                return self._private['Units']
#            except (KeyError, AttributeError):
#                raise AttributeError, "Can't get '%s' attribute 'Units'" % \
#                    self.__class__.__name__
#        #--- End: try
#    #--- End: def

    @Units.setter
    def Units(self, value):
        '''
Set the field's Units object.
'''
        if not isinstance(value, Units):
            raise TypeError, "'%s' attribute 'Units' must be a 'Units' object (got '%s')" % \
                (self.__class__.__name__, value.__class__.__name__)
        
        try:
            self._data.Units = value
        except AttributeError:
            self._set_special_attr('Units', value)
     #--- End: def

    @Units.deleter
    def Units(self):
        '''
Delete the field's Units object.
'''
        raise AttributeError, "Can't delete '%s' attribute 'Units'" % \
            self.__class__.__name__
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: _FillValue
    # ----------------------------------------------------------------
    @property
    def _FillValue(self):
        '''

The variable's CF `_FillValue` public attribute.

Assigning a value to this attribute also assigns the same value to the
`missing_value` public attribute, if the latter has previously been
defined.
'''
        try:
            return self._data._FillValue
        except AttributeError:
            return self._get_special_attr('_FillValue')
    #--- End: def

    @_FillValue.setter
    def _FillValue(self, value):
        '''
Set the field's _FillValue.
'''
        try:
            self._data._FillValue = value
        except AttributeError:
            self._set_special_attr('_FillValue', value)
    #--- End: def

    @_FillValue.deleter
    def _FillValue(self):
        '''
Delete the field's _FillValue.
'''
        try:
            del self._data._FillValue
        except AttributeError:
            self._del_special_attr('_FillValue')
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: units
    # ----------------------------------------------------------------
    # DCH possible inconsistency when setting self.Units.units
    @property
    def units(self):
        '''

The CF 'units' attribute.
'''        
        try:
            return self.Units.units
        except AttributeError:            
            raise AttributeError, "Can't get '%s' attribute 'units'" % \
                self.__class__.__name__
    #--- End: def

    @units.setter
    def units(self, value):
        self.Units.units = value
     #--- End: def

    @units.deleter
    def units(self):
        try:
            del self.Units.units
        except AttributeError:
            raise AttributeError, "Can't delete '%s' attribute 'units'" % \
                self.__class__.__name__       
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: calendar
    # ----------------------------------------------------------------
    @property
    def calendar(self):
        '''

The CF 'calendar' attribute.
'''
        try:
            return self.Units.calendar
        except AttributeError:            
            raise AttributeError, "Can't get '%s' attribute 'calendar'" % \
                self.__class__.__name__
    #--- End: def

    @calendar.setter
    def calendar(self, value):       
        self.Units.calendar = value

    @calendar.deleter
    def calendar(self):
        try:
            del self.Units.calendar
        except AttributeError:
            raise AttributeError, "Can't delete '%s' attribute 'calendar'" % \
                self.__class__.__name__       
    #--- End: def


    # ----------------------------------------------------------------
    # Property attribute: ncvar
    # ----------------------------------------------------------------
    @property
    def ncvar(self):        return self._getter('ncvar')
    @ncvar.setter
    def ncvar(self, value): self._setter('ncvar', value)
    @ncvar.deleter
    def ncvar(self):        self._deleter('ncvar')

    # ----------------------------------------------------------------
    # Property attribute: standard_name
    # ----------------------------------------------------------------
    @property
    def standard_name(self):
        '''

The variable's CF `standard_name` public attribute. 
'''
        return self.getattr('standard_name')
    #--- End: def
    @standard_name.setter
    def standard_name(self, value): self.setattr('standard_name', value)
    @standard_name.deleter
    def standard_name(self):        self.delattr('standard_name')

    # ----------------------------------------------------------------
    # Property attribute: long_name
    # ----------------------------------------------------------------
    @property
    def long_name(self):
        '''

The variable's CF `long_name` public attribute. 
'''
        return self.getattr('long_name')
    #--- End: def
    @long_name.setter
    def long_name(self, value): self.setattr('long_name', value)
    @long_name.deleter
    def long_name(self):        self.delattr('long_name')

    # ----------------------------------------------------------------
    # Property attribute: valid_min
    # ----------------------------------------------------------------
    @property
    def valid_min(self):
        '''
The CF 'valid_min' attribute. 
'''
        return self.getattr('valid_min')
    #--- End: def
    @valid_min.setter
    def valid_min(self, value): self.setattr('valid_min', value)
    @valid_min.deleter
    def valid_min(self):        self.delattr('valid_min')

    # ----------------------------------------------------------------
    # Property attribute: valid_max
    # ----------------------------------------------------------------
    @property
    def valid_max(self):
        '''
The CF 'valid_max' attribute. 
'''
        return self.getattr('valid_max')
    #--- End: def
    @valid_max.setter
    def valid_max(self, value): self.setattr('valid_max', value)
    @valid_max.deleter
    def valid_max(self):        self.delattr('valid_max')

    # ----------------------------------------------------------------
    # Property attribute: valid_range
    # ----------------------------------------------------------------
    @property
    def valid_range(self):
        '''
The CF 'valid_range' attribute. 
'''
        return self.getattr('valid_range')
    #--- End: def
    @valid_range.setter
    def valid_range(self, value): self.setattr('valid_range', value)
    @valid_range.deleter
    def valid_range(self):        self.delattr('valid_range')

    # ----------------------------------------------------------------
    # Property attribute: missing_value
    # ----------------------------------------------------------------
    @property
    def missing_value(self):
        '''

The CF 'missing_value' attribute.

Assigning a value to this attribute also assigns the same value to the
`_FillValue` attribute.
'''
        return self.getattr('missing_value')
    #--- End: def

    @missing_value.setter
    def missing_value(self, value):
        '''
Set the missing_value attribute
'''
        self.setattr('missing_value', value)
        self._FillValue = value
    #--- End: def
    @missing_value.deleter
    def missing_value(self):  
        '''
Delete the missing_value attribute
'''
        self.delattr('missing_value')
    #--- End: def

    # --------------------------------------------------------
    # Read-only property attributes
    # --------------------------------------------------------
    @property
    def privattrs(self):
        '''
        
Return a dictionary of the instance's private attributes.
        '''
        try:
            return self._private['privattrs'].copy()
        except (KeyError, AttributeError):
            return {}
    #--- End: def

#    @property
#    def pubattrs(self):
#        '''
#
#Return a dictionary of the instance's public attributes.
#'''
#        out = {}
#
##        if not test:
##            # If we're not going to test, include special cases which
##            # don't live in the self._private['pubattrs'] dictionary
##            for attr in ('units', 'calendar', '_FillValue'):
##                if hasattr(self, attr):
##                    out[attr] = getattr(self, attr)
##        #--- End: if
#        
#        try:
#            out.update(self._private['pubattrs'].copy())
#        except (KeyError, AttributeError):
#            pass
#
#        return out
#    #--- End: def

    @property
    def slice(self):
        '''

Slice the variable's data, returning a new variable.

**Examples:**

>>> v
<CF Variable: air_temperature(73, 96)>
>>> v.slice[0,0]
<CF Variable: air_temperature(1,1)>
'''
        return SliceVariable(self)
    #--- End: def

    @property
    def shape(self):
        '''

    Tuple of the data's dimension sizes.

    :returns: A tuple

    '''
        try:
            shape = self._data.shape
        except AttributeError:
            raise AttributeError, "Can't find 'shape' of '%s' instance" % \
                self.__class__.__name__

        return shape
    #--- End: def

    @property
    def ndim(self):
        '''

Number of data dimensions.

Equivalent to the number of elements in shape tuple.

:returns:     A non-negative integer
'''
        try:
            return self._data.ndim
        except AttributeError:
            raise AttributeError, "Can't find 'ndim' of '%s' instance" % \
                self.__class__.__name__
    #--- End: def

    @property
    def size(self):
        '''

    Number of elements in the data.

    Equivalent to the product of the data's dimension sizes.

    :returns:     A non-negative integer

    '''
        try:
            return self._data.size
        except AttributeError:  
            raise AttributeError, "Can't find 'size' of '%s' instance" % \
                self.__class__.__name__
    #--- End: def

    @property
    def dtype(self):
        '''

Data-type of the data's elements.

:returns:     A numpy dtype object.
'''
        try:
            return self._data.dtype
        except AttributeError:
            raise AttributeError, "Can't find 'dtype' of '%s' instance" % \
                self.__class__.__name__
    #--- End: def

    @property
    def array(self):
        '''

Create a numpy array deep copy of the data.

If the data was stored as a file pointer then it will be changed
in-place to be stored as a numpy array.

:returns:     A numpy array.

**Examples:**

>>> a = x.array
>>> type(a)
<type 'numpy.ndarray'>
>>> a = x.array[[0,1]]
>>> type(a)
<type 'numpy.ndarray'>
>>> a.shape
(2,)

'''
        if len(self._data) == 1:
            self._data.conform()
            a = self._data._list[0].copy()
        else:
            data = self._data.concatenate()
            a = data._list[0]

        try:
            a.set_fill_value(self._FillValue)
        except AttributeError: 
            pass

        return a
    #--- End: def

    @property
    def varray(self):
        '''
Create a numpy view of the data.

If the data was stored as a file pointer then it will be changed
in-place to be stored as a numpy array.

Note that making changes to elements of the returned view changes the
underlying data. Refer to :obj:`numpy.ndarray.view`.
    
:returns:     A numpy view.

**Examples:**

>>> a = x.varray
>>> print a
array([0, 1, 2, 3, 4])
>>> a[0] = 999
>>> print x.varray[0]
999
>>> a = 'a_different_object'
>>> print x.varray
array([999, 1, 2, 3, 4])
'''       
        self._concatenate_data()

        v = self._data._list[0].view()

        try:
            v.set_fill_value(self._FillValue)
        except AttributeError: 
            pass

        return v
    #--- End: def

#    @property
#    def tarray(self):
#        if not self.istime():
#            raise TypeError, "Can't create a 'Time' array from a non-time variable"
#            
#        try:
#            calendar = self.calendar
#        except AttributeError:
#            calendar = 'standard'
#
#        return cf.Time(self.varray, units=self.units, calendar=calendar)
#    #--- End: def
#
#    def set_time(self, cftime):
#        if not self.istime():
#            raise TypeError, "Can't set time for a non-time variable"
#
#        # Check that the Time array has the same size and shape as the
#        # Variable array
#
#        self.units    = cftime.units
#        self.calendar = cftime.calendar
#
#        self[...] = cftime.array
#    #--- End: def

    def istime(self):
        '''
'''
        try:
            if ' since ' in self.units:
                return True
            return False
        except AttributeError:
            return False
    #--- End: def

#    def change_calendar(self, calendar, fix=False):
#        t = self.tarray
#        t = t.change_calendar(calendar, fix=fix)
#        self.set_time(t)
#    #--- End: def
        
    @property
    def first_datum(self):
        ''' 

    Return the first element of the data without replacing a file
    pointer with a numpy array.
        
    :returns:     The scalar value of the first element of the data.

    '''
#        return numpy.asscalar(self.slice[(slice(0,1),) * self.ndim].varray)
        return self.slice[(slice(0,1),) * self.ndim].varray.item()
    #--- End: def
    
    @property
    def last_datum(self):
        ''' 

    Return the last element of the data without replacing a file
    pointer with a numpy array.
    
    :returns:     The scalar value of the last element of the data.

    '''
        return numpy.asscalar(self.slice[(slice(-1,None),) * self.ndim].varray)
    #--- End: def

    def type(self):
        '''

The types of the data stored in the variable's Data object.

:returns: A list of type objects for each of the data elements.

 **Examples:**

>>> v.type()
[<type 'netCDF4.Variable'>]
>>> v.slice[...]
>>> v.type()
[<type 'numpy.ndarray'>]
'''
        try:
            return map(type, self._data._list)
        except AttributeError:
            raise AttributeError, "Can't find types of '%s' data" % \
                self.__class__.__name__
    #--- End: def

    def copy(self, data=True):
        '''

Return a deep copy.

:keyword data: *Optional*. If False then assign the output variable's
               data elements to None, with the expectation that they
               will be replaced subsequently. The data object's
               metadata are deep copied in any case.

:type data: bool

:returns: A deep copy.
'''
        new = type(self)()

        try:
            new._data = self._data.copy(data=data)
        except AttributeError:
            pass

        new.Units = self.Units.copy()
        if hasattr(self, '_FillValue'):
            new._FillValue = self._FillValue

        # private
        attrs = set((self._private)) - set(('_data', 'Units', '_FillValue'))
        for attr in attrs:
            new._private[attr] = copy.deepcopy(self._private[attr])

        # Attributes
        attrs = set(self.__dict__) - set(('_private',))
        for attr in attrs:
            setattr(new, attr, copy.deepcopy(self.__dict__[attr]))

        return new
    #--- End: def

    def dump(self, id=None, omit=(), _include=()):
        '''

Return a string containing a full description of the variable.

:keyword id: *Optional*. Set the common prefix of component names. If
             None then defaults to the class name.

:type id: str

:keyword omit: *Optional*. Omit the given attributes from the
               description.

:type omit: sequence

:returns: A string containing the description of the variable.

.. seealso:: :func:`cf.dump`
'''
        if id is None:
            id = self.__class__.__name__
            
        string = []

        if hasattr(self, '_data'):
            string.append(self._data.dump(id=id))

        # Special attributes
        special = self._private['special_attributes'].copy()
        for attr in omit:
            if special.has_key(attr):
                special.pop(attr)
        for attr, value in special.iteritems():
            string.append('%s.%s = %s' % (id, attr, repr(value)))
        string.append('')

        # Simple attributes
        simple = self._simple_attributes()
        attrs  = sorted(set(simple) - set(omit))
        if attrs:
            for attr in attrs:
                name   = '%(id)s.%(attr)s = ' % locals()
                value  = repr(simple[attr])
                indent = ' ' * (len(name))
                if value.startswith("'") or value.startswith('"'):
                    indent = '%(indent)s ' % locals()
                string.append(textwrap.fill(name+value, 79,
                                            subsequent_indent=indent))
            #--- End: for
            string.append('')
        #--- End: if

        return '\n'.join(string)
    #--- End: def

    def equals(self, other, rtol=None, atol=None):
        '''

Return True if the variable is congruent to another variable in that

    1. They have identical data. Pointers to data on disk may be
       converted to numpy arrays in memory if required.

    2. They have identical reserved public attributes.

    3. They have identical public attributes.

Equality of numbers is to within a tolerance. Refer to :mod:`cf` for
details.

:argument other: The variable to compare against for equality.

:type other: object
             
:keyword atol: *Optional*. If None then use the default method for
               setting the absolute tolerance for numerical equality
               (refer to :mod:`cf` for details). If a number then set
               the absolute tolerance to this value for all such
               comparisons.

:type atol: None or float

:keyword rtol: *Optional*. If None then use the default method for
               setting the relative tolerance for numerical equality
               (refer to :mod:`cf` for details). If a number then set
               the relative tolerance to this value for all such
               comparisons.

:type rtol: None or float

:returns: True if the two objects are congruent, False otherwise.


**Examples:**

>>> y = x
>>> x.equals(y)
True
>>> y = x + 1
>>> x.equals(y)
False
>>> y -= 1
>>> x.equals(y)
True
>>> x.setattr('name', 'name0')
>>> y.setattr('name', 'name1')
>>> x.equals(y)
False
'''
        # Check each instance's id
        if self is other:
            return True

        # Check that each instance is the same type
        if type(self) != type(other):
             return False

        self_pub  = self.attributes
        other_pub = other.attributes
        if set(self_pub) != set(other_pub):
            return False

        if rtol is None:
            rtol = RTOL()
        if atol is None:
            atol = ATOL()                
            
        # Check that each instance has the same CF property values
        for attr, x in self_pub.iteritems():
            y = other_pub[attr]
            if not equals(x, y, rtol=rtol, atol=atol):
                return False

        # Check the special attributes which have 'equals' methods,
        # such as Units, etc.
        self_special  = self._private['special_attributes']
        other_special = other._private['special_attributes']
        if set(self_special) != set(other_special):
            return False
 
        for attr, x in self_special.iteritems():            
            y = other_special[attr]
            if not equals(x, y, rtol=rtol, atol=atol):
                return False

        # Check the data
        try:
            x = self._data
        except AttributeError:
            try:
                y = other._data
            except AttributeError:
                return False
        else:
            y = other._data
            if not x.equals(y, rtol=rtol, atol=atol):
                return False
        #--- End: try

#        # Check the private attributes which have 'equals' methods,
#        # such as _data, etc.
#        self_private  = self._private
#        other_private = other._private
#        for attr, x in self_private.iteritems():
#            if hasattr(x, 'equals'):
#                if not other_private.has_key(attr):
#                    return False
#                y = other_private[attr]            
#                if not x.equals(y, rtol=rtol, atol=atol):
#                    return False
#        #--- End: for

        return True
    #--- End: def

    def _match(self, exact, **kwargs):
        '''

See match and ematch for details.
'''        
        # Loop round criteria
        for name, value in kwargs.iteritems():

            if value is None:
                continue

            # ------------------------------------------------------
            # Find out how many phenomena match 'name'
            # ------------------------------------------------------
            count = 0
             
            # Check scalar or 1-d, size 1 coordinates, if this
            # variable has them
            try:
                x = self.coord(value, exact=exact, key=True)
            except AttributeError:
                pass
            else:
                try:
                    if x.size == 1 and x.ndim <= 1:
                        x = x.first_datum                
                        count += 1
                except AttributeError:
                    pass

            # Check simple attributes
            attr = None
            if exact:
                if name in self._simple_attributes() or hasattr(self, name):
                    attr = name
                    count += 1
            else:                    
                for pubattr in self._simple_attributes():
                    if re.match(name, pubattr):
                        attr = pubattr
                        count += 1
                if not count:
                    if hasattr(self, name):
                        attr = name
                        count += 1
            #--- End: if
 
            if count == 1:
                # Variable has exactly one phenomenon name matching 'name'
                if attr is not None:
                    try:
                        x = self.getattr(attr)
                    except AttributeError:
                        x = getattr(self, attr)
            elif count == 0:
                # Variable has no phenomenon name matching 'name'
                return False
            elif count > 1:                
                raise ValueError, "More than one phenomenon name matches '%(value)s'" % \
                    locals()
    
            # --------------------------------------------------------
            # See if the phenomenon conforms with the value
            # --------------------------------------------------------
            try:
                if not value.evaluate(x):
                    return False
                continue
            except AttributeError:
                pass

            if not isinstance(value, (list, tuple, numpy.ndarray)):
                values = (value,)
            else:
                values = value
            
            match = False
            for v in values:  
                try:
                    if re.search('^%s$' % v, x):
                        match = True
                        break
                except TypeError:
                    # v is not a string
                    if equals(x, v):
                        match = True
                        break
            #--- End: for

            if not match:
                return False
        #--- End: for

        # Still here? Then we have a match
        return True
    #--- End: def

    @property
    def mask(self):
        '''

'''
        self._concatenate_data()

        # Make sure that the data array is a masked array
        if not hasattr(self._data._list[0], 'mask'):
            return False

        # Still here?
        return self._data._list[0].mask
    #--- End: def

    @mask.setter
    def mask(self, m):
        '''

Set the mask for the data
'''
        # dch: this need to be pushed into the Data object
        self._concatenate_data()

        if not hasattr(self._data._list[0], 'mask'):
            # The data array is not masked, so turn it into a masked
            # array
            self._data._list[0] = numpy.ma.array(self._data._list[0], 
                                                 copy=False)
            if m is not False:
                self._data._list[0].mask = m
                self._data._list[0].shrink_mask()

            if hasattr(self, '_FillValue'):
                self._data._list[0].set_fill_value(self._FillValue)
        else:
            # The data array is already a masked array, so just update
            # the mask
            self._data._list[0].mask = m
            self._data._list[0].shrink_mask()
    #--- End: def

    @mask.deleter
    def mask(self):
        '''

Delete the mask for the data by setting it to False
'''
        # dch concat?
        self.mask = False
        self._data._list[0].shrink_mask()
    #--- End: def

    def match(self, **kwargs):      
        '''

Determine whether or not a variable matches conditions on its
phenomena, i.e. its public attributes and, if it has any, its scalar
or 1-d, size 1 coordinates. Unambiguous abbreviations of public
attribute or coordinate names are allowed.

The phenomena and their conditions are specified with `**kwargs`
parameters.

The variable matches the conditions if and only if it contains all of
the specified phenomena and they satisfy all of their given
criteria. A variable always matches no criteria.

:keyword **kwargs: The keyword arguments identify a public attribute's
                   name or a coordinate's standard name. Any
                   unambiguous abbreviation of the phenomenon's name
                   is acceptable.
          
                   A keyword's value may be one of the following:
                   
                   1. A number. A match requires that a numeric valued
                      phenomenon equals the number using the
                      numerically tolerant equality.
                   
                   2. A string. A match requires that a string-valued
                      phenomenon passes a regular expression match on
                      the string. The string may contain regular
                      expression special characters. To avoid
                      ambiguities, it is assumed that the regular
                      expression string matches the whole of the
                      phenomenon string, i.e. the regular expression
                      special characters ^ and $ are assumed if not
                      given.
                   
                   3. A sequence of numbers or strings. A match
                      requires that a numeric phenomenon equals,
                      within numerical tolerance, at least one of the
                      sequence's elements or a string-valued
                      phenomenon passes a regular expression match for
                      at least one string in the sequence (see 2.).
                   
                   4. A :class:`~cf.Comparison` object. A match
                      requires that the comparison for the phenomenon
                      evaluates to True.

:returns: True if the variable matches the criteria, False otherwise.
        
**Examples:**

>>> s
<CF Field: air_temperature(19, 30, 24)>
>>> s.pub('standard_name'), s.pub('units')
('air_temperature', 'K')
>>> s.extract(standard_name = 'air_temperature')
True
>>> s.extract(standard = '^air.*')
True
>>> s.extract(standard_name = lt('b'))
True
>>> s.extract(standard_name = outside('b', 'p'))
True
>>> s.extract(standard = ['.*temp.*', 'pressure'])
True
>>> s.extract(height = 2.0, units = 'K')
True
'''
        return self._match(False, **kwargs)
    #--- End: def

    def ematch(self, **kwargs):
        '''

Determine whether or not a variable matches conditions on its
phenomena, i.e. its public attributes and, if it has any, its scalar
or 1-d, size 1 coordinates. As :meth:`~cf.Variable.match`, but with
the restriction that only exact public attribute or coordinate names
are allowed.

:keyword **kwargs: The keyword arguments identify exact phenomenon
                   names, either a public attribute names or 1-d, size
                   1 coorinate standard names.

                   The keyword values are as for
                   :meth:`~cf.Variable.match`.

:returns: True if the variable matches the criteria, False otherwise.
        
**Examples:**

>>> s
<CF Field: air_temperature(19, 30, 24)>
>>> s.standard_name
'air_temperature'
>>> s.extract(standard_name = 'air_temperature')
True
>>> s.extract(standard_n = '^air.*')
False
'''
        return self._match(True, **kwargs)
    #--- End: def

    def extract(self, **kwargs):
        '''

Return the variable if and only if it matches conditions on its
phenomena, i.e. its public attributes and, if it has any, its scalar
or 1-d, size 1 coordinates. Unambiguous abbreviations of public
attribute or coordinate names are allowed.

Refer to :meth:`~cf.Variable.match` for further details.

keyword **kwargs: The keyword arguments identify a public attribute's
                  name or a coordinate's standard name. Any
                  unambiguous abbreviation of the phenomenon's name is
                  acceptable.

                  The keyword values are as for
                  :meth:`~cf.Variable.match`.

:returns: The variable if the it matches the criteria, otherwise a
          ValueError is raised.
'''
        if self.match(**kwargs):
            return self
        else:
            raise ValueError, "'%s' does not match criteria" % \
                self.__class__.__name__
    #--- End: def

    def eextract(self, **kwargs):
        '''

Return the variable if and only if it matches conditions on its
phenomena, i.e. its public attributes and, if it has any, its scalar
or 1-d, size 1 coordinates. As :meth:`~cf.Variable.extract`, but with
the restriction that only exact public attribute or coordinate names
are allowed.

keyword **kwargs: The keyword arguments identify exact phenomenon
                  names, either a public attribute names or 1-d, size
                  1 coorinate standard names.

                  The keyword values are as for
                  :meth:`~cf.Variable.match`.

:returns: The variable if the it matches the criteria, otherwise a
          ValueError is raised.
'''
        if self.ematch(**kwargs):
            return self
        else:
            raise ValueError, "'%s' does not match criteria" % \
                self.__class__.__name__
    #--- End: def





    def priv(self, *args, **kwargs):
        '''
        
A method which may be used in place of :meth:`~cf.Variable.getpriv` or
:meth:`~cf.Variable.setpriv`.

If called with positional parameters, then they are interpreted as for
:meth:`~cf.Variable.getpriv` and a public attribute (or default) value
is returned.

If called with a keyword parameter, then the keyword and its value act
as the two positional parameters of :meth:`~cf.Variable.setpriv` and a
public attribute is set.
'''
        if args:
            if kwargs:
                raise TypeError, "sdf 0u812 3 dch"
            return self.getpriv(*args)

        for attr, value in kwargs.iteritems():
            self.setpriv(attr, value)
    #--- End: def

    def setpriv(self, attr, value):
        '''
Set a private attribute.

:keyword attr: The name of the private attribute to be set.

:type attr: str

:keyword value: *Optional*. The value for the private attribute.
                 
:type value: object

:returns: None

'''
        try:
            self._private['privattrs'][attr] = value
        except KeyError:
            self._private['privattrs'] = {attr: value}       
    #--- End: def

    def getpriv(self, attr, *default):
        '''
Return a private attribute.

:keyword attr: The name of the private attribute to be returned.

:type attr: str

:keyword default: *Optional*. Return `default` if the requested private
                   attribute has not been set
                 
:type default: object

:returns: The value of a private attribute.
'''
        try:
            return self._private['privattrs'][attr]
        except KeyError:
            if default:
                return default[0]

        raise AttributeError, "'%s' object doesn't have private attribute '%s'" % \
            (self.__class__.__name__, attr)          
    #--- End: def

    def delpriv(self, attr):
        '''
Delete a private attribute.

:keyword attr: The name of the private attribute to be deleted.

:type attr: str

:returns: None
'''
        try:
            del self._private['privattrs'][attr]
        except KeyError:
            raise AttributeError, "'%s' object has no private attribute '%s'" % \
                (self.__class__.__name__, attr)
    #--- End: def

#    def pub(self, *args, **kwargs):
#        '''
#        
#A method which may be used in place of :meth:`~cf.Variable.getpub` or
#:meth:`~cf.Variable.setpub`.
#
#If called with positional parameters, then they are interpreted as for
#:meth:`~cf.Variable.getpub` and a public attribute (or default) value
#is returned.
#
#If called with a keyword parameter, then the keyword and its value act
#as the two positional parameters of :meth:`~cf.Variable.setpub` and a
#public attribute is set.
#
#
#**Examples:**
#
#>>> v.pub(units='K')
#>>> v.pub('units')
#'K'
#>>> v.pub('UNITS', 'default_value')
#'default_value'
#'''
#        if args:
#            if kwargs:
#                raise TypeError, "sdf 0u812 dch"
#            return self.getpub(*args)
#
#        for attr, value in kwargs.iteritems():
#            self.setpub(attr, value)
#    #--- End: def

    @property
    def attributes(self):
        '''
'''
        out = {}

        # Start with special attributes
        for attr in self._special_attributes:
            if hasattr(self, attr):
                out[attr] = getattr(self, attr)

        # Add simple attributes
        out.update(self._simple_attributes())

        return out
    #--- End: def

    def _simple_attributes(self):
        '''
'''
        return self._private['simple_attributes'].copy()
    #--- End: def

    def setattr(self, attr, value):
        '''
'''
        # Set a special attribute
        if attr in self._special_attributes:
            setattr(self, attr, value)
            return None

        # Still here? Then set a simple attribute
        self._private['simple_attributes'][attr] = value
    #--- End: def

    def hasattr(self, attr):
        '''
'''
        # Has a special attribute?
        if attr in self._special_attributes:
            return hasattr(self, attr)

        # Still here? Then has a simple attribute?
        return self._private['simple_attributes'].has_key(attr)
    #--- End: def

    def getattr(self, attr, *default):
        '''
'''
        # Get a special attribute
        if attr in self._special_attributes:
            return getattr(self, attr, *default)

        # Still here? Then get a simple attribute
        try:
            return self._private['simple_attributes'][attr] 
        except KeyError:
            if default:
                return default[0]

        # Still here?
        raise AttributeError, "'%s' object doesn't have CF attribute '%s'" % \
            (self.__class__.__name__, attr)          
    #--- End: def

    def delattr(self, attr):
        '''
'''
        # Delete a special attribute
        if attr in self._special_attributes:
            delattr(self, attr)
            return None

        # Still here? Then delete a simple attribute
        try:
            del self._private['simple_attributes'][attr] 
        except KeyError: 
            raise AttributeError, "'%s' object doesn't have CF attribute '%s'" % \
                (self.__class__.__name__, attr)
    #--- End: def

#    def setpub(self, attr, value):
#        '''
#Set a public attribute.
#
#:keyword attr: The name of the public attribute to be set.
#
#:type attr: str
#
#:keyword value: *Optional*. The value for the public attribute.
#                 
#:type value: object
#
#:returns: None
#'''
#        # Special cases which don't live in the
#        # self._private['pubattrs'] dictionary            
##        if attr in self._reserved_public_attrs:
#        if attr in self._special_attributes:
#            setattr(self, attr, value)
#            return None
#
##        if attr in _reserved_public_attrs:
##            raise AttributeError, "Can't set '%s' as a public attribute on '%s'" % \
##                 (attr, self.__class__.__name__) 
#
#        # Still here?
#        try:
#            self._private['pubattrs'][attr] = value
#        except KeyError:
#            self._private['pubattrs'] = {attr: value}
#    #--- End: def
#
#    def set(self, attr, value):
#        '''
#Set a public attribute.
#
#:keyword attr: The name of the public attribute to be set.
#
#:type attr: str
#
#:keyword value: *Optional*. The value for the public attribute.
#                 
#:type value: object
#
#:returns: None
#'''
#        # Special cases which don't live in the
#        # self._private['pubattrs'] dictionary            
#        if attr in self._reserved_public_attrs:
#            setattr(self, attr, value)
#            return None
#
##        if attr in _reserved_public_attrs:
##            raise AttributeError, "Can't set '%s' as a public attribute on '%s'" % \
##                 (attr, self.__class__.__name__) 
#
#        # Still here?
#        try:
#            self._private['pubattrs'][attr] = value
#        except KeyError:
#            self._private['pubattrs'] = {attr: value}
#    #--- End: def
#
#    def getpub(self, attr, *default):
#        '''
#
#Return a public attribute.
#
#:keyword attr: The name of the public attribute to be returned.
#
#:type attr: str
#
#:keyword default: *Optional*. Return `default` if the requested public
#                   attribute has not been set
#                 
#:type default: object
#
#:returns: The value of a public attribute.
#'''
#        # Special cases which don't live in the
#        # self._private['pubattrs'] dictionary
#        if attr in self._reserved_public_attrs:
#            return getattr(self, attr, *default)
##        
##        if attr in _reserved_public_attrs:
##            raise AttributeError, "Can't get '%s' as a public attribute from '%s'" % \
##                 (attr, self.__class__.__name__) 
#
#        try:
#            return self._private['pubattrs'][attr]
#        except (KeyError, AttributeError):
#            if default:
#                return default[0]
#
#        # Still here?
#        raise AttributeError, "'%s' object doesn't have public attribute '%s'" % \
#            (self.__class__.__name__, attr)          
#    #--- End: def
#
#    def delpub(self, attr):
#        '''
#Delete a public attribute.
#
#:keyword attr: The name of the public attribute to be deleted.
#
#:type attr: str
#
#:returns: None
#'''
#        # Special cases which don't live in the
#        # self._private['pubattrs'] dictionary
#        if attr in self._reserved_public_attrs:
#            delattr(self, attr)
#            return None
#
#        # Still here?
#        try:
#            del self._private['pubattrs'][attr]
#        except (KeyError, AttributeError):
#            raise AttributeError, "'%s' object doesn't have public attribute '%s'" % \
#                (self.__class__.__name__, attr)
#    #--- End: def

    def name(self, long_name=False, ncvar=False, default=None):
        ''' 

Return the `standard_name` attribute of the variable.

If there is no `standard_name` attribute then return one of the
`long_name` attribute, the `ncvar` attribute or the value of the
default parameter, depending on the values of the parameters.

:keyword long_name: *Optional*. If True, return the `long_name` if
                    `standard_name` does not exist.

:type long_name: bool

:keyword ncvar: *Optional* If True, return `ncvar` if neither
                `standard_name` not `long_name` have already been
                returned.

:type ncvar: bool

:keyword default: *Optional*. Return `default` if `standard_name`,
                  `long_name` nor `ncvar` have already been returned.

:type default: None or object


:returns: The name of the variable.

'''
        try:
            return self.standard_name
        except AttributeError:
            pass

        if long_name:
            try:
                return self.long_name
            except AttributeError:
                pass
            
        if ncvar:
            try:
                return self.ncvar
            except AttributeError: 
                pass            

        return default
    #--- End: def
        
    def override_Units(self, units):
        '''

Change the variable's units in place without changing the variable's
data.

:parameter units: The new units as a :class:`Units` object.

:type units: Units

:returns: None


**Examples:**

>>> v.Units
<CF Units: metre>
>>> v.varray
array([ 0.,  1.,  2.,  3.,  4.])

>>> v.override_Units(cf.Units('km'))
>>> v.Units
<CF Units: km>
>>> v.varray
array([ 0.,  1.,  2.,  3.,  4.])
'''
        self.Units = copy.deepcopy(units)
        try:
            for element in self._data.elements:
                element.Units = copy.deepcopy(units)
        except AttributeError:
            pass
    #--- End: def 

#--- End: class


# ====================================================================
#
# List of Variables object
#
# ====================================================================

class VariableList(CfList):
    '''

An ordered sequence of variables stored in a single, list-like
object. Refer to Variable Lists in the :mod:`cf` package for details.

**Methods and attributes defined here:**
'''

    _private_attrs = Variable._private_attrs[:]
        
    def __getattribute__(self, attr, *default):
        '''
x.__getattribute__(attr) <==> x.attr     
'''
        if (attr.startswith('__') or 
            attr in super(VariableList, self).__getattribute__('__dict__') or
            attr in dir(self)):
            # Return an attribute of the list-like object
            return super(VariableList, self).__getattribute__(attr, *default)

        # Return a built-in list of the attributes from each element
        # of the cf list
        return [getattr(v, attr, *default) 
                for v in super(VariableList, self).__getattribute__('_list')]
    #--- End: def

    def __setattr__(self, attr, value):
        '''
x.__setattr__(attr, value) <==> x.attr=value
'''
        if (attr.startswith('__') or 
            attr in super(VariableList, self).__getattribute__('__dict__') or
            attr in dir(self)):
            # Set an attribute on the list-like object
            super(VariableList, self).__setattr__(attr, value)
            return
        #--- End: if

        # Still here? Then do a built-in setattr on each of the list's
        # elements
        for v in self.__dict__['_list']:
            setattr(v, attr, value)
    #--- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)
'''
        return repr(self._list).replace('>, ', '>,\n ')
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)
'''
        return '\n\n'.join([v.__str__() for v in self._list])
    #--- End: def

    def __eq__(self, other):
        '''
x.__eq__(y) <==> x==y
'''
        return type(self)([v.__eq__(other) for v in self._list])
    #--- End: def

    def __ne__(self, other):
        '''
x.__ne__(y) <==> x!=y
'''
        return type(self)([v.__ne__(other) for v in self._list])
    #--- End: def

    def __neg__(self):
        '''
x.__neg__() <==> -x
'''
        return type(self)([v.__neg__() for v in self._list])
    #--- End: def

    def __pos__(self):
        '''
x.__pos__() <==> +x
'''
        return type(self)([v.__pos__() for v in self._list])
    #--- End: def

    def __abs__(self):
         '''
x.__abs__() <==> abs(x)
'''
         return type(self)([v.__abs__() for v in self._list])
    #--- End: def
    
    def __invert__(self):
         '''
x.__invert__() <==> ~x
'''
         return type(self)([v.__invert__() for v in self._list])
    #--- End: def
    
    def __add__(self, other):
        '''
x.__add__(y) <==> x+y
        '''
        return type(self)([v.__add__(other) for v in self._list])
    #--- End: def

    def __sub__(self, other):
        '''
x.__sub__(y) <==> x-y
        '''
        return type(self)([v.__sub__(other) for v in self._list])
    #--- End: def

    def __mul__(self, other):
        '''
x.__mul__(y) <==> x*y
        '''
        return type(self)([v.__mul__(other) for v in self._list])
    #--- End: def

    def __div__(self, other):
        '''
x.__div__(y) <==> x/y
'''
        return type(self)([v.__div__(other) for v in self._list])
    #--- End: def

    def __floordiv__(self, other):
        '''
x.__floordiv__(y) <==> x//y
'''
        return type(self)([v.__floordiv__(other) for v in self._list])
    #--- End: def


    def __truediv__(self, other):
        '''
x.__truediv__(y) <==> x/y
'''
        return type(self)([v.__truediv__(other) for v in self._list])
    #--- End: def

    def __radd__(self, other):
        '''
x.__radd__(y) <==> y+x
'''
        return type(self)([v.__radd__(other) for v in self._list])
    #--- End: def

    def __rmul__(self, other):
        '''
x.__rmul__(y) <==> y*x
'''
        return type(self)([v.__rmul__(other) for v in self._list])
    #--- End: def

    def __iadd__(self, other):
        '''
x.__iadd__(y) <==> x+=y
'''
        for v in self._list:
            v.__iadd__(other)
        return self
    #--- End def
        
    def __isub__(self, other):
        '''
x.__isub__(y) <==> x-=y
'''
        for v in self._list:
            v.__isub__(other)
        return self
    #--- End def

    def __imul__(self, other):
        '''
x.__imul__(y) <==> x*=y
'''
        for v in self._list:
            v.__imul__(other)
        return self
    #--- End def

    def __idiv__(self, other):
        '''
x.__idiv__(y) <==> x/=y
'''   
        for v in self._list:
            v.__idiv__(other)
        return self
    #--- End def

    def __ifloordiv__(self, other):
        '''
x.__ifloordiv__(y) <==> x//=y
'''   
        for v in self._list:
            v.__ifloordiv__(other)
        return self
    #--- End def

    def __itruediv__(self, other):
        '''
x.__itruediv__(y) <==> x/=y
'''   
        for v in self._list:
            v.__itruediv__(other)
        return self
    #--- End def

    @property
    def slice(self):
        '''
 
Slice each variable in the list, returning a new list of variables.

**Examples:**

>>> vl
[<CF Variable: air_temperature(73, 96)>,
 <CF Variable: air_temperature(73, 96)>]
>>> vl.slice[0,0]
[<CF Variable: air_temperature(1,1)>,
 <CF Variable: air_temperature(1,1)>]
'''
        return SliceVariableList(self)
    #--- End: def

    def _concatenate_data(self):
        '''

'''
        for v in self._list:
            v._concatenate_data()
    #--- End: def

    def dump(self, *arg, **kwargs):
        '''

Return a string containing the full descriptions of each variable in
the list.

.. seealso:: :func:`cf.dump`
'''
        return '\n\n\n'.join([v.dump(*arg, **kwargs) for v in self._list])
    #--- End: def

    def type(self):
        '''

The types of the data stored in each variable's Data object.

:returns: A list of list of type objects for each of the data.
elements.

 **Examples:**

>>> v.type()
[[<type 'netCDF4.Variable'>]]
>>> v.slice[...]
>>> v.type()
[[<type 'numpy.ndarray'>]]
'''
        return [v.type() for v in self._list]
    #--- End: def

#    def pub(self, *args, **kwargs):
#        '''
#
#Set a public attribute from each element of the list of variables or
#return a built-in list of the public attributes of each element of the
#list of variables.
#'''
#        if kwargs:
#            [v.pub(*args, **kwargs) for v in self._list]
#        else:
#            return [v.pub(*args) for v in self._list]
#    #--- End: def
#
#    def setpub(self, attr, value):
#        '''
#
#Set a public attribute from each element of the list of variables.
#'''
#        for v in self._list:
#            v.setpub(attr, value)
#    #--- End: def
#
#    def delpub(self, attr):
#        '''
#
#Delete a public attribute from each element of the list of variables.
#'''
#        for v in self._list:
#            v.delpub(attr)
#    #--- End: def
#
#    def getpub(self, *args):
#        '''
#
#Return a built-in list of the public attributes of each element of the
#list of variables.
#'''
#        return [v.getpub(*args) for v in self._list]
#    #--- End: def

    def setattr(self, attr, value):
        '''

Set a public attribute from each element of the list of variables.
'''
        for v in self._list:
            v.setattr(attr, value)
    #--- End: def

    def delattr(self, attr):
        '''

Delete a public attribute from each element of the list of variables.
'''
        for v in self._list:
            v.delattr(attr)
    #--- End: def

    def getattr(self, *args):
        '''

Return a built-in list of the public attributes of each element of the
list of variables.
'''
        return [v.getattr(*args) for v in self._list]
    #--- End: def

    def hasattr(self, *args):
        '''

Return a built-in list of the public attributes of each element of the
list of variables.
'''
        return [v.hasattr(*args) for v in self._list]
    #--- End: def

    def priv(self, *args, **kwargs):
        '''

Set a private attribute from each element of the list of variables or
return a built-in list of the private attributes of each element of
the list of variables.
'''
        if kwargs:
            [v.priv(*args, **kwargs) for v in self._list]
        else:
            return [v.priv(*args) for v in self._list]
    #--- End: def

    def setpriv(self, attr, value):
        '''

Set a private attribute on each element of the list of variables.
'''
        for v in self._list:
            v.setpriv(attr, value)
    #--- End: def

    def delpriv(self, attr):
        '''

Delete a private attribute from each element of the list of variables.
'''
        for v in self._list:
            v.delpriv(attr)
    #--- End: def

    def getpriv(self, attr, *default):
        '''

Return a built-in list of the private attributes of each element of the
list of variables.
'''
        return [v.getpriv(attr, *default) for v in self._list]
    #--- End: def

    def name(self, *arg, **kwargs):
        '''

Return a built-in list of the names of each element of the list of
variables.
'''
        return [v.name(*arg, **kwargs) for v in self._list]
    #--- End: def

    def match(self, *arg, **kwargs):        
        '''
 
Return a built-in list of the result of the match on each element of
the list of variables, with unambiguous abbreviations of phenomenon
names allowed.
'''
        return [v.match(*arg, **kwargs) for v in self._list]
    #--- End: def
       
    def ematch(self, *arg, **kwargs):
        '''

Return a built-in list of the result of the match on each element of
the list of variables, with only exact phenomenon names allowed.
'''
        return [v.ematch(*arg, **kwargs) for v in self._list]
    #--- End: def
       
    def extract(self, *arg, **kwargs):
        '''

Return the list of variables which match the given phenomena, with
unambiguous abbreviations of phenomenon names allowed.
'''

        return type(self)([v for v in self._list if v.match(*arg, **kwargs)])
    #--- End: def
  
    def eextract(self, *arg, **kwargs):
        '''

Return the list of variables which match the given phenomena, with
only exact phenomenon names allowed.
'''
        return type(self)([v for v in self._list if v.ematch(*arg, **kwargs)])
    #--- End: def
 
    def override_Units(self, *args, **kwargs):
        '''

Change the variables' units in place without changing the variables'
data.
'''
        for v in self._list:
            v.override_Units(*args, **kwargs)
    #--- End: def
  
#--- End: class


# ====================================================================
#
# Variable slice object
#
# ====================================================================

class SliceVariable(object):

    __slots__ = ['variable']

    def __init__(self, variable):
        '''
        
Set the contained variable.
'''
        self.variable = variable
    #--- End: def

    def __getitem__(self, index):
        '''
x.__getitem__(index) <==> x[index]
'''
        variable  = self.variable
        new       = variable.copy(data=False)
        new._data = variable._data.slice[index]

        return new
   #--- End: def

    def __setitem__(self, index, value):
        '''
x.__setitem__(index, value) <==> x[index] = value
'''
        self.variable._data.slice[index] = value
    #--- End: def

#--- End: class

# ====================================================================
#
# List of Variables slice object
#
# ====================================================================

class SliceVariableList(object):
    '''
'''
    __slots__ = ['variablelist']

    def __init__(self, variablelist):
        '''
        
Set the contained list of variables.
'''
        self.variablelist = variablelist
    #--- End: def

    def __getitem__(self, index):
        '''
x.__getitem__(index) <==> x[index]
'''
        variablelist = self.variablelist
        out          = [v.slice[index] for v in variablelist]
        return type(variablelist)(out)
    #--- End: def

    def __setitem__(self, index, value):
        '''
x.__setitem__(index, value) <==> x[index]=value
'''
        for v in self.variablelist:
            v.slice[index] = value
    #--- End: def

#--- End: class
