import netCDF4
import copy
import numpy
from .         import __Conventions__
from .variable import Variable
from .data     import Data
from .units    import Units
from .utils    import equals


def write(fields, file, format='NETCDF3_CLASSIC', 
          verbose=False, repoint=False, delete=False):
    '''

Write a field, or sequence of fields, to a CF compliant netCDF file.

NetCDF dimension and variable names will be taken from variables'
private attributes `ncvar` and the space attribute `nc_dimensions` if
present, otherwise they are inferred from standard names or set to
defaults. NetCDF names may be automatically given a numerical suffix
to avoid duplication.

Output netCDF file global attributes are those which occur in the set
of CF global attributes and non-standard data variable attributes and
which have equal values across all input fields.

It does not matter if an input field's data has not yet been read into
memory. The write operation will automatically read the data array
from disk if required, but the default is for it not to be saved in
the input field if it had to be read during the write
operation. However, if the `repoint` argument equals True then any
data array elements of an input field and its components which are
specified as file pointers (as opposed to arrays in memory) will be
changed in place so that they are arrays in memory. For example, for a
field, f:

>>> f.type()
[<type 'netCDF4.Variable'>]
>>> write(f, 'filename', repoint=True)
>>> f.type()
[<type 'netCDF4.Variable'>]
>>> write(f, 'filename', repoint=False)
>>> f.type()
[<type 'numpy.ndarray'>]

Identical field components (as determined by their `equals` methods)
are only written to the file once, apart from when they need to fulfil
both dimension coordinate and auxiliary coordinate roles for different
data variables.


:argument fields: The field or fields to write to the file.

:type fields: Field or sequence of fields

:argument file: The netCDF file to write fields to.

:type file: str

:parameter repoint: *Optional*. If False then any data array elements
                    of an input field and its components which are
                    file pointers will be changed in place so that
                    they are arrays in memory. If True then any data
                    array elements of an input field and its
                    components which are file pointers will be remain
                    as file pointers.

:type repoint: bool

:parameter delete: *Optional*. If True then delete each input field
                   from memory after it has been written to disk.

:type delete: bool

:parameter format: *Optional*. The format of the output netCDF
                   file. Valid formats are those supported by the
                   netCDF4 module: "NETCDF3_CLASSIC",
                   "NETCDF3_64BIT", "NETCDF4_CLASSIC" and
                   "NETCDF4". The default is
                   "NETCDF3_CLASSIC". Certain write operations may
                   be considerably faster using "NETCDF4_CLASSIC".

:type format: str

:parameter verbose: *Optional*. If True then print information to
                    stdout.

:type verbose: bool

:returns: None


**Examples:**

>>> f
[<CF Field: air_pressure(30, 24)>,
 <CF Field: u_compnt_of_wind(19, 29, 24)>,
 <CF Field: v_compnt_of_wind(19, 29, 24)>,
 <CF Field: potential_temperature(19, 30, 24)>]
>>> write(f, 'file')
'''

    # Initialize dictionary of useful variables
    g = {'netcdf'            : None,    # - netCDF4.Dataset instance
                                        
         'nc'                : {},      # - Dictionary of
                                        #   netCDF4.Variable instances
                                        #   keyed by their netCDF
                                        #   variable names
         'dim_to_ncdim'      : {},      # - Dictionary of netCDF
                                        #   dimension names keyed by
                                        #   field dimension names
         'ncdim_size'        : {},      # - Dictionary of dimension
                                        #   sizes keyed by netCDF
                                        #   dimension names
         'seen'              : {},      # - Dictionary netCDF variable
                                        #   names and netCDF
                                        #   dimensions keyed by
                                        #   elements of the field
                                        #   (such as a coordinate or a
                                        #   transform)
                                        
         'ncvar_names'       : set(()), # - set of all netCDF
                                        #   dimension and netCDF
                                        #   variable names
         'global_attributes' : set(()), # - Set of global or
                                        #   non-standard CF attributes
                                        #   which have identical
                                        #   values across all input
                                        #   fields.         

         'dim_coord_ncvars'  : set(()), # -
                                               
         'dimN'              : 1,       # - Counter
         'auxN'              : 1,       # - Counter
         'cmN'               : 1,       # - Counter
         'dataN'             : 1,       # - Counter
         'gmN'               : 1,       # - Counter
         'bndN'              : 1,       # - Counter
         'bnddimN'           : 1,       # - Counter
         'strlenN'           : 1,       # - Counter 
                                        
         'repoint'           : repoint,
         }

    # Setting delete to True forces repoint to be True
    if delete:
        g['repoint'] = True
 
    def _grid_ncdimensions(f, key, g=g):
        '''
'''
        space = f.space
        
        if space[key].ndim == 0:
            return ()
        else:
            return tuple(g['dim_to_ncdim'][dim] for dim in space.dimensions[key])
    #--- End: def
    
    def _grid_ncvar(f, key, default, counter, g=g):
        '''
'''
        coord = f.space[key]
        ncvar = getattr(coord, 'ncvar', coord.name(default=default))

        return _check_name(ncvar, counter)        
    #--- End: def
    
    def _create_dimension(ncdim, f, dim, g=g):
        '''

Create a netCDF dimension variable. Updates g['dim_to_ncdim']. Returns
None.
'''        
        ncdim_size = g['ncdim_size']

        ncdim_size[ncdim]      = f.space.dimension_sizes[dim]
        g['dim_to_ncdim'][dim] = ncdim

        g['netcdf'].createDimension(ncdim, ncdim_size[ncdim])
    #--- End: def

    def _create_dimension_coordinate(f, key, g=g):
        '''

Create a dimension coordinate variable, also creating:

 i) a new dimension variable if its not a scalar
ii) a new bounds dimension if required

Returns seen[id(f.space[key])]['ncvar'])
'''       
        seen  = g['seen']
        coord = f.space[key]

        create = False
        if not _seen(coord):            
            create = True
        elif seen[id(coord)]['ncdims'] != ():
            if seen[id(coord)]['ncvar'] != seen[id(coord)]['ncdims'][0]:
                # Already seen this coordinate but it was an auxiliary
                # coordinate, so it needs to be created as a dimension
                # coordinate.
                create = True
        #--- End: if

        if create:
            ncvar, g['dimN'] = _grid_ncvar(f, key, 'dim', g['dimN'])

            # Create a new dimension, if it is not a scalar coordinate
            if coord.ndim > 0:
                _create_dimension(ncvar, f, key, g=g)

            ncdimensions = _grid_ncdimensions(f, key)
            
            # If this dimension coordinate has bounds then create the
            # bounds netCDF variable and add the bounds or climatology
            # attribute
            extra = {}
            try:
                ncbounds = _create_bounds(f, key, ncdimensions, g=g)
            except AttributeError:
                pass
            else:
                try:
                    if coord.climatology:
                        extra['climatology'] = ncbounds
                    else:
                        extra['bounds'] = ncbounds
                except AttributeError:
                    extra['bounds'] = ncbounds
            #--- End: try

            # Create a new dimension coordinate variable
            _create_netcdf_variable(ncvar, ncdimensions, coord, 
                                    extra=extra, g=g)
            
#            # If this dimension coordinate has bounds then create the
#            # bounds netCDF variable and add the bounds or climatology
#            # attribute
#            try:
#                ncbounds = _create_bounds(f, key, ncdimensions, g=g)
#            except AttributeError:
#                pass
#            else:
#                try:
#                    if coord.climatology:
#                        g['nc'][ncvar].setncattr('climatology', ncbounds)
#                    else:
#                        g['nc'][ncvar].setncattr('bounds', ncbounds)
#                except AttributeError:
#                    g['nc'][ncvar].setncattr('bounds', ncbounds)
#            #--- End: try
        #--- End: if

        ncvar = seen[id(coord)]['ncvar']
        try:
            g['dim_to_ncdim'][key] = ncvar
        except KeyError:
            pass

        return ncvar
    #--- End: def

    def _create_auxiliary_coordinate(f, key, coordinates, g=g):
        '''

Create an auxiliary coordinate variable and create its bounds or
climatology variable if it has one. Return the coordinates list
updated with the ncvar for the auxiliary coordinate.
    
    '''
        coord = f.space[key]
        
        ncdimensions = _grid_ncdimensions(f, key)

        if not _seen(coord, ncdimensions):

            ncvar, g['auxN'] = _grid_ncvar(f, key, 'aux', g['auxN'])
            
            extra = {}
            try:
                ncbounds = _create_bounds(f, key, ncdimensions, g=g)
            except AttributeError:
                pass
            else:
                try:
                    if coord.climatology:
                        extra['climatology'] = ncbounds
                    else:
                        extra['bounds'] = ncbounds
                except AttributeError:
                    extra['bounds'] = ncbounds
            #--- End: try




            # Create a new auxiliary coordinate variable
            _create_netcdf_variable(ncvar, ncdimensions, coord,
                                    extra=extra, g=g)

            ## If this auxiliary coordinate has bounds then create the
            ## bounds netCDF variable and add the bounds or climatology
            ## attribute
            #try:
            #    ncbounds = _create_bounds(f, key, ncdimensions, g=g)
            #except AttributeError:
            #    pass
            #else:
            #    try:
            #        if coord.climatology:
            #            g['nc'][ncvar].setncattr('climatology', ncbounds)
            #        else:
            #            g['nc'][ncvar].setncattr('bounds', ncbounds)
            #    except AttributeError:
            #        g['nc'][ncvar].setncattr('bounds', ncbounds)
            ##--- End: try
        #--- End: if

        coordinates.append(g['seen'][id(coord)]['ncvar'])

        return coordinates
    #--- End: def

    def _create_bounds(f, key, coord_ncdimensions, g=g):
        '''

Create a coordinate's bounds netCDF variable, creating a new bounds
netCDF dimension if required. Return the bounds variable's netCDF
variable name.
'''
        coord = f.space[key]

        try:
            bounds = coord.bounds
            bounds._data
        except AttributeError as error:
            raise AttributeError, error

        # Still here? Then this coordinate has a bounds attribute
        # which contains data.
        size = bounds.shape[-1]

        ncdim, g['bnddimN'] = _check_name('bounds%(size)d' % locals(),
                                          g['bnddimN'], 
                                          dimsize=size)

        # Check if this bounds variable has not been previously
        # created.
        ncdimensions = coord_ncdimensions +(ncdim,)        
        if _seen(bounds, ncdimensions):
            return g['seen'][id(bounds)]['ncvar']    
        
        # Still here? Then this bounds variable has not been
        # previously created, so create it now.
        ncdim_size = g['ncdim_size']
        try:
            ncdim_size[ncdim]
        except KeyError:
            ncdim_size[ncdim] = size
            g['netcdf'].createDimension(ncdim, ncdim_size[ncdim])
        #--- End: try

        ncvar = getattr(bounds, 'ncvar', '%s_bounds' % coord.ncvar)

        ncvar, g['bndN'] = _check_name(ncvar, g['bndN'])

        # Note that is expected that bounds have equals units to their
        # parent coordinate

        # Create the bounds netCDF variable
        _create_netcdf_variable(ncvar, ncdimensions, bounds, 
                                omit=('units', 'calendar'), g=g)

        return ncvar
    #--- End: def
    
    def _seen(variable, ncdims=None, g=g):
        '''

Return True if a cf variable identical to the input 'variable' (via
variable.equals) is already in the 'seen' dictionary (and has
therefore already been written to the output file), otherwise return
False. When True is returned, the 'seen' dictionary is updated for the
input 'variable'.

If 'ncdims' is set then a further condition for returning True is that
'ncdims' is equal to netCDF dimensions (names and order) of the
identical variable in the 'seen' dictionary.
'''
        seen = g['seen']

        for ident in seen:
            if ncdims is not None and ncdims != seen[ident]['ncdims']:
                continue

            # Still here?
            if variable.equals(seen[ident]['variable']):
                seen[id(variable)] = {'variable': variable,
                                      'ncvar'   : seen[ident]['ncvar'],
                                      'ncdims'  : seen[ident]['ncdims']}
                return True
        #--- End: for

        return False
    #--- End: def

    def _datatype(variable):
        '''

Return the netCDF4.createVariable datatype corresponding to the
datatype of the array stored (possibly virtually) in the input
variable's 'data' attribute. E.g. if variable.dtype is 'float32', then
'f4' will be returned.

Numpy string data types will return 'S1' regardless of the numpy
string length. This means that the required conversion of
multi-character datatype numpy arrays into single-character datatype
numpy arrays (with an extra trailing dimension) is expected to be done
elsewhere (currently in def _create_netcdf_variable).

If the input variable has no data attribute then 'S1' is returned.
'''
#        try:
#            variable.dtype
#        except AttributeError:
#            return 'S1'

        if (variable.dtype is None     or
            variable.dtype.char == 'S' or
            not hasattr(variable, 'dtype')):
            return 'S1'            

        # Still here?
        return '%s%s' % (variable.dtype.kind, variable.dtype.itemsize)

        #if variable.dtype == 'float32': return 'f4'
        #if variable.dtype == 'float64': return 'f8'
        #if variable.dtype == 'float'  : return 'f8'
        #if variable.dtype == 'int32'  : return 'i4'
        #if variable.dtype == 'int64'  : return 'i8'
        #if variable.dtype == 'int8'   : return 'i1'
        #if variable.dtype == 'int16'  : return 'i2'
        #if variable.dtype is None     : return 'S1'        
        #if variable.dtype.char == 'S' : return 'S1'            
        #
        #return None
    #--- End: def

    def _offset_scale(array, add_offset, scale_factor):
        '''
'''
        if add_offset is not None or scale_factor is not None:
            array = array.copy()

        if add_offset is not None:                
            array -= add_offset

        if scale_factor is not None:
            array /= scale_factor

        return array
    #--- End: def

    def _create_netcdf_variable(ncvar, ncdimensions, input, 
                                omit=(), extra={}, g=g):
        '''

Create a netCDF variable from 'input' with name 'ncvar' and dimensions
'ncdimensions'. The netCDF variable's attributes are given by
input.properties(), less any given by the 'omit' argument. If a new
string-length netCDF dimension is required then it will also be
created. The 'seen' dictionary is updated for 'input'.

'''
        # Set the netCDF4.createVariable datatype
        datatype = _datatype(input)

        dimensions = ncdimensions
        
        data_out = None

        if input._data:
            # Convert a multi-character datatype numpy array into
            # single-character datatype numpy array with an extra trailing
            # dimensiong
            strlen = False
            if datatype == 'S1':
                 strlen = input.dtype.itemsize
                 if strlen > 1:

                     # Create a new dimension for the maximum string length
                     ncdim, g['strlenN'] = _check_name('strlen%(strlen)d' % locals(), 
                                                       g['strlenN'], 
                                                       dimsize=strlen)
            
                     ncdim_size = g['ncdim_size']
                     try:
                         ncdim_size[ncdim]
                     except KeyError:
                         ncdim_size[ncdim] = strlen
                         g['netcdf'].createDimension(ncdim, ncdim_size[ncdim])
                     #--- End: try

                     dimensions = ncdimensions + (ncdim,)
                     
                     units    = Units()
                     data_out = Data(units=units)
                     
                     # Conform the data
                     input._data.conform()

                     for array in input._data._list:
                         temp0 = array.flatten()
                         temp1 = numpy.empty((temp0.size, ncdim_size[ncdim]), '|S1')
                         for i in xrange(temp0.size):
                             temp1[i,:] = list(temp0[i].ljust(strlen, ' ')) 
                         data_out.append(Data(temp1.reshape(array.shape + (ncdim_size[ncdim],)),
                                              units=units))
                     #--- End: for
                     data_out.shape     = tuple(data_out.elements[0].shape)
                     data_out.order     = data_out.elements[0].order[:]
                     data_out.direction = dict(((d, True) for d in data_out.order))
                     
                     # Delete unwanted numpy arrays
                     del temp0
                     del temp1
            #--- End: if
        #--- End: if

        if data_out is None:
            data_out = input._data

        # Find the fill value (note that this is set in the call to
        # netCDF4.createVariable, rather than with setncattr).
        fill_value = getattr(input, '_FillValue', None)

        # Create a new netCDF variable        
        g['nc'][ncvar] = g['netcdf'].createVariable(ncvar, datatype, 
                                                    dimensions,
                                                    fill_value=fill_value)

        #-------------------------------------------------------------
        # Find offset and scale factors, if any
        #-------------------------------------------------------------
        add_offset   = input.getattr('add_offset',   None)   
        scale_factor = input.getattr('scale_offset', None)
        if add_offset == 0.0:
            add_offset = None
        if scale_factor == 1.0:
            scale_factor = None

        # Add simple attributes (and units and calendar) to the netCDF
        # variable
        netcdf_attrs = input._simple_attributes()        
        for attr in ('units', 'calendar'):
            if hasattr(input, attr):
                netcdf_attrs[attr] = getattr(input, attr)
        #--- End: for
        netcdf_attrs.update(extra)
        for attr in omit:
            if attr in netcdf_attrs:
                netcdf_attrs.pop(attr) 
        try:
            # Try using the faster setncatts
            g['nc'][ncvar].setncatts(netcdf_attrs)
        except AttributeError:
            # Otherwise use the slower setncattr
            for attr, value in netcdf_attrs.iteritems():
                g['nc'][ncvar].setncattr(attr, value)

        #-------------------------------------------------------------
        # Add data to the netCDF variable
        #-------------------------------------------------------------
        repoint = g['repoint']
        len_data = len(data_out)
        if len_data == 1:

            data = data_out[0]

            data.conform()
            array = data.varray

            array = _offset_scale(array, add_offset, scale_factor)

            g['nc'][ncvar][...] = array
           
            if not repoint:
                data_out[0] = data

        elif len_data > 1:
            adim  = input._data.adim
            index = [slice(None)] * input.ndim
            start = 0
            
            for i, data in enumerate(data_out):

                data.conform()
                array = data.varray
                array = _offset_scale(array, add_offset, scale_factor)

                stop        = start + array.shape[adim]
                index[adim] = slice(start, stop)
                start       = stop   

                g['nc'][ncvar][tuple(index)] = array
                
                if not repoint:
                    data_out[i] = data
            #--- End: for
        #--- End: if
                
        # Update the 'seen' dictionary
        g['seen'][id(input)] = {'variable': input,
                                'ncvar'   : ncvar,
                                'ncdims'  : ncdimensions,
                                }
    #--- End: def

    def _create_grid_mapping(ncvar, input, g=g):
        '''
    
Create a netCDF grid_mapping variable from :obj:`input` with name
:obj:`ncvar`. The netCDF variable's attributes are given by key/value
pairs of input.iteritems(). The 'seen' dictionary is updated for
'input'.
'''
        # Create a new netCDF variable
        g['nc'][ncvar] = g['netcdf'].createVariable(ncvar, 'S1', ())

        # Add attributes from key/value pairs
        try:
            # Try using the faster setncatts
            g['nc'][ncvar].setncatts(input)
        except AttributeError:
            # Otherwise use the slower setncattr
            for attr, value in input.iteritems():
                g['nc'][ncvar].setncattr(attr, value)

        # Update the 'seen' dictionary
        g['seen'][id(input)] = {'variable': input, 
                                'ncvar'   : ncvar, 
                                'ncdims'  : (), # Transforms have no dimensions
                                }
    #--- End: def

    def _check_name(base, counter, dimsize=None, g=g):
        '''


'''
        ncvar_names = g['ncvar_names']

        if dimsize is not None:
            if base in ncvar_names and dimsize == g['ncdim_size'][base]:
                return base, counter

        if base in ncvar_names:
            ncvar = '%(base)s_%(counter)d' % locals()
            while ncvar in ncvar_names:
                counter += 1
                ncvar = '%(base)s_%(counter)d' % locals()
        else:
            ncvar = base

        ncvar_names.add(ncvar)

        return ncvar, counter
    #--- End: def
    
    def _write_a_field(f, save=False, add_to_seen=False, g=g):
        '''

'''
        seen = g['seen']

        space = f.space

        if save:
            saved_dim_to_ncdim = copy.deepcopy(g['dim_to_ncdim'])

        g['dim_to_ncdim'] = {}

        # Initialize the list of the field's auxiliary coordinates
        coordinates = []

        for key in sorted(space.dimension_sizes):            
            try:
                space[key]
            except KeyError:
                # There is no dimension coordinate for this dimension,
                # so just create a netCDF dimension if required.
                try:
                    ncdim = space.nc_dimensions[key]
                except (AttributeError, KeyError):
                    ncdim = 'dim'

                if ncdim not in g['dim_coord_ncvars']:
                    g['dim_to_ncdim'][key] = ncdim 
                    if ncdim not in g['ncvar_names']:
                        # Create the new dimension
                        _create_dimension(ncdim, f, key, g=g)
                        g['ncvar_names'].add(ncdim)
                    continue

                ncdim, g['dimN'] = _check_name(ncdim, g['dimN'])
##                                               dimsize=space.dimension_sizes[key])

                if (space.dimension_sizes[key] > 1 or 
                    key in space.dimensions['data']):
                    # Create the new dimension
                    _create_dimension(ncdim, f, key, g=g)

            else:              
                # A dimension coordinate exists for this dimension, so
                # create a netCDF dimension and a netCDF coordinate
                # netCDF variable. 
                ncdim = _create_dimension_coordinate(f, key, g=g)
                g['dim_coord_ncvars'].add(ncdim)

                # Add a scalar coordinate to the list of auxiliary
                # coordinates
                if space[key].ndim == 0:
                    coordinates.append(ncdim)
            #--- End: try

        #--- End: for
                
        # ------------------------------------------------------------
        # Create auxiliary coordinate variables, except those which
        # might be completely specified elsewhere by a transformation.
        # ------------------------------------------------------------
        # Initialize the list of 'coordinates' attribute variable
        # values (each of the form 'name')
        for key in sorted(space.get_keys('^aux')):
            coordinates = _create_auxiliary_coordinate(f, key, 
                                                       coordinates,
                                                       g=g)
        #--- End: for

        # ------------------------------------------------------------
        # Create cell measures variables
        # ------------------------------------------------------------
        # Initialize the list of 'cell_measures' attribute values
        # (each of the form 'measure: name')
        cell_measures = []
        for key in sorted(space.get_keys('^cm')):
            
            ncdimensions = _grid_ncdimensions(f, key)
            if not _seen(space[key], ncdimensions):
                ncvar, g['cmN'] = _grid_ncvar(f, key, 'cm', g['cmN'])

                # Create a new cell_measures variable
                _create_netcdf_variable(ncvar, ncdimensions, space[key],
                                        g=g)
                
            cell_measures.append('%s: %s' % (space[key].standard_name,
                                             seen[id(space[key])]['ncvar']))
        #--- End: for

        # ------------------------------------------------------------
        # Create grid mapping and formula terms variables
        # ------------------------------------------------------------
        # Initialize the list of 'grid_mapping' attribute variable
        # values (each of the form 'name')
        grid_mapping  = []
        formula_terms = []
        try:
            transforms = space.transform.values()
        except AttributeError:
            pass
        else:     

            for transform in transforms:
                
                try:
                    grid_mapping_name = transform['grid_mapping_name']
                except KeyError:
                    # ------------------------------------------------
                    # This transform is a formula terms
                    # ------------------------------------------------

                    key = space.coord(transform.name, exact=True, key=True)
                    if key is None:
                        # This formula_terms doesn't match up with an
                        # existing coordinate
                        continue
                    
                    # Still here? Then create the formula_terms
                    # attribute string
                    for term, var in transform.iteritems():
                        
                        if var is None:
                            # Do not create broken formula_terms
                            formula_terms = None
                            break
                    
                        if isinstance(var, Variable):
                            if not seen.has_key(id(var)):
                                # var is a variable which hasn't been
                                # written yet (set save=True so we
                                # don't lose the current the
                                # dim_to_ncdim dictionary).
                                _write_a_field(var, save=True, add_to_seen=True, g=g)
                        elif space.has_key(var):
                            # var is a key of the space, such as 'dim1'
                            var = space[var]
                        else:
                            # Do not create broken formula_terms
                            formula_terms = None
                            break
                        #--- End: if
                       
                        # Record the netCDF variable name for the
                        # formula_terms string
                        ncvar = seen[id(var)]['ncvar']
                        formula_terms.append('%(term)s: %(ncvar)s' % locals())
                   #--- End: for
            
                    # Add the formula_terms attribute to the output
                    # variable
                    if formula_terms:
                        ncvar = seen[id(space[key])]['ncvar']
                        g['nc'][ncvar].setncattr('formula_terms', ' '.join(formula_terms))

                else:           
                    # ------------------------------------------------
                    # This transform is a grid mapping
                    # ------------------------------------------------
                    if _seen(transform):
                        ncvar = seen[id(transform)]['ncvar']
                    else:
                        # It has not yet been witten to the file, so
                        # create it
                        try:
                            ncvar = transform.ncvar
                        except AttributeError:
                            ncvar = grid_mapping_name
                        ncvar, g['gmN'] = _check_name(ncvar, g['gmN']) 
                            
                        _create_grid_mapping(ncvar, transform, g=g)
                        
                        ncvar = seen[id(transform)]['ncvar']
                    #--- End: if
                    grid_mapping.append(ncvar)
                #--- End: try

            #--- End: for

        #--- End: try

        # ------------------------------------------------------------
        # Create ancillary variables
        # ------------------------------------------------------------
        # Initialize the list of 'ancillary_variables' attribute
        # values
        ancillary_variables = []
        if hasattr(f, 'ancillary_variables'):
            for av in f.ancillary_variables:
                _write_a_field(av, save=True, add_to_seen=True, g=g)
                ancillary_variables.append(seen[id(av)]['ncvar'])
        #--- End: if

        # ------------------------------------------------------------
        # Create the data variable
        # ------------------------------------------------------------
        ncvar = getattr(f, 'ncvar', f.name(default='data'))
            
        ncvar, g['dataN'] = _check_name(ncvar, g['dataN'])

        dim_to_ncdim = g['dim_to_ncdim']
        ncdimensions = tuple(dim_to_ncdim[dim] 
                             for dim in space.dimensions['data'])

        extra = {}

        # Cell measures (in: list, out: str)
        if cell_measures:
            extra['cell_measures'] = ' '.join(cell_measures)           

        # Auxiliary/scalar coordinates (in: list, out: str)
        if coordinates:
            extra['coordinates']   = ' '.join(coordinates)

        # Grid mapping (in: list, out: str)
        if grid_mapping: 
            extra['grid_mapping']  = ' '.join(grid_mapping)

        # Ancillary variables (in: list, out: str)
        if ancillary_variables:
            extra['ancillary_variables'] = ' '.join(ancillary_variables)
            
        # Flag values (in: array, out: array)
        if hasattr(f, 'flag_values'):
            extra['flag_values'] = self.flag_values

        # Flag masks (in: array, out: array)
        if hasattr(f, 'flag_masks'):
            extra['flag_masks'] = self.flag_masks

        # Flag meanings  (in: array, out: str)
        if hasattr(f, 'flag_meanings'):
            extra['flag_meanings'] = ' '.join(f.flag_meanings)

        # Cell methods (in: CellMethods, out: str)
        try:
            cell_methods = copy.deepcopy(f.cell_methods)
        except AttributeError:
            pass
        else:
            for cm in cell_methods:
                for j in xrange(len(cm['dim'])):
                    dim = cm['dim'][j]
                    if dim not in (None, False) and not cm['no_coords'][j]:
                        cm['name'][j] = seen[id(space[dim])]['ncvar']
            #--- End: for
            extra['cell_methods'] = ' '.join(cell_methods.strings())
        #--- End: try

        # Create a new data variable
        _create_netcdf_variable(ncvar, ncdimensions, f,
                                omit  = g['global_attributes'],
                                extra = extra,
                                g=g)
        
#        try:
#            g['nc'][ncvar].flag_meanings = ' '.join(g['nc'][ncvar].flag_meanings)
#        except AttributeError:
#            pass
#       
#        if cell_measures:
#            g['nc'][ncvar].cell_measures = ' '.join(cell_measures)           
#        if coordinates:
#            g['nc'][ncvar].coordinates   = ' '.join(coordinates)
#        if grid_mapping:
#            g['nc'][ncvar].grid_mapping  = ' '.join(grid_mapping)
#
#        # CELL METHODS
#        try:
#            cell_methods = copy.deepcopy(f.cell_methods)
#        except AttributeError:
#            pass
#        else:
#            for cm in cell_methods:
#                for j in xrange(len(cm['dim'])):
#                    dim = cm['dim'][j]
#                    if dim not in (None, False) and not cm['no_coords'][j]:
#                        cm['name'][j] = seen[id(space[dim])]['ncvar']
#            #--- End: for
#            g['nc'][ncvar].cell_methods = ' '.join(cell_methods.strings())
#
#            for i in xrange(len(cell_methods)):
#                for j in xrange(len(cell_methods[i]['dim'])):
#                    dim = cell_methods[i]['dim'][j]
#                    if dim not in (None, False) and not cell_methods[i]['no_coords'][j]:
#                        cell_methods[i]['name'][j] = seen[id(space[dim])]['ncvar']
#            g['nc'][ncvar].cell_methods = ' '.join(cell_methods.strings())
            
        # ------------------------------------------------------------
        # If we are saving the original dim_to_ncdim then copy back
        # the original
        # ------------------------------------------------------------
        if save:
            g['dim_to_ncdim'] = copy.deepcopy(saved_dim_to_ncdim)        

        # Update the 'seen' dictionary, if required
        if add_to_seen:
            g['seen'][id(f)] = {'variable': f,
                                'ncvar'   : ncvar,
                                'ncdims'  : ncdimensions,
                                }
    #--- End: def
            
    def _create_global_attributes(fields, g=g):
        '''

Find the input fields' global attributes and write them to the
netCDF4.Dataset. These are the global or non-standard CF attributes
which have identical values across all input fields. Returns None.
'''
        # Data variable attributes, as defined in Appendix A
        data_attributes = set(('add_offset',
                               'ancillary_variables',
                               'cell_measures',
                               'cell_methods',
                               'coordinates',
                               '_FillValue',
                               'flag_masks',
                               'flag_meanings',
                               'flag_values',
                               'grid_mapping',
                               'long_name',
                               'missing_value',
                               'scale_factor',
                               'standard_error_multiplier',
                               'standard_name',
                               'units',
                               'valid_max',
                               'valid_min',
                               'valid_range',
                               ))

        # Global attributes, as defined in Appendix A
        g['global_attributes'] = set(('comment',
                                      'Conventions',
                                      'history',
                                      'institution',
                                      'references',
                                      'source',
                                      'title',
                                      ))
        global_attributes = g['global_attributes'] 

        # Put all non-standard CF attributes (i.e. those not in the
        # data_attributes set) into the global_attributes set.
        for f in fields:
            for attr in set(f._simple_attributes()) - global_attributes:
                if attr not in data_attributes:
                    global_attributes.add(attr)
        #--- End: for

        # Remove attributes from the global_attributes set which have
        # different values in different data variables
        f0 = fields[0]
        for attr in list(global_attributes):
            try:
                f0.getattr(attr)
            except AttributeError:
                global_attributes.remove(attr)
        #--- End: for

        if len(fields) > 1:
            for attr in list(global_attributes):
                attr0 = f0.getattr(attr)
                for f in fields[1:]:
                    try:                    
                        if not equals(f.getattr(attr), attr0):
                            global_attributes.remove(attr)
                            break
                    except AttributeError:
                        global_attributes.remove(attr)
                        break
            #--- End: for
        #--- End: if

        try:
            global_attributes.remove('Conventions')
        except KeyError:
            pass
        g['netcdf'].setncattr('Conventions', __Conventions__)

        for attr in global_attributes:
            g['netcdf'].setncattr(attr, f0.getattr(attr))

        # Put 'Conventions' back into the global attributes set so
        # that we know to exclude it from data variables' attributes
        global_attributes.add('Conventions')
    #--- End: def

    # ---------------------------------------------------------------
    # Open the output netCDF file. (NETCDF3_CLASSIC can be *really*
    # slow)
    # ---------------------------------------------------------------
    g['netcdf'] = netCDF4.Dataset(file, 'w', format=format)

    # ---------------------------------------------------------------
    # Set the fill mode for a Dataset open for writing to off. This
    # will prevent the data from being pre-filled with fill values,
    # which may result in some performance improvements.
    # ---------------------------------------------------------------
    g['netcdf'].set_fill_off()

    # ---------------------------------------------------------------
    # Write global attributes to the file first. This is important as
    # doing it later slows things down enormously. This function also
    # creates the g['global_attributes'] set, which is used in def
    # _write_a_field.
    # ---------------------------------------------------------------
    _create_global_attributes(fields, g=g)

    # ---------------------------------------------------------------
    # ---------------------------------------------------------------
    for f in fields[:]:
        _write_a_field(f, g=g)
        if verbose:
            print repr(f)
        if delete:
            del fields[0]

    # ---------------------------------------------------------------
    # Write all of the buffered data to disk
    # ---------------------------------------------------------------
    g['netcdf'].close()

    if verbose:
        n = len(fields)
        print 'Wrote',repr(n),'field'+('s' if n>1 else '')+' to',file
#--- End: def
